package uchicago.src.guiUtils;

import javax.swing.*;
import javax.swing.text.*;
import java.awt.event.*;
import java.awt.Color;

import cern.colt.map.OpenIntObjectHashMap;

public class CodeDocument extends DefaultStyledDocument {

  private Segment txt = new Segment();
  private Tokenizer tokenizer;;
  private OpenIntObjectHashMap styleMap = new OpenIntObjectHashMap();
  private Style basic;
  private int lastTokenType = Token.EOF;

  public CodeDocument(Tokenizer tokenizer) {
    this.tokenizer = tokenizer;

    basic = addStyle("text", null);
    StyleConstants.setFontFamily(basic, "Monospaced");
    StyleConstants.setFontSize(basic, 12);
    StyleConstants.setBackground(basic, Color.white);
    StyleConstants.setForeground(basic, Color.black);
    StyleConstants.setBold(basic, false);
    styleMap.put(Token.TEXT, basic);

    Style s = addStyle("error", basic);
    StyleConstants.setForeground(s, Color.red);


    s = addStyle("keyword", basic);
    StyleConstants.setForeground(s, Color.blue);
    styleMap.put(Token.KEYWORD, s);

    s = addStyle("string", basic);
    StyleConstants.setForeground(s, Color.green);
    styleMap.put(Token.STRING, s);

    s = addStyle("number", basic);
    StyleConstants.setForeground(s, Color.magenta);
    styleMap.put(Token.NUMBER, s);

    s = addStyle("comment", basic);
    StyleConstants.setForeground(s, Color.orange);
    styleMap.put(Token.COMMENT, s);
  }

  public int highlightError(int lineNum) {
    Element[] e = getRootElements();
    //System.out.println(e[0]);
    //System.out.println(lineNum);
    Element line = e[0].getElement(lineNum - 1);
    int length = line.getEndOffset() - line.getStartOffset();
    setCharacterAttributes(line.getStartOffset(), length, getStyle("error"),
                           true);
    return line.getStartOffset();
  }

  public void insertString(int offs, String str, AttributeSet a)
          throws BadLocationException {
    if (str.equals("\t")) {
      str = "  ";
    }
    //this calculates how many tabs are present.
    if (offs > 0 && str.equals("\n")) {
      StringBuffer buf = new StringBuffer("\n");
      Element e = getParagraphElement(offs);
      //get the current line
      String para = getText(e.getStartOffset(), e.getEndOffset() - e.getStartOffset());
      int numTabs = 0;
      //find out how many pairs of spaces are at the beginning
      for (int i = 0; i < para.length(); i++) {
        if (para.charAt(i) == ' ') {
          //add one for each space
          numTabs++;
        } else {
          break;
        }
      }
      //get the number of pairs
      if (numTabs > 0) numTabs = numTabs / 2;
      //if the previous line was a :, add a new tab
      if (getText(offs - 1, 1).equals(":")) {
        numTabs++;
      }
      //append a new tab for each one
      for (int i = 0; i < numTabs; i++) {
        buf.append("  ");
      }
      //kick back the string
      str = buf.toString();
    }
    try {
      super.insertString(offs, str, a);
    } catch (BadLocationException ex) {
      ex.printStackTrace();
      System.exit(0);
      throw ex;
    }

    if (str.length() > 1)
      colorBlock(offs, str, a);
    else
      colorLine(offs, str, a);

  }

  private void colorBlock(int offs, String str, AttributeSet a)
          throws BadLocationException {
    tokenizer.setString(str);
    try {
      Token t = tokenizer.nextToken();
      while (t.type != Token.EOF) {
        AttributeSet as = (AttributeSet) styleMap.get(t.type);
        if (as == null) as = basic;
        int start = t.start + offs;
        /*
          System.out.println(t.type + ": " + t.text +
          " at " + t.lineEnd + ", " + start + ", " +
          t.length);
        */
        setCharacterAttributes(start, t.length, as, true);
        t = tokenizer.nextToken();
      }
    } catch (java.io.IOException ex) {
      ex.printStackTrace();
    }
  }


  private void colorLine(int offs, String str, AttributeSet a)
          throws BadLocationException {
    if (!str.equals("\n")) {
      // gets the line at this offset.
      Element e = getParagraphElement(offs);
      int length = e.getEndOffset() - e.getStartOffset();
      getText(e.getStartOffset(), length, txt);
      //setCharacterAttributes(e.getStartOffset(),length, basic, false);
      int offset = e.getStartOffset();
      tokenizer.setString(txt.toString());

      try {
        Token t = tokenizer.nextToken();
        while (t.type != Token.EOF) {
          AttributeSet as = (AttributeSet) styleMap.get(t.type);
          if (as == null) as = basic;
          int start = t.start + offset;
          /*
           System.out.println(t.type + ": " + t.text +
               " at " + t.lineEnd + ", " + start + ", " +
               t.length);
          */
          setCharacterAttributes(start, t.length, as, true);
          t = tokenizer.nextToken();
        }
      } catch (java.io.IOException ex) {
        ex.printStackTrace();
      }
    }
  }

  public void remove(int offs, int len) throws BadLocationException {
    super.remove(offs, len);
    // recolorize rest of line
    Element e = getParagraphElement(offs);
    int length = e.getEndOffset() - e.getStartOffset();
    getText(e.getStartOffset(), length, txt);
    //setCharacterAttributes(e.getStartOffset(),length, basic, false);
    int offset = e.getStartOffset();
    tokenizer.setString(txt.toString());

    try {
      Token t = tokenizer.nextToken();
      while (t.type != Token.EOF) {
        AttributeSet as = (AttributeSet) styleMap.get(t.type);
        if (as == null) as = basic;
        int start = t.start + offset;
        /*
          System.out.println(t.type + ": " + t.text +
          " at " + t.lineEnd + ", " + start + ", " +
          t.length);
        */
        setCharacterAttributes(start, t.length, as, true);
        t = tokenizer.nextToken();
      }
    } catch (java.io.IOException ex) {
      ex.printStackTrace();
    }
  }

  public static void main(String[] args) {
    JFrame f = new JFrame("CodeDocument Test");
    f.setSize(400, 500);
    f.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent evt) {
        System.exit(0);
      }
    });
    Tokenizer t = new uchicago.src.simbuilder.NQPyTokenizer();
    CodeDocument doc = new CodeDocument(t);
    Style s = doc.getStyle("text");
    JTextPane pane = new JTextPane(doc);
    f.getContentPane().add(pane);
    f.setVisible(true);
    pane.requestFocus();
  }
}

