package uchicago.src.nqpy.compiler;

import uchicago.src.nqpy.parser.SimpleNode;
import com.go.trove.classfile.Label;

public abstract class AbstractEmitter implements Emitter {

  protected NQPyClass nqPyClass;
  
  public AbstractEmitter(NQPyClass nqclass) {
    nqPyClass = nqclass;
  }

  /**
   * Gets the result type (as an NQPyClass) of arithmetic operations
   * on this Emitter.
   */
  public NQPyClass getArithResType(NQPyClass rhs) {
    String message = "Arithmetic operations not supported between " +
      nqPyClass.getJClassName() + " and " + rhs.getJClassName();
    throw ErrorFactory.createCompilerException(message);
  }

  /**
   * Gets the result type (as an NQPyClass) of single arg arithmetic operations
   * on this Emitter.
   */
  public NQPyClass getArithResType() {
    String message = "Arithmetic operation not supported on " +
      nqPyClass.getJClassName();
    throw ErrorFactory.createCompilerException(message);
  }
  
  /**
   * Emits the appropriatxe byte code to go to the specified branch
   * if the specified node compiles to true. This tests whether
   * the result of node.compile() is equal to zero, and branches if the
   * result is not equal to zero.
   */
  public void branchOnTrue(SimpleNode node, NQPyBlockCompiler compiler,
		   Label branch)
  {
    String message = "branchOnTrue not supported on " + this;
    throw ErrorFactory.createCompilerException(message);
  }


  private void throwArithError(String type) {
    String message = "Arithmetic operations not supported between " +
      nqPyClass.getJClassName() + " and " + type;
    throw ErrorFactory.createCompilerException(message);
  }

  private void throwArithError() {
    String message = "Arithmetic operation not supported on " +
      nqPyClass.getJClassName();
    throw ErrorFactory.createCompilerException(message);
  }
    
  
  /**
   * Emits the appropriate byte code to go to the specified branch
   * if the single node in nodes compiles to false. This tests whether
   * the result of node.compile() is equal to zero, and branches if so.
   */
  public void branchOnFalse(SimpleNode node, NQPyBlockCompiler compiler,
		   Label branch)
  {
    String message = "branchOnFalse not supported on " + this;
    throw ErrorFactory.createCompilerException(message);
  }

  /**
   * Branches to the specified branch if the specified nodes are equal.<p>
   *
   * This checks the type of rhs and then calls the
   * the appropriate eqCmp on this NQPyClass.
   */
  public void eq(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn, Label branch,
		 NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      eqCmpInt(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      eqCmpDouble(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      eqCmpLong(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      eqCmpFloat(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      eqCmpString(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.BOOLEAN_CLASS:
      eqCmpBoolean(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.JAVA_CLASS:
      eqCmpJava(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DYNAMIC_CLASS:
      eqCmpJava(lhsn, rhsn, branch, compiler);
      break;
    default:
      String message = "Cannot compare " + this + " and " + rhs;
      throw ErrorFactory.createCompilerException(message);
    }
  }

  public void eqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    throwCmpError("int");
  }

  public void eqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("double");
  }

  public void eqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError("long");
  }

  public void eqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    throwCmpError("float");
  }

  public void eqCmpBoolean(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			   NQPyBlockCompiler compiler)
  {
    throwCmpError("boolean");
  }

  public void eqCmpString(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("java.lang.String");
  }

  public void eqCmpJava(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError(rhsn.getNQPyClass(compiler).getJClassName());
  }

  /**
   * Branches to the specified branch if the specified nodes are not
   * equal.<p>
   *
   * This checks the type of rhs and then calls the the appropriate notEqCmp
   * on this NQPyClass.*
   */
  public void notEq(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		    Label branch, NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      notEqCmpInt(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      notEqCmpDouble(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      notEqCmpLong(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      notEqCmpFloat(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      notEqCmpString(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.BOOLEAN_CLASS:
      notEqCmpBoolean(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.JAVA_CLASS:
      System.out.println("here " + this);
      notEqCmpJava(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DYNAMIC_CLASS:
      notEqCmpJava(lhsn, rhsn, branch, compiler);
      break;
    default:
      String message = "Cannot compare " + this + " and " + rhs;
      throw ErrorFactory.createCompilerException(message);
    }
  }

  public void notEqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    throwCmpError("int");
  }

  public void notEqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("double");
  }

  public void notEqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError("long");
  }

  public void notEqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    throwCmpError("float");
  }

  public void notEqCmpBoolean(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			   NQPyBlockCompiler compiler)
  {
    throwCmpError("boolean");
  }

  public void notEqCmpString(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("java.lang.String");
  }

  public void notEqCmpJava(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError(rhsn.getNQPyClass(compiler).getJClassName());
  }

  private void throwCmpError(String type) {
    String message = "Cannot compare " + nqPyClass.getJClassName() +
      " and " + type;
    throw ErrorFactory.createCompilerException(message);
  }

  /**
   * Branches to the specified branch if the lhs node > rhs node.<p>
   *
   * This checks the type of rhs and then calls the the appropriate gtCmp
   * on this NQPyClass.*
   */
  public void gt(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		    Label branch, NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      gtCmpInt(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      gtCmpDouble(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      gtCmpLong(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      gtCmpFloat(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      gtCmpString(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.BOOLEAN_CLASS:
      gtCmpBoolean(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.JAVA_CLASS:
      gtCmpJava(lhsn, rhsn, branch, compiler);
      break;
    default:
      String message = "Cannot compare " + this + " and " + rhs;
      throw ErrorFactory.createCompilerException(message);
    }
  }

  public void gtCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    throwCmpError("int");
  }

  public void gtCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("double");
  }

  public void gtCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError("long");
  }

  public void gtCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    throwCmpError("float");
  }

  public void gtCmpBoolean(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			   NQPyBlockCompiler compiler)
  {
    throwCmpError("boolean");
  }

  public void gtCmpString(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("java.lang.String");
  }

  public void gtCmpJava(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError(rhsn.getNQPyClass(compiler).getJClassName());
  }

  /**
   * Branches to the specified branch if the lhs node < rhs node.<p>
   *
   * This checks the type of rhs and then calls the the appropriate ltCmp
   * on this NQPyClass.*
   */
  public void lt(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		    Label branch, NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      ltCmpInt(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      ltCmpDouble(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      ltCmpLong(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      ltCmpFloat(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      ltCmpString(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.BOOLEAN_CLASS:
      ltCmpBoolean(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.JAVA_CLASS:
      System.out.println("here " + this);
      ltCmpJava(lhsn, rhsn, branch, compiler);
      break;
    default:
      String message = "Cannot compare " + this + " and " + rhs;
      throw ErrorFactory.createCompilerException(message);
    }
  }

  public void ltCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    throwCmpError("int");
  }

  public void ltCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("double");
  }

  public void ltCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError("long");
  }

  public void ltCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    throwCmpError("float");
  }

  public void ltCmpBoolean(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			   NQPyBlockCompiler compiler)
  {
    throwCmpError("boolean");
  }

  public void ltCmpString(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("java.lang.String");
  }

  public void ltCmpJava(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError(rhsn.getNQPyClass(compiler).getJClassName());
  } 

  /**
   * Branches to the specified branch if the lhs node >= rhs node.<p>
   *
   * This checks the type of rhs and then calls the the appropriate gtEqCmp
   * on this NQPyClass.*
   */
  public void gtEq(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		    Label branch, NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      gtEqCmpInt(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      gtEqCmpDouble(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      gtEqCmpLong(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      gtEqCmpFloat(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      gtEqCmpString(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.BOOLEAN_CLASS:
      gtEqCmpBoolean(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.JAVA_CLASS:
      System.out.println("here " + this);
      gtEqCmpJava(lhsn, rhsn, branch, compiler);
      break;
    default:
      String message = "Cannot compare " + this + " and " + rhs;
      throw ErrorFactory.createCompilerException(message);
    }
  }

  public void gtEqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    throwCmpError("int");
  }

  public void gtEqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("double");
  }

  public void gtEqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError("long");
  }

  public void gtEqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    throwCmpError("float");
  }

  public void gtEqCmpBoolean(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			   NQPyBlockCompiler compiler)
  {
    throwCmpError("boolean");
  }

  public void gtEqCmpString(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("java.lang.String");
  }

  public void gtEqCmpJava(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError(rhsn.getNQPyClass(compiler).getJClassName());
  }


  /**
   * Branches to the specified branch if the lhs node <= rhs node.<p>
   *
   * This checks the type of rhs and then calls the the appropriate ltEqCmp
   * on this NQPyClass.*
   */
  public void ltEq(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		    Label branch, NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      ltEqCmpInt(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      ltEqCmpDouble(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      ltEqCmpLong(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      ltEqCmpFloat(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      ltEqCmpString(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.BOOLEAN_CLASS:
      ltEqCmpBoolean(lhsn, rhsn, branch, compiler);
      break;
    case NQPyClass.JAVA_CLASS:
      ltEqCmpJava(lhsn, rhsn, branch, compiler);
      break;
    default:
      String message = "Cannot compare " + this + " and " + rhs;
      throw ErrorFactory.createCompilerException(message);
    }
  }

  public void ltEqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    throwCmpError("int");
  }

  public void ltEqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("double");
  }

  public void ltEqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError("long");
  }

  public void ltEqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    throwCmpError("float");
  }

  public void ltEqCmpBoolean(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			   NQPyBlockCompiler compiler)
  {
    throwCmpError("boolean");
  }

  public void ltEqCmpString(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    throwCmpError("java.lang.String");
  }

  public void ltEqCmpJava(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    throwCmpError(rhsn.getNQPyClass(compiler).getJClassName());
  }

  /**
   * Performs addition between two nodes.
   *
   * This checks the type of rhs and then calls the the appropriate addX
   * on this NQPyClass.
   */
  public void add(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		  NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      addInt(lhsn, rhsn, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      addDouble(lhsn, rhsn, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      addLong(lhsn, rhsn, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      addFloat(lhsn, rhsn, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      addString(lhsn, rhsn, compiler);
      break;
    case NQPyClass.BOOLEAN_CLASS:
      // can add booleans to strings and strings to booleans
      addBoolean(lhsn, rhsn, compiler);
      break;
    default:
      throwArithError(rhs.getJClassName());
    }
  }

  public void addBoolean(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    throwArithError("boolean");
  }
  
  public void addInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    throwArithError("int");
  }

  public void addDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    throwArithError("double");
  }

  public void addLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    throwArithError("long");
  }

  public void addFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    throwArithError("float");
  }

  public void addString(SimpleNode lhsn, SimpleNode rhsn,
			NQPyBlockCompiler compiler)
  {
    throwArithError("java.lang.String");
  }

  /**
   * Performs subtraction between two nodes.
   *
   * This checks the type of rhs and then calls the the appropriate subX
   * on this NQPyClass.
   */
  public void sub(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		  NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      subInt(lhsn, rhsn, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      subDouble(lhsn, rhsn, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      subLong(lhsn, rhsn, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      subFloat(lhsn, rhsn, compiler);
      break;
    default:
      throwArithError(rhs.getJClassName());
    }
  }
  
  public void subInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    throwArithError("int");
  }

  public void subDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    throwArithError("double");
  }

  public void subLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    throwArithError("long");
  }

  public void subFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    throwArithError("float");
  }

  /**
   * Performs multiplication between two nodes.
   *
   * This checks the type of rhs and then calls the the appropriate multX
   * on this NQPyClass.
   */
  public void mult(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		  NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      multInt(lhsn, rhsn, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      multDouble(lhsn, rhsn, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      multLong(lhsn, rhsn, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      multFloat(lhsn, rhsn, compiler);
      break;
    case NQPyClass.STRING_CLASS:
      multString(lhsn, rhsn, compiler);
      break;
    default:
      throwArithError(rhs.getJClassName());
    }
  }
  
  public void multInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    throwArithError("int");
  }

  public void multDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    throwArithError("double");
  }

  public void multLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    throwArithError("long");
  }

  public void multFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    throwArithError("float");
  }

  public void multString(SimpleNode lhsn, SimpleNode rhsn,
			NQPyBlockCompiler compiler)
  {
    throwArithError("java.lang.String");
  }

  /**
   * Performs division between two nodes.
   *
   * This checks the type of rhs and then calls the the appropriate divX
   * on this NQPyClass.
   */
  public void div(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		  NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      divInt(lhsn, rhsn, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      divDouble(lhsn, rhsn, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      divLong(lhsn, rhsn, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      divFloat(lhsn, rhsn, compiler);
      break;
    default:
      throwArithError(rhs.getJClassName());
    }
  }
  
  public void divInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    throwArithError("int");
  }

  public void divDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    throwArithError("double");
  }

  public void divLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    throwArithError("long");
  }

  public void divFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    throwArithError("float");
  }

  public void divString(SimpleNode lhsn, SimpleNode rhsn,
			NQPyBlockCompiler compiler)
  {
    throwArithError("java.lang.String");
  }
  
  /**
   * Performs division between two nodes.
   *
   * This checks the type of rhs and then calls the the appropriate divX
   * on this NQPyClass.
   */
  public void mod(NQPyClass rhs, SimpleNode lhsn, SimpleNode rhsn,
		  NQPyBlockCompiler compiler)
  {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      modInt(lhsn, rhsn, compiler);
      break;
    case NQPyClass.DOUBLE_CLASS:
      modDouble(lhsn, rhsn, compiler);
      break;
    case NQPyClass.LONG_CLASS:
      modLong(lhsn, rhsn, compiler);
      break;
    case NQPyClass.FLOAT_CLASS:
      modFloat(lhsn, rhsn, compiler);
      break;
    default:
      throwArithError(rhs.getJClassName());
    }
  }
  
  public void modInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    throwArithError("int");
  }

  public void modDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    throwArithError("double");
  }

  public void modLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    throwArithError("long");
  }

  public void modFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    throwArithError("float");
  }

  public void modString(SimpleNode lhsn, SimpleNode rhsn,
			NQPyBlockCompiler compiler)
  {
    throwArithError("java.lang.String");
  }

  public void neg(SimpleNode node, NQPyBlockCompiler compiler) {
    throwArithError();
  }
}
