package uchicago.src.nqpy.compiler;

import uchicago.src.nqpy.parser.SimpleNode;
import com.go.trove.classfile.Label;

public class IntEmitter extends AbstractEmitter {
  
  public IntEmitter(NQPyClass nqclass) {
    super(nqclass);
  }

  /**
   * Gets the result type (as an NQPyClass) of single arg arithmetic operations
   * on this Emitter.
   */
  public NQPyClass getArithResType() {
    return nqPyClass;
  }

  /**
   * Gets the result type (as an NQPyClass) of arithmetic operations
   * on this Emitter.
   */
  public NQPyClass getArithResType(NQPyClass rhs) {
    switch (rhs.id()) {
    case NQPyClass.INT_CLASS:
      return NQPyClassFactory.makeIntegerClass();
    case NQPyClass.DOUBLE_CLASS:
      return NQPyClassFactory.makeDoubleClass();
    case NQPyClass.LONG_CLASS:
      return NQPyClassFactory.makeLongClass();
    case NQPyClass.FLOAT_CLASS:
      return NQPyClassFactory.makeFloatClass();
    case NQPyClass.STRING_CLASS:
      return NQPyClassFactory.makeStringClass();
    default:
      String message = "Arithmetic operations not supported between " +
	nqPyClass.getJClassName() + " and " + rhs.getJClassName();
      throw ErrorFactory.createCompilerException(message);
    }
  }

  /**
   * Emits the appropriatxe byte code to go to the specified branch
   * if the specified node compiles to true. This tests whether
   * the result of node.compile() is equal to zero, and branches if not so.
   */
  public void branchOnTrue(SimpleNode node, NQPyBlockCompiler compiler,
		   Label branch)
  {
    node.compile(compiler);
    compiler.emitNotEqZero(branch);
  }

  /**
   * Emits the appropriate byte code to go to the specified branch
   * if the single node in nodes compiles to false. This tests whether
   * the result of node.compile() is equal to zero, and branches if so.
   */
  public void branchOnFalse(SimpleNode node, NQPyBlockCompiler compiler,
		   Label branch)
  {
    node.compile(compiler);
    compiler.emitEqZero(branch);
  }


  private void intOp(SimpleNode lhsn, SimpleNode rhsn,
		     NQPyBlockCompiler compiler)
  {
    lhsn.compile(compiler);
    rhsn.compile(compiler);
  }

  private void doubleOp(SimpleNode lhsn, SimpleNode rhsn,
			NQPyBlockCompiler compiler)
  {
    lhsn.compile(compiler);
    compiler.emitPrimitiveConversion(int.class, double.class);
    rhsn.compile(compiler);
  }

  private void floatOp(SimpleNode lhsn, SimpleNode rhsn,
		       NQPyBlockCompiler compiler)
  {
    lhsn.compile(compiler);
    compiler.emitPrimitiveConversion(int.class, float.class);
    rhsn.compile(compiler);
  }

  private void longOp(SimpleNode lhsn, SimpleNode rhsn,
		      NQPyBlockCompiler compiler)
  {
    lhsn.compile(compiler);
    compiler.emitPrimitiveConversion(int.class, long.class);
    rhsn.compile(compiler);
  }
  
  public void eqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.emitIntComp(branch, "==");
  }
  
  public void eqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.emitDCMPL();
    compiler.emitEqZero(branch);

  }
  
  public void eqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.emitLCMP();
    compiler.emitEqZero(branch);
  }
  
  public void eqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.emitFCMPL();
    compiler.emitEqZero(branch);
  }

  public void notEqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.emitIntComp(branch, "!=");
  }
  
  public void notEqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.emitDCMPL();
    compiler.emitNotEqZero(branch);

  }
  
  public void notEqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.emitLCMP();
    compiler.emitNotEqZero(branch);
  }
  
  public void notEqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.emitFCMPL();
    compiler.emitNotEqZero(branch);
  }


  public void gtCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.emitIntComp(branch, ">");
  }
  
  public void gtCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.emitDCMPL();
    compiler.emitGtZero(branch);
  }
  
  public void gtCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.emitLCMP();
    compiler.emitGtZero(branch);
  }
  
  public void gtCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.emitFCMPL();
    compiler.emitGtZero(branch);
  }

    public void gtEqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.emitIntComp(branch, ">=");
  }
  
  public void gtEqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.emitDCMPL();
    compiler.emitGtEqZero(branch);
  }
  
  public void gtEqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.emitLCMP();
    compiler.emitGtEqZero(branch);
  }
  
  public void gtEqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.emitFCMPL();
    compiler.emitGtEqZero(branch);
  }

  public void ltCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.emitIntComp(branch, "<");
  }
  
  public void ltCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.emitDCMPL();
    compiler.emitLtZero(branch);
  }
  
  public void ltCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.emitLCMP();
    compiler.emitLtZero(branch);
  }
  
  public void ltCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.emitFCMPL();
    compiler.emitLtZero(branch);
  }

  public void ltEqCmpInt(SimpleNode lhsn, SimpleNode rhsn, Label branch,
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.emitIntComp(branch, "<=");
  }
  
  public void ltEqCmpDouble(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.emitDCMPL();
    compiler.emitLtEqZero(branch);
  }
  
  public void ltEqCmpLong(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.emitLCMP();
    compiler.emitLtEqZero(branch);
  }
  
  public void ltEqCmpFloat(SimpleNode lhsn, SimpleNode rhsn, Label branch,
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.emitFCMPL();
    compiler.emitLtEqZero(branch);
  }

  public void addInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.intAdd();
  }

  public void addDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.doubleAdd();
  }

  public void addLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.longAdd();
  }

  public void addFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.floatAdd();
  }

  public void addString(SimpleNode lhsn, SimpleNode rhsn,
			NQPyBlockCompiler compiler)
  {
    lhsn.compile(compiler);
    compiler.emitInvokeStatic("java.lang.String", "valueOf",
			      String.class, new Class[] {int.class});
    rhsn.compile(compiler);
    compiler.emitInvokeVirtual("java.lang.String", "concat", String.class,
			       new Class[] {String.class});
    
  }

  public void subInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.intSub();
  }

  public void subDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.doubleSub();
  }

  public void subLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.longSub();
  }

  public void subFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.floatSub();
  }

  public void multInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.intMult();
  }

  public void multDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.doubleMult();
  }

  public void multLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.longMult();
  }

  public void multFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.floatMult();
  }

  /**
   * This creates x copies of the string where x is the value of the
   * lhsn integer.
   */
  public void multString(SimpleNode lhsn, SimpleNode rhsn,
			NQPyBlockCompiler compiler)
  {
    compiler.emitInt(0);
    String iName = compiler.createTmp(int.class);
    compiler.emitStoreTmp(iName);
    lhsn.compile(compiler);
    String sizeName = compiler.createTmp(int.class);
    compiler.emitStoreTmp(sizeName);
    rhsn.compile(compiler);
    String stringName = compiler.createTmp(String.class);
    compiler.emitStoreTmp(stringName);
    
    compiler.emitNew(StringBuffer.class);
    compiler.emitDup();
    compiler.emitConstructor("java.lang.StringBuffer", new Class[] {});
    String bufName = compiler.createTmp(StringBuffer.class);
    compiler.emitStoreTmp(bufName);
    
    Label check = compiler.createLabel();
    Label end = compiler.createLabel();
    Label start = compiler.createLabel();
    compiler.emitGoto(check);
    start.setLocation();
    compiler.emitLoadTmp(bufName);
    compiler.emitLoadTmp(stringName);
    compiler.emitInvokeVirtual("java.lang.StringBuffer", "append",
			       StringBuffer.class, new Class[] {String.class});
    compiler.emitPop();
    compiler.emitIncTmp(iName, 1);
    check.setLocation();
    compiler.emitLoadTmp(iName);
    compiler.emitLoadTmp(sizeName);
    compiler.emitIntComp(start, "<");
    end.setLocation();
    compiler.emitLoadTmp(bufName);
    compiler.emitInvokeVirtual("java.lang.StringBuffer", "toString",
    			       String.class, new Class[] {});
    
  }

  public void divInt(SimpleNode lhsn, SimpleNode rhsn, 
		       NQPyBlockCompiler compiler)
  {
    intOp(lhsn, rhsn, compiler);
    compiler.intDiv();
  }

  public void divDouble(SimpleNode lhsn, SimpleNode rhsn, 
			  NQPyBlockCompiler compiler)
  {
    doubleOp(lhsn, rhsn, compiler);
    compiler.doubleDiv();
  }

  public void divLong(SimpleNode lhsn, SimpleNode rhsn, 
			NQPyBlockCompiler compiler)
  {
    longOp(lhsn, rhsn, compiler);
    compiler.longDiv();
  }

  public void divFloat(SimpleNode lhsn, SimpleNode rhsn, 
			 NQPyBlockCompiler compiler)
  {
    floatOp(lhsn, rhsn, compiler);
    compiler.floatDiv();
  }

  public void modInt(SimpleNode lhsn, SimpleNode rhsn, NQPyBlockCompiler compiler) {
    intOp(lhsn, rhsn, compiler);
    compiler.intMod();
  }
  
  public void modDouble(SimpleNode lhsn, SimpleNode rhsn, NQPyBlockCompiler compiler) {
    doubleOp(lhsn, rhsn, compiler);
    compiler.doubleMod();
  }
  
  public void modFloat(SimpleNode lhsn, SimpleNode rhsn, NQPyBlockCompiler compiler) {
    floatOp(lhsn, rhsn, compiler);
    compiler.floatMod();
  }
  
  public void modLong(SimpleNode lhsn, SimpleNode rhsn, NQPyBlockCompiler compiler) {
    longOp(lhsn, rhsn, compiler);
    compiler.longMod();
  }

  public void neg(SimpleNode node, NQPyBlockCompiler compiler) {
    node.compile(compiler);
    compiler.intNeg();
  }
    
}
