package uchicago.src.nqpy.compiler;

import ViolinStrings.Strings;
import com.go.trove.classfile.*;
import uchicago.src.nqpy.parser.NQPyParser;
import uchicago.src.nqpy.parser.ParseException;
import uchicago.src.nqpy.parser.SimpleNode;
import uchicago.src.nqpy.util.NQPyClassLoader;
import uchicago.src.sim.engine.BasicAction;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.StringReader;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Stack;

public class NQPyBlockCompiler {

  public static final NQPyClass NULL_NQPyClass = new NullNQPyClass();
  private int tmpCount = 0;

  private ClassFile cf;
  private CodeBuilder builder;

  // key: var name, value: VarClass object
  private HashMap localVars = new HashMap();
  private HashMap tmps = new HashMap();
  private HashSet classLookup = new HashSet();
  private boolean classMethMade = false;

  private Stack loopData = new Stack();
  private NQPyClass returnType = NQPyClassFactory.makeVoidClass();
  // this is to check when need a return but none found
  private boolean returnNotFound = true;

  class VarClass {
    LocalVariable variable;
    NQPyClass nqclass;

    public VarClass(LocalVariable v, NQPyClass nqclass) {
      this.variable = v;
      this.nqclass = nqclass;
    }
  }


  class LoopData {
    Label start, end, check;

    public LoopData(Label start, Label end, Label check) {
      this.start = start;
      this.end = end;
      this.check = check;
    }
  }


  /**
   * Creates a NQPyBlockCompiler of the specified class. This class
   * will extend uchicago.src.sim.engine.BasicAction and have a single
   * method public void execute(). This method will contain the compiled
   * block.
   */
  public NQPyBlockCompiler(String className) throws NoSuchMethodException {
    cf = new ClassFile(className, BasicAction.class);
    //cf.markSynthetic();
    makeConstructor();
    initExecuteBuilder();
  }

  public NQPyBlockCompiler(Method m, ClassFile cf) {
    this.cf = cf;
    MethodInfo mi = cf.addMethod(m);
    builder = new CodeBuilder(mi);
  }

  /**
   * Creates a BlockCompiler that will compile a public method with the
   * specified name into the specified class.
   */
  public NQPyBlockCompiler(String name, ClassFile cf) {
    this.cf = cf;
    createPublicMethod(name, void.class, null);
  }

  public NQPyBlockCompiler(String name, String retType, ClassFile cf) {
    this.cf = cf;
    createPublicMethod(name, retType, null);
  }

  public NQPyBlockCompiler(String name, Class retVal, ClassFile cf) {
    this.cf = cf;
    createPublicMethod(name, retVal, null);
  }

  public NQPyBlockCompiler(ClassFile cf) {
    this.cf = cf;
  }

  public void createMainMethod() {
    AccessFlags pubStatic = new AccessFlags();
    pubStatic.setPublic(true);
    pubStatic.setStatic(true);
    TypeDescriptor[] params = new TypeDescriptor[1];
    params[0] = new TypeDescriptor(String[].class);
    MethodInfo mi = cf.addMethod(pubStatic, "main", null, params);
    builder = new CodeBuilder(mi);
  }

  private void createPublicMethod(String name, String retVal, Class[] args) {
    TypeDescriptor td = new TypeDescriptor(retVal);
    createPublicMethod(name, td, args);
  }

  private void createPublicMethod(String name, Class retVal, Class[] args) {
    TypeDescriptor ret = null;
    if (retVal != null) {
      ret = new TypeDescriptor(retVal);
    }

    createPublicMethod(name, ret, args);
  }

  private void createPublicMethod(String name, TypeDescriptor ret, Class[] args) {
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);
    MethodInfo mi = cf.addMethod(pub, name, ret,
            CompilerUtil.makeTypeDescriptors(args));
    builder = new CodeBuilder(mi);
    returnType = NQPyClassFactory.makeNQPyClass(ret.getClassName());
  }

  public String getClassName() {
    return cf.getClassName();
  }


  /*
   * Makes a no-arg constructor.
   */
  private void makeConstructor() {
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);

    MethodInfo ctor = cf.addConstructor(pub, null);
    builder = new CodeBuilder(ctor);
    builder.loadThis();
    builder.invokeSuperConstructor(null);
    builder.returnVoid();
  }

  /*
   * Initializes the creation of the public void execute() method of
   * a BasicAction class. Subsequent calls to emit* will add bytecode
   * to this method.
   */
  private void initExecuteBuilder() throws NoSuchMethodException {
    Method execMethod = BasicAction.class.getMethod("execute", new Class[]{});
    MethodInfo mi = cf.addMethod(execMethod);
    builder = new CodeBuilder(mi);
  }

  public void setBlockName(String val) {
    ErrorFactory.setBlockName(val);
  }

  /**
   * Adds a java class to the list of imports.
   */
  public void addImport(String name) {
    NQPyClassFactory.addImport(name);
  }

  /**
   * Sets the NQPyDynamicClass that the "self" variable
   * refers to.
   */
  public void setSelf(NQPyDynamicClass selfType) {
    localVars.put("self", new VarClass(null, selfType));
  }

  /**
   * Sets the return type.
   */
  public void setReturnType(Class c) {
    returnType = NQPyClassFactory.makeNQPyClass(c);
  }

  /**
   * gets the return type.
   */
  public NQPyClass getReturnType() {
    return returnType;
  }

  /**
   * Sets whether a return stmt was encountered or not.
   */
  public void setReturnNotFound(boolean val) {
    returnNotFound = val;
  }


  /**
   * Gets the NQPyClass for the specified local variable.
   */
  public NQPyClass getNQPyClassForVar(String varName) {
    VarClass var = (VarClass) localVars.get(varName);
    if (var == null || var.nqclass == null) return NULL_NQPyClass;
    return var.nqclass;
  }

  /**
   * Returns whether the specified name is a local variable.
   */
  public boolean isLocalVar(String varName) {
    return localVars.containsKey(varName);
  }

  /**
   * Pushes the loop data onto the loop data stack.
   */
  public void pushLoopData(Label start, Label end, Label check) {
    loopData.push(new LoopData(start, end, check));
  }

  /**
   * Pops the loop data off the stack.
   */
  public void popLoopData() {
    loopData.pop();
  }

  /**
   * Gets the size of the loop data stack.
   */
  public int getLoopDataSize() {
    return loopData.size();
  }

  /**
   * Adds field and code for a class lookup to the current class.
   */
  private void addClassLookup(NQPyClass nqclass) {
    String className = Strings.change(nqclass.getJClassName(), ".", "$");
    // only do this if we haven't already created a field to cache this Class data.
    if (!classContainsField("class$" + className)) {
      AccessFlags flag = new AccessFlags();
      flag.setPrivate(true);
      flag.setStatic(true);

      FieldInfo info = cf.addField(flag, "class$" + className,
              new TypeDescriptor(Class.class));
      info.markSynthetic();
      makeClassLookupMethod();
      classLookup.add("class$" + className);
    }
  }

  private boolean classContainsField(String fieldName) {
    FieldInfo[] info = cf.getFields();
    for (int i = 0; i < info.length; i++) {
      if (fieldName.equals(info[i].getName())) {
        return true;
      }
    }

    return false;
  }

  private void makeClassLookupMethod() {
    // if the class already has the class lookup method in it
    // return
    MethodInfo[] methodInfo = cf.getMethods();
    for (int i = 0; i < methodInfo.length; i++) {
      MethodInfo info = methodInfo[i];
      if (info.getName().equals("_class$_")) return;
    }

    AccessFlags mflags = new AccessFlags();
    //mflags.setPrivate(true);
    mflags.setStatic(true);
    MethodInfo mi = cf.addMethod(mflags, "_class$_",
            new TypeDescriptor(Class.class),
            new TypeDescriptor[]{
              new TypeDescriptor(String.class)});

    mi.markSynthetic();
    CodeBuilder forClass = new CodeBuilder(mi);
    Label start = forClass.createLabel();
    Label end = forClass.createLabel();
    start.setLocation();
    forClass.loadLocal(forClass.getParameters()[0]);

    TypeDescriptor[] tds = new TypeDescriptor[]{
      new TypeDescriptor(String.class)};
    forClass.invokeStatic("java.lang.Class", "forName",
            new TypeDescriptor(Class.class),
            tds);
    end.setLocation();
    forClass.returnValue(Class.class);
    forClass.exceptionHandler(start, end, "java.lang.ClassNotFoundException");
    TypeDescriptor td = new TypeDescriptor("java.lang.ClassNotFoundException");
    LocalVariable ex = forClass.createLocalVariable("ex", td);
    forClass.storeLocal(ex);
    forClass.newObject(new
            TypeDescriptor(java.lang.NoClassDefFoundError.class));
    forClass.dup();
    forClass.loadLocal(ex);
    forClass.invokeVirtual("java.lang.Throwable",
            "getMessage",
            new TypeDescriptor(String.class),
            null);
    forClass.invokeConstructor("java.lang.NoClassDefFoundError",
            CompilerUtil.makeTypeDescriptors(String.class));
    forClass.throwObject();

    classMethMade = true;
  }

  /**
   * Emits code for doing a Integer.class type call.
   */
  public void emitClassLookup(NQPyClass nqclass) {
    addClassLookup(nqclass);

    TypeDescriptor ctd = new TypeDescriptor(Class.class);
    String name = "class$" + Strings.change(nqclass.getJClassName(),
            ".", "$");
    builder.loadStaticField(name, ctd);
    Label call = builder.createLabel();
    Label end = builder.createLabel();
    builder.ifNullBranch(call, true);
    // null pops the value from the static so need to reload it
    builder.loadStaticField(name, ctd);
    builder.branch(end);
    call.setLocation();
    //builder.loadThis();
    builder.loadConstant(nqclass.getJClassName());
    builder.invokeStatic("_class$_", ctd,
            CompilerUtil.makeTypeDescriptors(String.class));
    builder.dup();
    builder.storeStaticField(name, ctd);
    end.setLocation();

  }

  /** emit* methods. The emit* methods do that actual emission of bytecode */

  /**
   * Emits the cast byte code.
   */
  public void emitCast(NQPyClass nqclass) {
    builder.checkCast(nqclass.makeTypeDescriptor());
  }

  /**
   * Emits a break which breaks out of the top most loop on the stack.
   * Assumes that the stack is not empty (i.e. that the check has been done
   * by the caller).
   */
  public void emitBreak() {
    LoopData data = (LoopData) loopData.peek();
    builder.branch(data.end);
  }

  /**
   * Emits a continue which branches to the start of the topmost loop on
   * the stack. Assumes that the stack is not empty (i.e. that the check
   * has been done by the caller).
   */
  public void emitContinue() {
    LoopData data = (LoopData) loopData.peek();
    builder.branch(data.start);
  }

  /**
   * Emits byte code for return.
   */
  public void emitReturn(NQPyClass c) {
    if (c.id() == NQPyClass.VOID_CLASS)
      builder.returnVoid();
    else
      builder.returnValue(c.getJClass());
  }

  /**
   * Creates local variables for the parameters passed to the
   * current method under construction. These names can then be used
   * within the method body.
   */
  public void createMethodParams(ArrayList names) {
    LocalVariable[] mparams = builder.getParameters();
    if (names.size() != mparams.length) {
      String message = "Internal compiler error; invalid number of names " +
              "passed to createMethodParams";
      throw ErrorFactory.createCompilerException(message);
    }

    for (int i = 0; i < mparams.length; i++) {
      String name = (String) names.get(i);
      LocalVariable var = mparams[i];
      NQPyClass clazz = NQPyClassFactory.makeNQPyClass(var.getType().
              getClassName());
      localVars.put(name, new VarClass(var, clazz));
      var.setName(name);
    }
  }
  
  /**
   * Emits the "arraylength" instruction
   */ 
  public void emitArrayLength() {
    builder.arrayLength();
  }
  
  /**
   * Loads value from an array. The array reference, followed by the index must be on the
   * stack.
   */ 
  public void emitLoadFromArray(Class componentType) {
    builder.loadFromArray(componentType);
  }
  
  /**
   * Creates a local variable with the specified name and the specified
   * class. If the local var already exists, then this checks if its NQPyClass
   * allows assignment from varClass. If not, this will return
   * NQPyBlockCompiler.NullNQPyClass.
   *
   * @return the NQPyClass of the created or already existing variable,
   *         or NQPyBlockCompiler.NullNQPyClass if assignment cannot be performed.
   */
  public NQPyClass emitLocalVar(String varName, NQPyClass varClass) {
    VarClass var = (VarClass) localVars.get(varName);
    if (var == null) {
      LocalVariable lvar = builder.createLocalVariable(varName,
              varClass.
              makeTypeDescriptor());
      localVars.put(varName, new VarClass(lvar, varClass));
      return varClass;
    } else {
      if (var.nqclass.isAssignableFrom(varClass)) {
        if (var.nqclass.equals(NQPyClassFactory.makeNQPyClass(Object.class))) {
          // change the nqclass of the var
          var.nqclass = varClass;
        } 
        return var.nqclass;
      }
    }

    return NULL_NQPyClass;
  }

  /**
   * Creates a local variable with the specified name and the specified
   * class. If the local var already exists, then it is overwritten by
   * this one.
   */
  public void emitLocalVarNoCheck(String varName, NQPyClass varClass) {
    LocalVariable lvar = builder.
            createLocalVariable(varName, varClass.makeTypeDescriptor());
    localVars.put(varName, new VarClass(lvar, varClass));
  }

  /**
   * Stores the value on the stack in the specified local variable.
   * Assumes the appropriate type checking has already been done.
   */
  public void emitStoreLocal(String name) {
    LocalVariable v = ((VarClass) localVars.get(name)).variable;
    builder.storeLocal(v);
  }

  /**
   * Loads the value of the specified local variable onto the stack.
   */
  public void emitLoadLocalVar(String name) {
    if (name.equals("self"))
      builder.loadThis();
    else {
      LocalVariable v = ((VarClass) localVars.get(name)).variable;
      builder.loadLocal(v);
    }
  }

  /**
   * Creates a temporay local variable. This is used when variables
   * are needed for things like for loops etc. Returns a unique variable name.
   */
  public synchronized String createTmp(Class c) {
    String name = "__tmp__" + tmpCount++;
    LocalVariable v = builder.createLocalVariable(name, new TypeDescriptor(c));
    tmps.put(name, v);
    return name;
  }

  /**
   * Stores the value on the stack in the specified tmp variable.
   */
  public void emitStoreTmp(String name) {
    LocalVariable v = (LocalVariable) tmps.get(name);
    if (v == null) {
      String message = "Internal compiler error, undefined tmp variable " +
              name;
      throw ErrorFactory.createCompilerException(message);
    }
    builder.storeLocal(v);
  }

  /**
   * Loads the value of the specified tmp variable onto the stack.
   */
  public void emitLoadTmp(String name) {
    LocalVariable v = (LocalVariable) tmps.get(name);
    if (v == null) {
      String message = "Internal compiler error, undefined tmp variable " +
              name;
      throw ErrorFactory.createCompilerException(message);
    }
    builder.loadLocal(v);
  }

  /**
   * Increments the value of the temporary variable by the specified amount.
   * The variable must an int.
   */
  public void emitIncTmp(String name, int amount) {
    LocalVariable v = (LocalVariable) tmps.get(name);
    if (v == null) {
      String message = "Internal compiler error, undefined tmp variable " +
              name;
      throw ErrorFactory.createCompilerException(message);
    }
    builder.integerIncrement(v, amount);
  }

  /**
   * Emits a static method call. Assumes the args are
   * on the stack.
   */
  public void emitInvokeStatic(String fullClassName, String methodName,
                               NQPyClass retType, Class[] args) {
    TypeDescriptor ret = retType.makeTypeDescriptor();
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(args);
    builder.invokeStatic(fullClassName, methodName, ret, params);
  }

  /**
   * Emits a static method call. Assumes the args are
   * on the stack.
   */
  public void emitInvokeStatic(String fullClassName, String methodName,
                               Class retType, Class[] args) {
    TypeDescriptor ret = new TypeDescriptor(retType);
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(args);
    builder.invokeStatic(fullClassName, methodName, ret, params);
  }

  /**
   * Emits a static method call using the data in the specified MethodCall.
   */
  public void emitInvokeStatic(MethodCall call) {
    String className = call.getDeclaringClass().getJClassName();
    String methName = call.getMethodName();
    TypeDescriptor ret = call.getRetType().makeTypeDescriptor();
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(call.getArgs());
    builder.invokeStatic(className, methName, ret, params);
  }

  /**
   * Invokes the specified method of the class with the specified name,
   * and the specified argtypes. The method must return void. Note that
   * this does not check if the named class has such a method.
   */
  public void emitInvokeVQuickNoRet(String className, String methodName,
                                    TypeDescriptor[] tds) {
    builder.invokeVirtual(className, methodName, null, tds);
  }

  /**
   * Emits a virtual method call using the specified arguments.
   */
  public void emitInvokeVirtual(String className, String methName,
                                Class retType, Class[] args) {
    TypeDescriptor ret = new TypeDescriptor(retType);

    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(args);
    builder.invokeVirtual(className, methName, ret, params);
  }

  /**
   * Invokes the specified method by either caling emitInvokeVirtual or
   * emitInvokeInterface where appropriate.
   */
  public void emitInvokeMethod(MethodCall call) {
    Class c = call.getDeclaringClass().getJClass();
    // to work with dynamic classes that have no jclass
    if (c == null)
      emitInvokeVirtual(call);
    else if (call.isStatic())
      emitInvokeStatic(call);
    else if (c.isInterface())
      emitInvokeInterface(call);
    else
      emitInvokeVirtual(call);
  }
  
  public void emitInvokeSuper(MethodCall call) {
    String className = call.getDeclaringClass().getJClassName();
    String methName = call.getMethodName();
    TypeDescriptor ret = call.getRetType().makeTypeDescriptor();
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(call.getArgs());
    builder.invokeSuper(className, methName, ret, params);
  }

  /**
   * Emits a virtual method call using the data in the specified MethodCall.
   */
  public void emitInvokeVirtual(MethodCall call) {
    String className = call.getDeclaringClass().getJClassName();
    String methName = call.getMethodName();
    TypeDescriptor ret = call.getRetType().makeTypeDescriptor();
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(call.getArgs());
    builder.invokeVirtual(className, methName, ret, params);
  }

  public void emitInvokeInterface(String className, String methodName, Class returnType, Class[] args) {
    TypeDescriptor ret = new TypeDescriptor(returnType);
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(args);
    builder.invokeInterface(className, methodName, ret, params);
  }

  /**
   * Emits an interface method call using the data in the specified MethodCall.
   */
  public void emitInvokeInterface(MethodCall call) {
    String className = call.getDeclaringClass().getJClassName();
    String methName = call.getMethodName();
    TypeDescriptor ret = call.getRetType().makeTypeDescriptor();
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(call.getArgs());
    builder.invokeInterface(className, methName, ret, params);
  }

  /**
   * Emits a constructor call using data in the specified MethodCall.
   */
  public void emitConstructor(MethodCall call) {
    String className = call.getDeclaringClass().getJClassName();
    TypeDescriptor[] params = CompilerUtil.makeTypeDescriptors(call.getArgs());
    builder.invokeConstructor(className, params);

  }

  /**
   * Emits a constructor call using data in the specified MethodCall.
   */
  public void emitConstructor(String className, Class[] params) {
    TypeDescriptor[] tparams = CompilerUtil.makeTypeDescriptors(params);
    builder.invokeConstructor(className, tparams);
  }

  /**
   * Emits code to load a field using the FieldCall data.
   */
  public void emitLoadField(FieldCall call) {
    if (call.isStatic()) {
      //System.out.println(call.getDeclaringClassName());
      //System.out.println(call.getFieldName());
      //System.out.println(call.getStaticFieldType());

      builder.loadStaticField(call.getDeclaringClassName(),
              call.getFieldName(),
              call.getFieldType().makeTypeDescriptor());
    } else {
      builder.loadField(call.getDeclaringClassName(),
              call.getFieldName(),
              call.getFieldType().makeTypeDescriptor());
    }
  }

  /**
   * Emits code to store a field using the FieldCall data.
   */
  public void emitStoreField(FieldCall call) {
    if (call.isStatic()) {
      builder.storeStaticField(call.getDeclaringClassName(),
              call.getFieldName(),
              call.getFieldType().makeTypeDescriptor());
    } else {
      builder.storeField(call.getDeclaringClassName(),
              call.getFieldName(),
              call.getFieldType().makeTypeDescriptor());
    }
  }

  /**
   * Adds two ints and leaves result on stack.
   */
  public void intAdd() {
    builder.math(Opcode.IADD);
  }

  public void intSub() {
    builder.math(Opcode.ISUB);
  }

  public void intMult() {
    builder.math(Opcode.IMUL);
  }

  public void intDiv() {
    builder.math(Opcode.IDIV);
  }

  public void intNeg() {
    builder.math(Opcode.INEG);
  }

  /**
   * Adds two doubles and leaves result on stack.
   */
  public void doubleAdd() {
    builder.math(Opcode.DADD);
  }

  public void doubleSub() {
    builder.math(Opcode.DSUB);
  }

  public void doubleMult() {
    builder.math(Opcode.DMUL);
  }

  public void doubleDiv() {
    builder.math(Opcode.DDIV);
  }

  public void doubleNeg() {
    builder.math(Opcode.DNEG);
  }

  public void floatAdd() {
    builder.math(Opcode.FADD);
  }

  public void floatSub() {
    builder.math(Opcode.FSUB);
  }

  public void floatMult() {
    builder.math(Opcode.FMUL);
  }

  public void floatDiv() {
    builder.math(Opcode.FDIV);
  }

  public void floatNeg() {
    builder.math(Opcode.FNEG);
  }

  public void longAdd() {
    builder.math(Opcode.LADD);
  }

  public void longSub() {
    builder.math(Opcode.LSUB);
  }

  public void longMult() {
    builder.math(Opcode.LMUL);
  }

  public void longDiv() {
    builder.math(Opcode.LDIV);
  }

  public void longNeg() {
    builder.math(Opcode.LNEG);
  }

  public void intMod() {
    builder.math(Opcode.IREM);
  }

  public void doubleMod() {
    builder.math(Opcode.DREM);
  }

  public void floatMod() {
    builder.math(Opcode.FREM);
  }

  public void longMod() {
    builder.math(Opcode.LREM);
  }

  /**
   * Emits the new byte code.
   */
  public void emitNew(NQPyClass nq) {
    builder.newObject(nq.makeTypeDescriptor());
  }

  /**
   * Emits the new byte code.
   */
  public void emitNew(Class c) {
    builder.newObject(new TypeDescriptor(c));
  }

  /**
   * Emits the "dup" byte code
   */
  public void emitDup() {
    builder.dup();
  }

  /**
   * Emits the pop byte code.
   */
  public void emitPop() {
    builder.pop();
  }

  /**
   * Converts between primitive types.
   */
  public void emitPrimitiveConversion(Class from, Class to) {
    builder.convert(from, to);
  }

  /**
   * Creates a label.
   */
  public Label createLabel() {
    return builder.createLabel();
  }

  /**
   * Emits the dcmpl byte code, comparing two doubles on the stack.
   */
  public void emitDCMPL() {
    builder.math(Opcode.DCMPL);
  }

  /**
   * Emits the fcmpl byte code, comparing two doubles on the stack.
   */
  public void emitFCMPL() {
    builder.math(Opcode.FCMPL);
  }

  /**
   * Emits the lcmp byte code, comparing two longs on the stack.
   */
  public void emitLCMP() {
    builder.math(Opcode.LCMP);
  }

  /**
   * Emits to branch to specified label if obj ref on stack is not null.
   */
  public void emitIfNotNull(Label branch) {
    builder.ifNullBranch(branch, false);
  }

  /**
   * Emits to branch to specified label if obj ref on stack is null.
   */
  public void emitIfNull(Label branch) {
    builder.ifNullBranch(branch, true);
  }

  /**
   * Emits byte code to branch to specified label if int on stack equals
   * 0.
   */
  public void emitEqZero(Label branch) {
    builder.ifZeroComparisonBranch(branch, "==");
  }

  /**
   * Emits byte code to branch to specified label if int on stack does not
   * equal 0.
   */
  public void emitNotEqZero(Label branch) {
    builder.ifZeroComparisonBranch(branch, "!=");
  }

  /**
   * Emits byte code to branch to specified label if int on stack > 0
   */
  public void emitGtZero(Label branch) {
    builder.ifZeroComparisonBranch(branch, ">");
  }

  /**
   * Emits byte code to branch to specified label if int on stack >= 0
   */
  public void emitGtEqZero(Label branch) {
    builder.ifZeroComparisonBranch(branch, ">=");
  }

  /**
   * Emits byte code to branch to specified label if int on stack < 0
   */
  public void emitLtZero(Label branch) {
    builder.ifZeroComparisonBranch(branch, "<");
  }

  /**
   * Emits byte code to branch to specified label if int on stack <= 0
   */
  public void emitLtEqZero(Label branch) {
    builder.ifZeroComparisonBranch(branch, "<=");
  }


  /**
   * Emits byte code to branch to the specified label according to the
   * specified comparison of two ints on stack.
   */
  public void emitIntComp(Label branch, String choice) {
    builder.ifComparisonBranch(branch, choice);
  }

  /**
   * Emits byte code to jump to the specified label.
   */
  public void emitGoto(Label branch) {
    builder.branch(branch);
  }


  /**
   * Converts between an int on the stack and a boolean (an int of 1 or 0).
   */
  public void emitIntToBoolean() {
    Label branch = builder.createLabel();
    Label end = builder.createLabel();
    builder.ifZeroComparisonBranch(branch, "!=");
    builder.loadConstant(0);
    builder.branch(end);
    branch.setLocation();
    builder.loadConstant(1);
    end.setLocation();
  }

  /**
   * Converts between an double on the stack and a boolean
   * (an int of 1 or 0).
   */
  public void emitDoubleToBoolean() {
    builder.convert(double.class, int.class);
    emitIntToBoolean();
  }

  /**
   * Converts between an float on the stack and a boolean
   * (an int of 1 or 0).
   */
  public void emitFloatToBoolean() {
    builder.convert(float.class, int.class);
    emitIntToBoolean();
  }

  /**
   * Converts between an long on the stack and a boolean
   * (an int of 1 or 0).
   */
  public void emitLongToBoolean() {
    builder.convert(long.class, int.class);
    emitIntToBoolean();
  }


  /**
   * Pushes the specified String on the stack.
   */
  public void emitString(String str) {
    builder.loadConstant(str);
  }

  /**
   * Pushes the specified int on the stack.
   */
  public void emitInt(int val) {
    builder.loadConstant(val);
  }

  /**
   * Pushes the specified double on the stack.
   */
  public void emitDouble(double val) {
    builder.loadConstant(val);
  }

  /**
   * Pushes the specified long on the stack.
   */
  public void emitLong(long val) {
    builder.loadConstant(val);
  }

  /**
   * Pushes the specified float on the stack.
   */
  public void emitFloat(float val) {
    builder.loadConstant(val);
  }

  public void emitNull() {
    builder.loadConstant(null);
  }


  /**
   * Push System.out on the stack. This must be followed by an emit*
   * that places the object on the stack, and then an emitPrint*End() call
   * which will invoke println on the System.out.
   */
  public void emitPrintStart() {
    builder.loadStaticField("java.lang.System", "out",
            new TypeDescriptor(java.io.PrintStream.class));
  }


  /**
   * Emits a PrintStream.print. Expects the argument to be on the stack,
   * and described by the specified pyclass.
   */
  public void emitPrintEnd(NQPyClass pyclass) {
    TypeDescriptor[] tds;

    if (pyclass.id() == NQPyClass.JAVA_CLASS) {
      tds = CompilerUtil.makeTypeDescriptors(Object.class);
    } else {
      tds = CompilerUtil.makeTypeDescriptors(pyclass);
    }
    this.emitInvokeVQuickNoRet("java.io.PrintStream", "print", tds);

  }

  /**
   * Emits a PrintStream.println. Expects the argument to be on the stack,
   * and described by the specified pyclass.
   */
  public void emitPrintlnEnd(NQPyClass pyclass) {
    TypeDescriptor[] tds;

    if (pyclass.id() == NQPyClass.JAVA_CLASS) {
      tds = CompilerUtil.makeTypeDescriptors(Object.class);
    } else {
      tds = CompilerUtil.makeTypeDescriptors(pyclass);
    }
    this.emitInvokeVQuickNoRet("java.io.PrintStream", "println", tds);
  }

  public void compile(String code) {
    StringReader reader = new StringReader(code);
    NQPyParser parser = new NQPyParser(reader);
    ErrorFactory.setCurNode(null);
    ErrorFactory.setCode(code);

    try {
      SimpleNode root = parser.file_input();
      //root.dump("");
      root.compile(this);
      if (returnType != NQPyClassFactory.makeVoidClass() && returnNotFound) {
        ErrorFactory.setCurNode(root);
        String message = "Method must return " + returnType.getJClassName();
        throw ErrorFactory.createCompilerException(message);
      }

      if (returnType == NQPyClassFactory.makeVoidClass()) {
        builder.returnVoid();
      }

    } catch (ParseException ex) {
      throw ErrorFactory.createParserException(ex);
    }
  }

  public void execute() {
    NQPyClassLoader cl = new NQPyClassLoader(this.getClass().getClassLoader());
    cl.addClass("BCTest", cf);
    cl.writeClass("BCTest", "c:/temp");

    try {
      Class clazz = cl.loadClass(cf.getClassName());
      Constructor baCtor = clazz.getConstructor(new Class[]{});
      BasicAction action = (BasicAction) baCtor.newInstance(new Object[]{});
      action.execute();
    } catch (Exception ex) {
      ex.printStackTrace();
      System.exit(0);
    }
  }

  public static void main(String[] args) {

    /* dynamic class test
    NQPyDynamicClass dc = new NQPyDynamicClass("EvNode",
					       uchicago.src.sim.network.
					       DefaultNode.class);
    dc.addMethod("EvNode", "EvNode", new String[]{});
    dc.addMethod("foo", void.class, new Class[]{int.class});
    dc.addField("bar", int.class);
    dc.addMethod("setNode", "void", new String[]{"EvNode"});
    NQPyClassFactory.addDynType(dc.getJClassName(), dc);
    */
    //NQPyClassFactory.addImport("javax.swing.JFrame");
    //NQPyClassFactory.addImport("uchicago.src.nqpy.util.NQPyUtil");


    try {
      BufferedReader in = new BufferedReader(new FileReader(args[0]));
      StringBuffer b = new StringBuffer();
      String line;

      while ((line = in.readLine()) != null) {
        b.append(line);
        b.append("\n");
      }

      NQPyBlockCompiler compiler = new NQPyBlockCompiler("BCTest");
      //compiler.setReturnType(int.class);
      ErrorFactory.setBlockName(args[0]);
      compiler.compile(b.toString());
      compiler.execute();
    } catch (CompilerException ex) {
      ex.printStackTrace();
    } catch (ParserException ex) {
      System.out.println(ex.getMessage());
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }
}
