package uchicago.src.simbuilder;

import uchicago.src.simbuilder.beans.*;

import javax.swing.*;
import javax.swing.tree.*;
import java.awt.*;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * JTree for displaying the Evolver project tree. Events and
 * interaction with other classes are handled via methods in
 * EvolverController.
 * <p/>
 * $Revision: 1.2 $ $Date: 2004/12/10 22:09:50 $
 */

public class EvolverBeanTree extends JTree {

  private Hashtable beanNodeMap = new Hashtable();

  public EvolverBeanTree() {
    super();
    setRowHeight(0);
    setRootVisible(true);
    setShowsRootHandles(true);
    EvolverBeanNode projNode = new EvolverBeanNode(new EvolverProject(),
            "Environment");
    beanNodeMap.put(projNode.bean, projNode);
    setModel(new DefaultTreeModel(projNode));
    this.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
    setCellRenderer(new EvolverTreeRenderer());
  }

  /**
   * Adds a node (an EvolverBean) to the project tree.
   */
  public TreePath addNode(TreePath path, Object bean, String name) {
    EvolverBeanNode parent = (EvolverBeanNode) path.getLastPathComponent();
    EvolverBeanNode node = new EvolverBeanNode(bean, name);
    beanNodeMap.put(bean, node);

    parent.add(node);
    path = path.pathByAddingChild(node);
    ((DefaultTreeModel) getModel()).reload(parent);
    //System.out.println(parent.getChildCount());
    return path;
  }

  /**
   * Removes a node (an EvolverBean) from the project tree and
   * more specifically the specified path.
   */
  public void removeNode(TreePath path) {
    EvolverBeanNode node = (EvolverBeanNode) path.getLastPathComponent();
    if (node != getModel().getRoot()) {
      EvolverBeanNode parent = (EvolverBeanNode) node.getParent();
      EvolverBean parentBean = (EvolverBean) parent.getBean();
      EvolverBean child = (EvolverBean) node.getBean();
      child.removeFromParent((EvolverParent) parentBean);
      node.removeFromParent();
      ((DefaultTreeModel) getModel()).reload(parent);
      setSelectionPath(path.getParentPath());
      beanNodeMap.remove(node.getBean());
    }

  }

  /**
   * Selects the tree node associated with the specified Object.
   *
   * @param bean
   * @return
   */
  public TreePath selectNodeForObject(Object bean) {
    DefaultTreeModel model = (DefaultTreeModel) getModel();
    EvolverBeanNode root = (EvolverBeanNode) model.getRoot();
    for (Enumeration e = root.breadthFirstEnumeration(); e.hasMoreElements();) {
      EvolverBeanNode node = (EvolverBeanNode) e.nextElement();
      if (node.getBean().equals(bean)) {
        TreeNode[] path = model.getPathToRoot(node);
        TreePath p = new TreePath(path);
        setSelectionPath(p);
        return p;
      }
    }

    return null;
  }

  /**
   * Selects a node by name.
   */
  public TreePath selectNodeByBean(Object bean) {
    EvolverBeanNode node = (EvolverBeanNode) beanNodeMap.get(bean);
    TreeNode[] path = ((DefaultTreeModel) getModel()).getPathToRoot(node);
    TreePath p = new TreePath(path);
    setSelectionPath(p);
    return p;
  }

  /**
   * Renames a node.
   */
  public void renameNode(Object bean, String oldName, String newName) {
    EvolverBeanNode node = (EvolverBeanNode) beanNodeMap.remove(bean);
    node.name = newName;
    beanNodeMap.put(bean, node);
    ((DefaultTreeModel) getModel()).reload(node);
  }

  /**
   * Gets the currently selected bean (not the node wrapper).
   */
  public Object getSelectedBean(TreePath path) {
    EvolverBeanNode node = (EvolverBeanNode) path.getLastPathComponent();
    return node.getBean();
  }

  public EvolverProject getProjectBean() {
    EvolverBeanNode root = (EvolverBeanNode) this.getModel().getRoot();
    return (EvolverProject) root.getBean();
  }

  public void expandAll() {
    TreeNode root = (TreeNode) this.getModel().getRoot();
    // Traverse tree from root
    expandAll(new TreePath(root), true);
  }

  private void expandAll(TreePath parent, boolean expand) {
    // Traverse children
    TreeNode node = (TreeNode) parent.getLastPathComponent();
    if (node.getChildCount() >= 0) {
      for (Enumeration e = node.children(); e.hasMoreElements();) {
        TreeNode n = (TreeNode) e.nextElement();
        TreePath path = parent.pathByAddingChild(n);
        expandAll(path, expand);
      }
    }
    
    // Expansion or collapse must be done bottom-up
    if (expand) {
      this.expandPath(parent);
    } else {
      this.collapsePath(parent);
    }
  }

  // adds the specified list of children to the specified EvolverBeanNode
  private void addChildren(EvolverBeanNode parent, ArrayList children) {
    for (int i = 0; i < children.size(); i++) {
      try {
        EvolverBean bean = (EvolverBean) children.get(i);
        EvolverBeanNode node = new EvolverBeanNode(bean,
                Introspector.getBeanInfo(bean.getClass()).getBeanDescriptor().getName());
        parent.add(node);
        beanNodeMap.put(bean, node);
        if (bean instanceof EvolverParent) {
          addChildren(node, ((EvolverParent) bean).children());
        }
      } catch (IntrospectionException ex) {
        ex.printStackTrace();
      }
    }
  }

  public void setProjectBean(EvolverProject project) {
    DefaultTreeModel model = (DefaultTreeModel) getModel();
    EvolverBeanNode newRoot = new EvolverBeanNode(project,
            project.getProjectName());
    beanNodeMap.clear();
    beanNodeMap.put(project, newRoot);
    addChildren(newRoot, project.children());
    model.setRoot(newRoot);
    model.reload();
    selectNodeByBean(project);
  }

  /**
   * A tree node wrapper class around EvolverBeans.
   */
  class EvolverBeanNode extends DefaultMutableTreeNode {

    Object bean;
    String name;

    public EvolverBeanNode(Object bean, String name) {
      this.bean = bean;
      this.name = name;

      // yuck!
      String n = null;
      if (bean instanceof EvolverModel) {
        n = ((EvolverModel) bean).displayName();
      } else if (bean instanceof EvolverCompileable) {
        n = ((EvolverCompileable) bean).className();
      }

      if (n != null && n.length() > 0) {
        this.name = n;
      }
    }

    public String toString() {
      return name;
    }

    public Object getBean() {
      return bean;
    }
  }

  /**
   * A renderer to render the tree nodes with the icon specified
   * in the bean info.
   */
  class EvolverTreeRenderer extends DefaultTreeCellRenderer {

    Hashtable icons = new Hashtable();
    ImageIcon proj;

    class NodeInfo {
      ImageIcon icon;
      String tip;

      public NodeInfo(ImageIcon icon, String tip) {
        this.icon = icon;
        this.tip = tip;
      }
    }

    public EvolverTreeRenderer() {
      proj = new ImageIcon(this.getClass().getResource("/uchicago/src/simbuilder/images/Box.gif"));
    }

    public Component getTreeCellRendererComponent(JTree tree, Object value,
                                                  boolean sel, boolean expanded, boolean leaf,
                                                  int row, boolean hasFocus) {
      JLabel label = (JLabel) super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf,
              row, hasFocus);

      EvolverBeanNode node = (EvolverBeanNode) value;
      if (node.isRoot()) {
        label.setIcon(proj);
      } else {
        NodeInfo info = (NodeInfo) icons.get(node.toString());
        if (info == null) {
          info = makeInfo(node);
        }

        if (info.icon != null) label.setIcon(info.icon);
        if (info.tip != null) EvolverBeanTree.this.setToolTipText(info.tip);
      }

      return label;
    }

    private NodeInfo makeInfo(EvolverBeanNode node) {
      BeanInfo bInfo = BeanIntrospector.getBeanInfo(node.getBean().getClass());
      Image image = bInfo.getIcon(bInfo.ICON_COLOR_16x16);
      NodeInfo info = null;
      if (image != null)
        info = new NodeInfo(new ImageIcon(image),
                bInfo.getBeanDescriptor().getShortDescription());
      else
        info = new NodeInfo(null,
                bInfo.getBeanDescriptor().getShortDescription());
      return info;

    }
  }
}



