package uchicago.src.simbuilder;

import uchicago.src.nqpy.compiler.AbstractCompilerException;
import uchicago.src.nqpy.compiler.CompilerException;
import uchicago.src.nqpy.compiler.NQPyClassFactory;
import uchicago.src.nqpy.compiler.ParserException;
import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.engine.SimModel;
import uchicago.src.sim.util.RepastException;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.simbuilder.util.JarClassLoader;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.jar.JarFile;

/**
 * The "mediator" for the various classes that make up the Evolver
 * tool. The EvolverController coordinates the various parts of the
 * gui. Receiving events (e.g. mouse clicks) and performing the
 * appropriate action.
 * <p/>
 * $Revision: 1.7 $ $Date: 2005/03/30 22:16:21 $
 */

public class EvolverController implements ActionListener,
        PropertyChangeListener {

  private PaletteItem paletteItem = null;
  private PaletteButton selButton = null;
  private MouseListener treeListener = null;
  private KeyListener treeKeyListener = null;
  private EvolverGui gui;

  private ByteCodeGenerator generator;
  private String curFile = "";
  private String curDir = ".";
  public static final String VERSION = "1.1.1 (20050426)";
  public static final float NUM_VERSION = 1.1f;
  private String projectName = "";
  private JarClassLoader loader;
  private boolean isArcGIS = false;

  public EvolverController() {
    generator = new ByteCodeGenerator();
    uchicago.src.sim.util.ByteCodeBuilder.loader = generator.getLoader();
    try {
      curDir = new File(".").getCanonicalPath();
    } catch (IOException e) {
      e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
    }
  }

  public boolean isArcGIS() {
    return isArcGIS;
  }

  public void setArcGIS(boolean arcGIS) {
    isArcGIS = arcGIS;
  }

  /**
   * Is an item on the palette selected?
   */
  public boolean isPaletteSelected() {
    return paletteItem != null;
  }

  public void setGui(EvolverGui g) {
    gui = g;
  }

  private void setProject(EvolverProject proj) {
    Hashtable typeList = (Hashtable) generator.getDetail(EvolverBean.DYN_TYPE_LIST);
    if (typeList != null) {
      for (Iterator iter = typeList.keySet().iterator(); iter.hasNext();) {
        NQPyClassFactory.removeDynamicType((String) iter.next());
      }
      typeList.clear();
    }


    //System.out.println(generator.getDetail(EvolverBean.DYN_TYPE_LIST));
    proj.addPropertyChangeListener(this);
    gui.setProject(proj);
    EvolverModel model = proj.getModel();
    if (model != null)
      projectName = model.displayName();
    else
      projectName = "";
    treeNodeSelected(gui.tree, gui.selectNodeByBean(proj));
    gui.tree.expandAll();
    setTitle();
  }

  private void setTitle() {
    if (!isArcGIS) {
      StringBuffer b = new StringBuffer(projectName);
      if (curFile.length() == 0)
        b.append(" []");
      else {
        File f = new File(curDir);
        b.append(" [");
        b.append(f.getAbsolutePath());
        b.append(File.separator);
        b.append(curFile);
        b.append("]");
      }
      gui.setTitle(b.toString() + " - Repast Py");
    } else {
      gui.setTitle("Agent Analyst");
    }
  }


  public void startNewProject() {
    setProject(new EvolverProject());
  }

  ChangeListener tabChange = new ChangeListener() {
    public void stateChanged(ChangeEvent evt) {
      JTabbedPane tab = (JTabbedPane) evt.getSource();
      if (tab.getSelectedIndex() == tab.indexOfTab("XML")) {
        StringWriter writer = new StringWriter();
        gui.getProject().saveXMLToString(writer);
        gui.setXML(writer.toString());
      } else {
        //System.out.println ("props tab");
      }
    }
  };

  void open(String openPath) {
    try {
      String dir = "";
      int index = openPath.lastIndexOf(File.separator);
      String file = openPath.substring(index + 1, openPath.length());
      if (index == -1)
        dir = ".";
      else
        dir = openPath.substring(0, index);

      File f = new File(dir + File.separator + file);

      if (!isArcGIS || curFile.length() == 0) {
        // check if curFile is length 0 so that
        // we can do the first load and not change the file 
        // after that.
        String path = f.getCanonicalPath();
        index = path.lastIndexOf(File.separator);
        curFile = path.substring(index + 1, path.length());
        curDir = f.getParentFile().getCanonicalPath();
      }

      XMLLoader xmlLoader = new XMLLoader(openPath);
      EvolverProject project = xmlLoader.load(loader);
      setProject(project);
      String infoName = project.getModel().getClass().getName() + "BeanInfo";
      Class clazz = Class.forName(infoName);
      ModelProducerBeanInfo info = (ModelProducerBeanInfo) clazz.newInstance();
      info.loadModelProducer(this, false);
    } catch (Exception ex) {
      ex.printStackTrace();
      String msg = ex.getMessage();
      if (msg == null) {
        msg = ex.toString();
      }
      gui.showError("Error opening project: " + msg);
    }
  }

  public void load(String filePath) {
    File file = new File(filePath);
    if (file.exists())
      open(filePath);
    else {
      gui.showError(filePath + " does not exist");
    }
  }

  /**
   * Loads the beans in the jar file on the bean palette.
   */
  public void loadJarFile(String[] jarFile) {
    try {
      if (loader != null)
        loader.addJarFile(jarFile[1]);
      else {

        URL url = new URL("file:" + jarFile);
        loader = new JarClassLoader(url, this.getClass().getClassLoader());
      }
    } catch (IOException ex) {
      gui.showError(ex.getMessage());
    }

    try {
      JarFile jar = new JarFile(jarFile[1]);
      gui.addJarToPalette(jarFile[0], loader, jar);
    } catch (IOException ex) {
      gui.showError(ex.getMessage());
    }
  }

  /*
  Action addJar = new AbstractAction("Add Jar File ...") {
    public void actionPerformed(ActionEvent evt) {
      String[] loadPath = gui.getJarPath(curDir, "");
      if (loadPath != null) {
        loadJarFile(loadPath);
      }
    }
  };
  */

  /**
   * Performed when the open icon or menu is selected.
   */
  Action open = new AbstractAction("Open...") {
    public void actionPerformed(ActionEvent evt) {
      int response = gui.askYesNoCancel("Save current model?");
      if (response == EvolverGui.YES) {
        String savePath = gui.getSavePath(curDir, curFile);
        if (savePath != null)
          saveAs(savePath);
        else
          return;
      } else if (response == EvolverGui.CANCEL) return;

      // this executes if we get a yes and a good save,
      // or a no to the save request. Cancel does nothing.
      String openPath = getOpenPath();
      if (openPath != null) open(openPath);
    }
  };

  public String getOpenPath() {
    return gui.getOpenPath(curDir, curFile);
  }

  /*
   * This is called after the save dialog comes up. Note that windows
   * natively checks for file overwrites and re-shows the save dialog.
   * Consequently, this only pops up an overwrite message on non-windows
   * platforms.
   */
  public void saveAs(String savePath) {
    int index = savePath.lastIndexOf(File.separator);
    if (!isArcGIS) {
      curFile = savePath.substring(index + 1, savePath.length());
      curDir = savePath.substring(0, index);
    }

    File outfile = new File(savePath);
    String os = System.getProperty("os.name").toUpperCase();
    if (!os.startsWith("WIN")) {
      if (outfile.exists()) {
        if (!gui.askYesNo("file \"" + savePath + "\" exists. " +
                "Do you want to replace it?"))
          return;
      }
    }


    try {
      gui.getProject().saveXMLToFile(savePath);
    } catch (IOException ex) {
      gui.showError("Save Error: " + ex.getMessage());
    }
    setTitle();
  }


  /**
   * Performed when the Save As menu item is selected.
   */
  Action saveAs = new AbstractAction("Save As...") {
    public void actionPerformed(ActionEvent evt) {
      String savePath = gui.getSavePath(curDir, curFile);
      if (savePath != null) saveAs(savePath);
    }
  };

  /**
   * Performed when the save icon or menu item is selected.
   */
  Action save = new AbstractAction("Save") {
    public void actionPerformed(ActionEvent evt) {
      try {
        if (curFile.length() == 0) {
          String savePath = gui.getSavePath(curDir, curFile);
          if (savePath != null) saveAs(savePath);
        } else {
          String savePath = curDir + File.separator + curFile;
          gui.getProject().saveXMLToFile(savePath);
        }
      } catch (IOException ex) {
        gui.showError("Save Error: " + ex.getMessage());
      }
    }
  };

  /**
   * Performed when the new icon or menu item is selected.
   */
  Action actNew = new AbstractAction("New...") {
    public void actionPerformed(ActionEvent evt) {
      String askText = "Save Current Model?";
      if (isArcGIS) askText = "Export Current Model?";
      int response = gui.askYesNoCancel(askText);
      if (response == EvolverGui.YES) {
        String savePath = gui.getSavePath(curDir, curFile);
        if (savePath != null)
          saveAs(savePath);
        else
          return;
      } else if (response == EvolverGui.CANCEL) return;

      chooseNewModel(true);
    }
  };

  /**
   * Popup gui dialog for choosing new model and select that model type.
   *
   * @param allowCancel whether or not the user can cancel out the dialog
   */
  public void chooseNewModel(boolean allowCancel) {
    if (!isArcGIS) curFile = "";

    NewModelDialog newModel = new NewModelDialog(allowCancel);
    newModel.display((JFrame) gui.getTopLevelAncestor());

    if (newModel.getSelectedModelInfo() != null) {
      ModelProducerBeanInfo info = newModel.getSelectedModelInfo();
      startNewProject();
      info.loadModelProducer(this, true);
    }
  }


  /**
   * Performed when the exit icon or menu item is selected.
   */
  Action exit = new AbstractAction("Exit") {
    public void actionPerformed(ActionEvent evt) {
      doExit();
    }
  };

  /**
   * Performed when the compile icon or menu item is selected.
   */
  Action compile = new AbstractAction("Compile") {
    public void actionPerformed(ActionEvent evt) {
      compile();
    }
  };

  private void exportToJava() {
    if (compile()) {

      ArrayList toExport = new ArrayList();
      EvolverProject proj = gui.getProject();
      String projDir = proj.getProjectDirectory();
      String packageName = proj.getPackageName();
      String modelName = proj.modelName();

      JFileChooser chooser = new JFileChooser(projDir);
      chooser.setCurrentDirectory(new File(projDir));
      chooser.setDialogTitle("Select Directory");
      chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
      String destDir = ".";
      if (chooser.showOpenDialog(gui) == JFileChooser.APPROVE_OPTION) {
        destDir = chooser.getSelectedFile().getAbsolutePath();
      } else {
        return;
      }


      String pkg = packageName.trim().toLowerCase().replace(' ', '_');
      pkg = pkg.replace('\'', '_');
      pkg = pkg.replace('\"', '_');
      String modelClassName = pkg + "." + modelName;
      toExport.add(modelClassName);

      for (Iterator iter = proj.getModel().compileables().iterator(); iter.hasNext();) {
        EvolverCompileable agent = (EvolverCompileable) iter.next();
        toExport.add(pkg + "." + agent.className());
      }

      String[] decompilerArgs = new String[5];
      decompilerArgs[0] = "-classpath";
      StringBuffer buf = new StringBuffer();
      buf.append(projDir);
      buf.append(",");
      String classpath = proj.getClassPath();
      classpath.replace(';', ',');
      buf.append(classpath);
      decompilerArgs[1] = buf.toString();
      if (decompilerArgs[1].endsWith(",")) decompilerArgs[1] = decompilerArgs[1].substring(0, decompilerArgs[1].length() - 1);
      decompilerArgs[2] = "-d";
      decompilerArgs[3] = destDir;

      for (int i = 0; i < toExport.size(); i++) {
        decompilerArgs[4] = toExport.get(i).toString();
        jode.decompiler.Main.decompile(decompilerArgs);
      }
    }
  }

  // returns true if compile is a success, otherwise false.
  private boolean compile() {
    boolean success = true;
    generator.reInit();
    uchicago.src.sim.util.ByteCodeBuilder.loader = generator.getLoader();
    //System.out.println ("Compiling ...");
    EvolverProject proj = gui.getProject();
    ArrayList paths = proj.getParsedClassPath();
    for (int i = 0, n = paths.size(); i < n; i++) {
      generator.addClassPath((String) paths.get(i));
    }
    NQPyClassFactory.removeAllDynamicTypes();

    try {
      proj.compile(generator);
    } catch (RepastException ex) {
      ex.printStackTrace();
      gui.showError(ex.getMessage());
      success = false;
    } catch (CompilerException ex) {
      generator = null;
      System.gc();
      ex.printSuperStackTrace();
      generator = new ByteCodeGenerator();
      uchicago.src.sim.util.ByteCodeBuilder.loader = generator.getLoader();
      proj.postCompile();
      onCompilerError(ex);
      success = false;
    } catch (ParserException ex) {
      generator = null;
      System.gc();
      generator = new ByteCodeGenerator();
      ex.printStackTrace();
      uchicago.src.sim.util.ByteCodeBuilder.loader = generator.getLoader();
      proj.postCompile();
      onCompilerError(ex);
      success = false;
    } catch (RuntimeException ex) {
      generator = new ByteCodeGenerator();
      uchicago.src.sim.util.ByteCodeBuilder.loader = generator.getLoader();
      ex.printStackTrace();
      String message = ex.getMessage();
      if (message == null || message.length() == 0) message = ex.toString();
      gui.showError(message);
      success = false;
    } catch (Error er) {
      // this may happen if the bytecode generated by the compiler is incorrect
      // in which case Nick has really screwed up
      generator = new ByteCodeGenerator();
      uchicago.src.sim.util.ByteCodeBuilder.loader = generator.getLoader();
      er.printStackTrace();
      String message = er.getMessage();
      if (message == null || message.length() == 0) message = er.toString();
      gui.showError(message);
      success = false;
    } finally {
      proj.postCompile();
    }

    return success;
  }

  /**
   * Performed when the run icon or menu item is selected.
   */
  Action run = new AbstractAction("Run") {
    public void actionPerformed(ActionEvent evt) {
      run();
    }
  };

  Action exportToJava = new AbstractAction("Export To Java") {
    public void actionPerformed(ActionEvent e) {
      exportToJava();
    }
  };

  public void run() {
    EvolverProject proj = gui.getProject();
    String projDir = proj.getProjectDirectory();
    String packageDir = proj.getPackageName();
    String modelName = proj.modelName();

    if (modelName == null) {
      gui.showError("Nothing to run!");
      return;
    }

    try {
      if (!compile()) return;
      String pkg = packageDir.trim().toLowerCase().replace(' ', '_');
      pkg = pkg.replace('\'', '_');
      pkg = pkg.replace('\"', '_');
      //System.out.println("pkg = " + pkg);
      Class c = generator.loadClass(pkg + "." + modelName, projDir);
      loadModel(c);
    } catch (IllegalAccessException ex) {
      ex.printStackTrace();
    } catch (ClassNotFoundException ex) {
      gui.showError("SimBuilder encountered an error while trying to run" +
              " the model.\nHave you compiled it yet?");
    } catch (Error er) {
       // this may happen if the bytecode generated by the compiler is incorrect
      // in which case Nick has really screwed up
      er.printStackTrace();
      String message = er.getMessage();
      if (message == null || message.length() == 0) message = er.toString();
      gui.showError(message);
    }
  }

  /**
   * Loads the palette for specified model type into the tool bar
   */
  public void loadPalette(String name) {
    gui.loadPalette(name);
  }

  /**
   * Adds the specified paletteItem to the node at the specified row.
   */
  public Object addBeanToTree(String tabName, String beanName, int row) {
    paletteItem = gui.getPaletteItemByName(tabName, beanName);
    return treeNodeSelected(gui.tree, gui.tree.getPathForRow(row));
  }

  public EvolverProject getProjectBean() {
    return gui.getProject();
  }

  private void loadModel(Class c) throws IllegalAccessException {
    SimModel model;
    try {
      model = (SimModel) c.newInstance();
    } catch (ClassCastException ex) {
      throw new IllegalArgumentException("Model does not implement the SimModel interface");
    } catch (InstantiationException ex2) {
      throw new IllegalArgumentException("Can't instantiate " + c);
    }

    SimInit init = new SimInit();
    init.setExitOnExit(false);

    String pathname = curDir + File.separator + "run_" + curFile;
    pathname = pathname.substring(0, pathname.length() - 3) + "pf";
    File pfFile = new File(pathname);
    if (pfFile.exists()) {
      try {
        // should be fine with exception here as we check if file exists beforehand
        init.loadModel(model, pfFile.getCanonicalPath(), false);
      } catch (IOException e) {
        e.printStackTrace();
      } catch (RuntimeException ex) {
        ex.printStackTrace();
        // try to load the model without the parameter file
        init.loadModel(model, "", false);
      }
    } else {
      init.loadModel(model, "", false);
    }
  }

  private void doExit() {
    int response = gui.askYesNoCancel("Save Current Model?");
    if (response == EvolverGui.YES) {
      String savePath = gui.getSavePath(curDir, curFile);
      if (savePath != null)
        saveAs(savePath);
      else
        return;
    } else if (response == EvolverGui.CANCEL) return;

    System.exit(0);
  }

  /**
   * Called when an EvolverBean on the palette is unselected either
   * via the mouse or programmatically.
   */
  public void unselect(Component c) {
    paletteItem = null;
    //System.out.println ("unselected");
    Component root = SwingUtilities.getRoot(c);
    // root may equal null if we are adding beans before 
    // the gui is displayed
    if (root != null) root.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    if (selButton != null) selButton.unselect();

  }

  /**
   * Called when an EvolverBean on the palette is selected.
   */
  public void actionPerformed(ActionEvent e) {
    PaletteButton b = (PaletteButton) e.getSource();
    selButton = b;
    paletteItem = b.getItem().getItem();
    b.setRootCursor();
  }


  /*
   * Called when a EvolverBean in the project tree is selected. If
   * a palette item is currently selected then this will attempt
   * to add that item to the selected bean in the project tree. If
   * no palette item is selected (not adding beans to the current project) The
   * selected bean will be introspected and the results displayed in the
   * property pane by the appropriate editors.
   *
   * Returns the bean associated with the tree node.
   */
  private Object treeNodeSelected(EvolverBeanTree tree, TreePath path) {
    Object bean = null;
    if (paletteItem == null) {
      //System.out.println("Show properties for selected item");
      bean = tree.getSelectedBean(path);
    } else {
      try {
        bean = paletteItem.makeBean();
        BeanInfo info = Introspector.getBeanInfo(bean.getClass());
        EvolverBean child = (EvolverBean) bean;
        EvolverBean parent = (EvolverBean) tree.getSelectedBean(path);
        String childName = info.getBeanDescriptor().getName();
        if (!(parent instanceof EvolverParent)) {
          String parentName = (Introspector.
                  getBeanInfo(parent.getClass())).
                  getBeanDescriptor().getName();

          gui.showError("Cannot add " + childName + " to " + parentName);
          return null;
        }
        child.addToParent((EvolverParent) parent);

        if (bean instanceof EvolverCompileable) {
          childName = ((EvolverCompileable) bean).className();
          ((EvolverCompileable) bean).addPropertyChangeListener(this);
        }

        path = tree.addNode(path, bean, childName);
        tree.addSelectionPath(path);
      } catch (IOException ex) {
        gui.showError("IO error instantiating bean");
      } catch (IntrospectionException ex) {
        gui.showError("Bean Introspection error");
        ex.printStackTrace();
      } catch (ClassNotFoundException ex) {
        gui.showError("Bean class not found error");
        ex.printStackTrace();
      } catch (RepastException ex) {
        // throws RepastException if child cannot be added to parent
        gui.showError(ex.getMessage());
        return null;
      } finally {
        unselect(tree);
      }
    }
    gui.setObject(bean);
    return bean;
  }

  public void onCompilerError(AbstractCompilerException ex) {
    //System.out.println (ex.getMessage ());
    String block = ex.getBlockName();
    ActionCodeProducer producer = EvolverModelCompiler.beanTableLookup(block);
    TreePath path = gui.selectNodeForObject(producer);
    treeNodeSelected(gui.tree, path);
    gui.showSourceEditor(producer, ex);

    /*
    EvolverBean bean = EvolverModelCompiler.beanTableLookup(block);
    block = block.substring(0, block.indexOf("."));
    TreePath path = gui.selectNodeByBean(block);
    EvolverCompileable bean =
      (EvolverCompileable)treeNodeSelected(gui.tree, path);
    gui.showSourceEditor(bean, ex);
    */
  }

  /**
   * Returns a KeyListener for the project tree. Allows for the
   * deletion of beans from the project tree.
   */
  public KeyListener getTreeKeyListener() {
    if (treeKeyListener == null) {
      treeKeyListener = new KeyAdapter() {
        public void keyPressed(KeyEvent evt) {
          if (evt.getKeyCode() == KeyEvent.VK_DELETE ||
                  evt.getKeyCode() == KeyEvent.VK_BACK_SPACE) {
            EvolverBeanTree tree = (EvolverBeanTree) evt.getSource();
            TreePath path = tree.getSelectionPath();
            if (path != null) {
              if (path.getLastPathComponent() != tree.getModel().getRoot()) {
                Object bean = tree.getSelectedBean(path);
                if (!(bean instanceof EvolverModel)) {
                  // cannot delete a model anymore
                  int result = JOptionPane.showConfirmDialog(gui, "Delete selected item from the project tree?", "Delete?", JOptionPane.YES_NO_OPTION);
                  if (result == JOptionPane.YES_OPTION) {
                    tree.removeNode(path);
                    treeNodeSelected(tree, path.getParentPath());
                  }
                }
              }
            }

          }
        }
      };
    }

    return treeKeyListener;
  }

  /**
   * Returns a MouseListener for the project bean tree. This calls
   * treeNodeSelected (see this class) when appropriate.
   */
  public MouseListener getTreeListener() {
    if (treeListener == null) {
      treeListener = new MouseAdapter() {
        public void mouseClicked(MouseEvent evt) {
          EvolverBeanTree tree = (EvolverBeanTree) evt.getSource();
          TreePath path = tree.getPathForLocation(evt.getX(), evt.getY());
          if (path != null) {
            treeNodeSelected(tree, path);
          }
        }
      };
    }

    return treeListener;
  }

  public void propertyChange(PropertyChangeEvent evt) {
    Object o = evt.getSource();
    if (o instanceof EvolverModel) {
      EvolverModel model = (EvolverModel) o;
      String name = model.displayName();
      if (!projectName.equals(name)) {
        projectName = name;
        setTitle();
      }

      if (evt.getPropertyName().equals(PropertyName.DISPLAY_NAME.getName())) {
        gui.renameTreeNode(o, (String) evt.getOldValue(), (String) evt.getNewValue());
      }
    } else if (o instanceof EvolverCompileable && evt.getPropertyName().equals(PropertyName.CLASS_NAME.getName())) {
      gui.renameTreeNode(o, (String) evt.getOldValue(), (String) evt.getNewValue());
    }
  }

  public void setFilename(File file) {
    try {
      if (file.getParent() == null) {
        file = new File(file.getCanonicalPath());
      }
      curDir = file.getParentFile().getCanonicalPath();
      String filename = file.getCanonicalPath();
      curFile = filename.substring(filename.lastIndexOf(File.separator), filename.length());
    } catch (IOException ex) {
      ex.printStackTrace();
    }
  }
}
