package uchicago.src.simbuilder;

import uchicago.src.nqpy.compiler.NQPyBlockCompiler;
import uchicago.src.nqpy.compiler.NQPyClassFactory;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.nqpy.compiler.CompilerException;
import uchicago.src.nqpy.parser.TokenMgrError;
import uchicago.src.sim.engine.DefaultGroup;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;

import java.util.*;
import java.io.StringWriter;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;

/**
 * Encapsulates EvolverModel compilation.
 */

public class EvolverModelCompiler {

  private ArrayList setupStmts = new ArrayList();
  private ArrayList buildStmts = new ArrayList();
  private ArrayList displayStmts = new ArrayList();
  private ArrayList fields = new ArrayList();
  private ArrayList compActions = new ArrayList();

  private static HashMap nameClass = new HashMap();
  private static HashMap beanTable = new HashMap();

  private EvolverModel model;
  private ByteCodeGenerator generator;
  private String pkg;

  static {
    nameClass.put("double", double.class);
    nameClass.put("float", float.class);
    nameClass.put("int", int.class);
    nameClass.put("long", long.class);
    nameClass.put("boolean", boolean.class);
    nameClass.put("Object", Object.class);
    nameClass.put("java.lang.Object", Object.class);
  }

  class Field {
    String fType;
    String fname;
    Class fClass;
    boolean createGet = false;

    public Field(String fType, String fname) {
      this.fType = fType;
      this.fname = fname;
    }

    public Field(String fName, Class fClass, boolean createGet) {
      this.fType = fClass.getName();
      this.fClass = fClass;
      this.fname = fName;
      this.createGet = createGet;
    }

    public boolean createGet() {
      return createGet;
    }
  }

  // integrates the agent producer initalization and setup code with
  // the produced model.
  private void processAgents() {
    List compileables = model.compileables();
    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        String groupName = producer.getGroupName();
        fields.add(new Field(groupName, DefaultGroup.class, true));
        setupStmts.add(producer.getNQPySetup());
        buildStmts.add(producer.getNQPyInit());
      }
    }
  }

  private void processComponents() {
    List comps = model.components();
    for (int i = 0; i < comps.size(); i++) {
      EvolverComponent comp = (EvolverComponent) comps.get(i);
      Hashtable compFields = comp.fields();
      Enumeration e = compFields.keys();
      while (e.hasMoreElements()) {
        String compName = (String) e.nextElement();
        String className = (String) compFields.get(compName);
        fields.add(new Field(className, compName));
      }

      String stmt = comp.setupStmts();
      if (stmt != null) setupStmts.add(stmt);
      stmt = comp.buildStmts();
      if (stmt != null) buildStmts.add(stmt);
      stmt = comp.displayStmts();
      if (stmt != null) displayStmts.add(stmt);

      ArrayList methods = comp.methods();
      if (methods != null) compActions.addAll(methods);

      //ScheduleDetails dets = comp.schedule();


      // step null problem occurs in here somewhere
      //model.masterSchedule(); //.addScheduleDetails(dets);
    }

    model.masterSchedule();
  }

  private void buildFields() {
    for (int i = 0; i < fields.size(); i++) {
      Field f = (Field) fields.get(i);
      generator.createPrivateField(f.fname, f.fType);
      if (f.createGet()) {
        String capName = Character.toUpperCase(f.fname.charAt(0)) + f.fname.substring(1);
        generator.createGetMethodRetField("get" + capName, f.fname, f.fClass);
      }
    }
  }

  private void buildActions() {
    if (model.preSetupStmts().length() > 0) setupStmts.add(0, model.preSetupStmts());
    if (model.preBuildStmts().length() > 0) buildStmts.add(0, model.preBuildStmts());
    if (model.preDisplayStmts().length() > 0) displayStmts.add(0, model.preDisplayStmts());
    buildStmts.add(model.buildStmts());
    setupStmts.addAll(1, model.parameters().getSetupStmts());
    model.actions().addToAction("_setup", setupStmts);
    model.actions().addToAction("_build", buildStmts);
    model.actions().addToAction("_display", displayStmts);

    Actions actions = model.masterSchedule().createActions(model.actions());
    //String modelTypeName = (String) generator.getDetail(EvolverBean.MODEL_NAME);
    //String agentTypeName = (String) generator.getDetail(EvolverBean.AGENT_NAME);
    //actions.setDynTypeNames(modelTypeName, agentTypeName);
    actions.compile(generator, pkg + "." + model.className());
    compileCompActions();

    model.actions().remove("_schedule");
  }

  private void compileCompActions() {
    String selfName = pkg + "." + model.className();
    Hashtable h = (Hashtable) generator.getDetail(EvolverBean.DYN_TYPE_LIST);
    NQPyDynamicClass dt = (NQPyDynamicClass) h.get(selfName);

    /*
    for (int i = 0; i < actions.size(); i++) {
      ActionCode code = (ActionCode)actions.get(i);
      System.out.println(code.getName());
      System.out.println(code.getCode());
    }
    */

    for (int i = 0; i < compActions.size(); i++) {
      ActionCode code = (ActionCode) compActions.get(i);

      //System.out.println (code.getName ());
      //System.out.println (code.getCode ());

      String retType = code.getRetType();
      NQPyBlockCompiler compiler;
      if (retType.equals("void")) {
        compiler = new NQPyBlockCompiler(code.getName(), generator.getClassFile());
      } else {
        try {
          Class retClass = (Class) nameClass.get(retType);
          if (retClass == null) retClass = Class.forName(retType);
          compiler = new NQPyBlockCompiler(code.getName(), retClass,
                  generator.getClassFile());
        } catch (ClassNotFoundException e) {
          CompilerException ex = new CompilerException("Return type '" + retType + "' not found", e);
          ex.setBlockName(selfName + "." + code.getName());
          ex.setLineNum(-1);
          throw ex;
        }
      }

      compiler.setSelf(dt);
      String blockName = selfName + "." + code.getName();
      compiler.setBlockName(blockName);
      EvolverModelCompiler.putInBeanTable(blockName, code.getDefiningProducer());
      ArrayList imports = code.getImports();
      for (int j = 0; j < imports.size(); j++) {
        //System.out.println("ADDING IMPORT " + (String)imports.get(j));
        compiler.addImport((String) imports.get(j));
      }
      try {
        compiler.compile(code.getCode());
      } catch (TokenMgrError er) {
        CompilerException ex = new CompilerException(er.getMessage());
        ex.setBlockName(blockName);
        ex.setLineNum(er.errorLine);
        throw ex;
      }
    }

    //postCompile();
  }

  public static void putInBeanTable(String blockName, ActionCodeProducer bean) {
    beanTable.put(blockName, bean);
  }

  public static ActionCodeProducer beanTableLookup(String blockName) {
    ActionCodeProducer bean = (ActionCodeProducer) beanTable.get(blockName);
    if (bean == null) throw new IllegalArgumentException("No defining producer found");
    return bean;
  }

  private void clear() {
    setupStmts.clear();
    buildStmts.clear();
    displayStmts.clear();
    fields.clear();
    compActions.clear();
    //model.masterSchedule().clear();
  }

  public void init(EvolverModel model, String pkg) {
    this.model = model;
    this.pkg = pkg;
  }

  public void compile(ByteCodeGenerator gen, String dir) {
    NQPyClassFactory.addImport("anl.repast.gis.data.*");
    NQPyClassFactory.addImport("anl.repast.gis.display.*");
    generator = gen;
    generator.createClass(pkg + "." + model.className(), model.getDynamicSuperclass());
    generator.createDefaultConstructor(null);
    model.parameters().compile(generator, "params");
    generator.initStringField("name", model.displayName(), true);
    generator.finishCtor();

    processComponents();
    processAgents();
    buildFields();
    buildActions();
    buildMain();
    generator.finishClass();
    generator.writeToFile(pkg + "." + model.className(), dir);
    clear();
  }
  
  private void buildMain() {
    NQPyBlockCompiler compiler = new NQPyBlockCompiler(generator.getClassFile());
    compiler.createMainMethod();
    VelocityContext context = new VelocityContext();
    context.put("modelName", model.className());
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "ModelMain.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    //System.out.println(stmt);
    compiler.compile(stmt.toString());
  }
}
