package uchicago.src.simbuilder;

import uchicago.src.guiUtils.Token;
import uchicago.src.guiUtils.Tokenizer;

import java.io.*;
import java.util.HashSet;


public class NQPyTokenizer implements Tokenizer {

  public static final int EOF = -1;
  public static final int KEYWORD = 0;
  public static final int STRING = 1;
  public static final int NUMBER = 2;
  public static final int COMMENT = 3;
  public static final int TEXT = 4;
  public static final int ORDINARY_CHAR = 5;
  public static final int ONE_CHAR_STRING = 6;


  private Lexer lexer;

  public NQPyTokenizer() {
    lexer = new Lexer();
  }

  public void setReader(Reader reader) {
    lexer.reset(reader);
  }

  public void setString(String str) {
    lexer.reset(str);
  }

  public Token nextToken() throws IOException {
    int tok = lexer.nextToken();
    switch (tok) {

      case NQPyTokenizer.EOF:
        return new Token(Token.EOF, "", -1, -1, -1);

      case NQPyTokenizer.KEYWORD:
        return new Token(Token.KEYWORD, lexer.getString(), lexer.getLineNum(),
                lexer.startCol, lexer.col - lexer.startCol - 1);

      case NQPyTokenizer.NUMBER:
        return new Token(Token.NUMBER, "num",
                lexer.getLineNum(),
                lexer.startCol, lexer.col - lexer.startCol - 1);

      case NQPyTokenizer.COMMENT:
        // this will returned after a new line so we need to col
        // vals back to 0
        int startCol = lexer.startCol;
        int col = lexer.col;
        lexer.startCol = lexer.col;

        return new Token(Token.COMMENT, lexer.getString(), lexer.getLineNum(),
                startCol, col - startCol - 1);

      case NQPyTokenizer.STRING:
        startCol = lexer.startCol;
        col = lexer.col;
        lexer.startCol = col;

        return new Token(Token.STRING, lexer.getString(), lexer.getLineNum(),
                startCol, col - startCol - 1);

      case NQPyTokenizer.ORDINARY_CHAR:
        return new Token(Token.TEXT, lexer.getString(), lexer.getLineNum(),
                lexer.startCol - 1, 1);

      case NQPyTokenizer.ONE_CHAR_STRING:
        return new Token(Token.STRING, lexer.getString(), lexer.getLineNum(),
                lexer.startCol - 1, 1);

      default:
        return new Token(Token.TEXT, lexer.getString(), lexer.getLineNum(),
                lexer.startCol, lexer.col - lexer.startCol - 1);
    }
  }

  class LexerReader extends Reader {

    Lexer lexer;
    private Reader reader;

    public LexerReader(Lexer l, Reader reader) {
      super();
      this.reader = reader;
      lexer = l;
      lexer.col = 0;
    }

    public void close() throws IOException {
      reader.close();
    }

    public int read(char[] buf, int off, int len) throws IOException {
      return reader.read(buf, off, len);
    }

    public int read() throws IOException {
      lexer.col++;
      return reader.read();
    }
  }

  class Lexer {

    private static final int COMMENT = 0;
    private static final int STRING = 1;
    private static final int INITIAL = 2;

    private StreamTokenizer input;
    int col;
    int startCol;
    private int state;
    private char stringChar;
    private HashSet keywords = new HashSet();

    public Lexer() {
      state = INITIAL;
      keywords.add("or");
      keywords.add("and");
      keywords.add("not");
      keywords.add("is");
      keywords.add("in");
      keywords.add("lambda");
      keywords.add("if");
      keywords.add("else");
      keywords.add("elif");
      keywords.add("for");
      keywords.add("try");
      keywords.add("except");
      keywords.add("def");
      keywords.add("class");
      keywords.add("finally");
      keywords.add("print");
      keywords.add("pass");
      keywords.add("break");
      keywords.add("continue");
      keywords.add("return");
      keywords.add("import");
      keywords.add("from");
      keywords.add("del");
      keywords.add("raise");
      keywords.add("global");
      keywords.add("exec");
      keywords.add("assert");
      keywords.add("as");
      keywords.add("while");
    }

    public void reset(String str) {
      reset(new StringReader(str));
    }

    public void reset(Reader r) {
      input = new StreamTokenizer(new LexerReader(this, r));
      input.resetSyntax();
      input.parseNumbers();
      input.wordChars('a', 'z');
      input.wordChars('A', 'Z');
      input.wordChars('_', '_');
      input.wordChars('!', '!');
      input.whitespaceChars('\n', '\n');
      input.eolIsSignificant(true);

      startCol = col = 0;
      state = INITIAL;
    }

    public String getString() {
      return input.sval;
    }

    public double getNumber() {
      return input.nval;
    }

    public int getLineNum() {
      return input.lineno();
    }

    public int getCol() {
      return col;
    }

    public void pushBack() {
      input.pushBack();
    }

    public int nextToken() throws IOException {

      int token = NQPyTokenizer.EOF;

      while (true) {
        switch (input.nextToken()) {

          case StreamTokenizer.TT_EOF:
            return NQPyTokenizer.EOF;

            // '#'
          case 35:
            if (state == INITIAL)
              state = COMMENT;
            else if (state == STRING) {
              startCol = col;
              return NQPyTokenizer.ONE_CHAR_STRING;
            }
            break;

          case StreamTokenizer.TT_EOL:
            if (state == COMMENT) {
              state = INITIAL;
              return NQPyTokenizer.COMMENT;
            }

            break;

          case StreamTokenizer.TT_NUMBER:
            if (state == INITIAL) return NQPyTokenizer.NUMBER;
            if (state == STRING) return NQPyTokenizer.STRING;
            break;

          case StreamTokenizer.TT_WORD:
            if (state == INITIAL) {
              if (keywords.contains(input.sval)) return NQPyTokenizer.KEYWORD;
              return NQPyTokenizer.TEXT;
            }

            if (state == STRING) return NQPyTokenizer.STRING;
            break;

          default:
            char c = (char) input.ttype;
            if (state == INITIAL) {
              //input.sval = String.valueOf((char)input.ttype);
              if (c == ' ' || c == '\t' || c == '\n' || c == '\r' ||
                      c == '\f') {
                startCol = col;
                break;
              } else if (c == '"' || c == '\'') {
                state = STRING;
                stringChar = c;

                startCol = col;
                return NQPyTokenizer.ORDINARY_CHAR;

              } else {
                startCol = col;
                return NQPyTokenizer.ORDINARY_CHAR;
              }
            } else if (state == STRING) {
              if (c == stringChar) {
                state = INITIAL;
                startCol = col;
                return NQPyTokenizer.ORDINARY_CHAR;
              } else {
                startCol = col;
                return NQPyTokenizer.ONE_CHAR_STRING;
              }
            }
        }

        // return Token.EOF;
      }
    }
  }

  public static void main(String[] args) {
    java.util.HashMap test = new java.util.HashMap();
    test.put(new Integer(0), "KEYWORD");
    test.put(new Integer(1), "STRING");
    test.put(new Integer(2), "NUMBER");
    test.put(new Integer(3), "COMMENT");
    test.put(new Integer(4), "TEXT");
    test.put(new Integer(5), "ORDINARY_CHAR");

    try {
      InputStreamReader reader = new InputStreamReader(System.in);
      NQPyTokenizer tok = new NQPyTokenizer();
      tok.setReader(reader);
      Token t = tok.nextToken();
      while (t.type != Token.EOF) {
        //System.out.println(test.get(new Integer(t.type)) + ": " + t.text +
        //" at " + t.lineEnd + ", " + t.start + ", " +
        // t.length);
        t = tok.nextToken();
      }
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }
}













