package uchicago.src.simbuilder;

import javax.swing.*;
import java.awt.event.ActionListener;
import java.beans.IntrospectionException;
import java.util.Enumeration;

/**
 * Panel for displaying EvolverBeans as PaletteBeanItems. The
 * actual selection events for beans is handled by the EvolverController.
 *
 * @author Nick Collier
 * @version $Revision: 1.2 $ $Date: 2005/04/25 13:25:54 $
 */

public class PalettePanel extends JPanel {

  private String paletteName;
  private ButtonGroup group = new ButtonGroup();
  private PaletteButton selector;
  private ActionListener listener;

  /**
   * Creates a PalettePanel from the String[] array of bean names,
   * specified palette name, the specified ActionListener and the
   * specified PaletteButton.
   *
   * @param beans    a String[] whose elements are the fully-qualified
   *                 names of EvolverBeans.
   * @param name     the name to display for this particular palette.
   * @param listener the actionListener that is called when a Bean
   *                 is selected. This will be the EvolverController.
   * @param sel      the selecton (arrow) button.
   */

  public PalettePanel(String[] beans, String name, ActionListener listener,
                      PaletteButton sel) {
    this(beans, name, listener, sel, listener.getClass().getClassLoader());
  }

  public PalettePanel(String[] beans, String name, ActionListener listener,
                      PaletteButton sel, ClassLoader loader) {
    super();
    setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
    paletteName = name;
    this.listener = listener;
    selector = sel;
    try {
      for (int i = 0; i < beans.length; i++) {
        addButton(beans[i], loader);
      }
    } catch (Exception ex) {
      ex.printStackTrace();
    }

    group.add(selector);
  }

  public void addHiddenModel(String className) {
    PaletteBeanItem item = null;
    try {
      item = new PaletteBeanItem(className);
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    PaletteButton b = new PaletteButton(item, selector);
    group.add(b);
  }

  /**
   * It seems a button can only be in a single button group, so we need
   * to some way to insure that the selector is part of the currently
   * selected panel's group.
   */
  public void addSelectorToGroup() {
    group.add(selector);
  }

  /**
   * Adds a button to this palette. Given the specified beanName,
   * a PaletteBeanItem is created and this item is placed as a button
   * on this palette.
   */
  public void addButton(String beanName) throws ClassNotFoundException,
          IntrospectionException {
    PaletteBeanItem item = new PaletteBeanItem(beanName);
    PaletteButton b = new PaletteButton(item, selector);
    b.addActionListener(listener);
    add(b);
    group.add(b);
  }

  /**
   * Adds a button to this palette. Given the specified beanName,
   * a PaletteBeanItem is created and this item is placed as a button
   * on this palette.
   */
  public void addButton(String beanName, ClassLoader loader) throws
          ClassNotFoundException, IntrospectionException {
    Class clazz = loader.loadClass(beanName);
    PaletteBeanItem item = new PaletteBeanItem(beanName, clazz);
    PaletteButton b = new PaletteButton(item, selector);
    b.setRolloverEnabled(true);
    b.addActionListener(listener);
    add(b);
    group.add(b);
  }

  public PaletteItem getItemByName(String name) {
    for (Enumeration elements = group.getElements(); elements.hasMoreElements();) {
      PaletteButton button = (PaletteButton) elements.nextElement();
      if (button.getItem().getBeanName().equals(name)) {
        return button.getItem();
      }
    }

    return null;
  }


  public String getName() {
    return paletteName;
  }
}
