package uchicago.src.simbuilder.beans;

import uchicago.src.simbuilder.util.BooleanMessage;
import uchicago.src.simbuilder.util.RepastPyUtilities;
import org.w3c.dom.Element;

import java.util.*;

/**
 * AbstractAgentProducer producer for GIS agents.
 *
 * @author Nick Collier
 * @version $Revision: 1.3 $ $Date: 2004/12/14 22:10:48 $
 */
public abstract class AbstractAgentProducer extends AbstractEvolverAgent implements AgentProducer {
  
  public static final String NO_PARENT = "(None)";

  protected String groupName = "group";
  protected ConstrainedProperty possibleParents;
  private String initialSuperClassName;
  private boolean isParent = false;

  public AbstractAgentProducer() {
    String[] vals = {NO_PARENT};
    String[] views = {NO_PARENT};
    possibleParents = new ConstrainedProperty(vals, 0, views);
  }

  public ConstrainedProperty getPossibleSuperClasses() {
    return possibleParents;
  }

  public void setPossibleSuperClasses(ConstrainedProperty possibleParents) {
    this.possibleParents = possibleParents;
  }

 
    

  /**
   * Returns the AgentProducer that is the "parent" of this AgentProducer. May return null;
   *
   * @return
   */
  public AgentProducer getParent() {
    Object obj = possibleParents.getSelected();
    AgentProducer parent = obj.equals(NO_PARENT) ? null : (AgentProducer)obj;
    return parent;
  }

  public void addPossibleParentClass(AgentProducer producer) {
    if (producer.getClass().equals(this.getClass())) {
      // restrict parent to same kind of producer
      possibleParents.add(producer.className(), producer);
      possibleParents.sort();
    }
  }
  
  public void removePossibleParentClass(AgentProducer producer) {
    Object obj = possibleParents.getSelected();
    possibleParents.removePossible(producer);
    if (producer.equals(obj)) {
      possibleParents.setSelected(NO_PARENT);
    } else {
      possibleParents.setSelected(obj);
    }
  }
  
  public void updatePossibleParent(AgentProducer producer, String newName) {
    possibleParents.edit(producer, newName);
  }

  /**
   * Returns the nqpy that becomes part of the model's build (not the produced agent's)
   * method.
   *
   * @return
   */
  public String getNQPyInit() {
    String stmt = "for __agent__ as " + getName() +
            " in self." + getGroupName() + ": \n";
    StringBuffer b = new StringBuffer(stmt);
    b.append("\t__agent__.setModel(self)\n");
    b.append("\n");
    return b.toString();
  }

  protected String getFQName() {
    StringBuffer buf = new StringBuffer();
    buf.append(pkg);
    buf.append(".");
    buf.append(getName());
    return buf.toString();
  }

  /**
   * NQPy that becomes part of the model's setup (not the produced agent's setup).
   *
   * @return
   */
  public String getNQPySetup() {
    // groupName = DefaultGroup()
    return "self." + getGroupName() + " = DefaultGroup(jclass(\"" + getFQName() + "\"), \"step\")\n";
  }

  /**
   * Implementation of EvolverBean.schedule(). This returns the ScheduleDetails
   * used for compilation, as opposed to that which is the schedule property.
   */
  public ScheduleDetails schedule() {

    String stmt = "for __agent__ as " + getName() +
            " in self." + getGroupName() + ": \n";

    for (int i = 0; i < details.size(); i++) {
      ScheduleDetail d = details.getDetail(i);
      StringBuffer b = new StringBuffer(stmt);
      b.append("\t__agent__.");
      b.append(d.getActionName());
      b.append("()\n");
      d.setNQPy(b.toString());
      d.setTarget("for all " + name);
    }
    return details;
  }

  public void setGroupName(String groupName) {
    this.groupName = groupName;
  }

  public String getGroupName() {
    return groupName;
  }

  /**
   * Returns a BooleanMessage indicating whether or not this AgentProducer is in a compilable
   * state.
   * @return
   */ 
  public BooleanMessage isValid() {
    if (!RepastPyUtilities.isValidJavaIdentifier(className())) {
      return new BooleanMessage("Invalid agent name '" + className() + "'. Valid agent names may contain only letters, numbers, or '_'", false);
    }
    
    if (!RepastPyUtilities.isValidJavaIdentifier(getGroupName())) {
      return new BooleanMessage("Invalid group name '" + getGroupName() + "'. Valid group names may contain only letters, numbers, or '_'", false);
    }
    
    // follow chain of parent's parent's and see if any include this
    // AgentProducer
    Object selectedParent = getPossibleSuperClasses().getSelected();
    while (!selectedParent.equals(NO_PARENT)) {
      AbstractAgentProducer parent = (AbstractAgentProducer) selectedParent;
      if (selectedParent.equals(this)) {
        return new BooleanMessage("Invalid parent class for '" + className() + "'. Cyclic dependency detected.", false);
      }
      selectedParent = parent.getPossibleSuperClasses().getSelected();
    }
    
    
    Object obj = possibleParents.getSelected();
    if (!obj.equals(NO_PARENT)) {
      // check if parent has any fields named the same as this 
      Set paramNames = new HashSet();
      for (Iterator iter = parameters.namesIter(); iter.hasNext(); ) {
        paramNames.add(iter.next());
      }
      
      AbstractEvolverAgent parent = (AbstractEvolverAgent) obj;
      for (Iterator iter = parent.parameters.namesIter(); iter.hasNext(); ) {
        String paramName = iter.next().toString();
        if (paramNames.contains(paramName)) {
          return new BooleanMessage("Agent producer '" + getName() + "' cannot duplicate its parent's field '" +
                  paramName + "'", false);
        }
      }
      
      // check if any actions have same name but different return value
      Map actionMap = new HashMap();
      for (Iterator iter = actions.iterator(); iter.hasNext(); ) {
        ActionCode code = (ActionCode) iter.next();
        actionMap.put(code.getName(), code.getRetType());
      }
      
      for (Iterator iter = parent.actions.iterator(); iter.hasNext(); ) {
        ActionCode code = (ActionCode) iter.next();
        String retType = (String)actionMap.get(code.getName());
        if (retType != null && (!retType.equals(code.getRetType()))) {
          return new BooleanMessage("Agent producer '" + getName() + "' cannot override its parent's action '" +
                  code.getName() + "' with a different return type", false);
        }
      }
    }
    
    return BooleanMessage.TRUE;
  }


  /**
   * Returns true if this AgentProducer is the parent class of another AgentProducer.
   *
   * @return
   */
  public boolean isParent() {
    return isParent;
  }

  /**
   * Sets whether or not this AgentProducer is the parent of another agent producer.
   * @param isParent
   */ 
  public void setIsParent(boolean isParent) {
    this.isParent = isParent;
  }

  /**
   * Return the xml string for saving parent info
   * @return
   */ 
  protected String getParentXML() {
    return "\" superClass=\"" + possibleParents.getSelectedView();
  }
  
  /**
   * We can't set the superclass while loading because the produce may not know about
   * all the other producers yet. So, we need a separate method to call after all have been
   * loaded.
   */ 
  public void setInitialSuperClass() {
    if (initialSuperClassName != null) possibleParents.setSelectedView(initialSuperClassName);
  }
  
  public void loadFromXML(Element nodeE) {
    // loads name, actions, parameters and schedule
    super.loadFromXML(nodeE);
    setGroupName(nodeE.getAttribute("groupName"));
    initialSuperClassName = nodeE.getAttribute("superClass");
  }
}
