package uchicago.src.simbuilder.beans;

import org.w3c.dom.Element;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.simbuilder.property.EvolverPropertyEditor;
import uchicago.src.nqpy.compiler.AbstractCompilerException;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.sim.util.RepastException;

import javax.swing.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyEditorManager;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Abstract agent class for network models created with Evolver. When this
 * is compiled a default step method is created as well as ivar and
 * accessor methods for the EvolverModel for which this is an agent.
 * <p/>
 * $Revision: 1.2 $ $Date: 2004/12/10 21:59:42 $
 */
public abstract class AbstractEvolverAgent extends AbstractEvolverCompileable
        implements EvolverParent, ActionCodeProducer, Schedulable {

  protected String name = "MyAgent";
  protected Fields parameters = new Fields();
  protected Actions actions = new Actions();
  protected ScheduleDetails details = new ScheduleDetails();
  protected ArrayList compileables = new ArrayList();

  static {
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.Fields.class,
                    uchicago.src.simbuilder.beans.FieldsPropertyEditor.class);
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.Actions.class,
                    uchicago.src.simbuilder.beans.ActionsPropertyEditor.class);
  }

  public AbstractEvolverAgent() {
    actions.createDefaultAction("step", "");
    //actions.createDefaultAction("init", "## user defined init code below\n");
    details.addDetail(new ScheduleDetail("step", 1,
            ScheduleDetails.EVERY_TICK, false));
    pcs = new PropertyChangeSupport(this);
  }

  /**
   * Sets the name of this NetworkNodeProducer. This name becomes the
   * class name of class compiled from this NetworkNodeProducer.
   */
  public void setName(String val) {
    PropertyChangeEvent evt = new PropertyChangeEvent(this, 
            PropertyName.CLASS_NAME.getName(), name, val);
    pcs.firePropertyChange(evt);
    name = val;
  }

  /**
   * Gets the name of this NetworkNodeProducer. This name becomes the
   * class name of class compiled from this NetworkNodeProducer.
   */
  public String getName() {
    return name;
  }

  public void setParameters(Fields newParameters) {
    parameters = newParameters;
    actions.clearVarNames();
    actions.addVarName("model");
    Iterator i = parameters.namesIter();
    while (i.hasNext()) {
      actions.addVarName((String) i.next());
    }
  }

  public Fields getParameters() {
    return parameters;
  }

  public Actions getActions() {
    actions.clearVarNames();
    for (Iterator i = parameters.namesIter(); i.hasNext();) {
      actions.addVarName((String) i.next());
    }
    actions.addVarName("model");
    return actions;
  }

  public void setActions(Actions val) {
    for (Iterator iter = this.actions.iterator(); iter.hasNext();) {
      ActionCode code = (ActionCode) iter.next();
      // actions has that code, but val doesn't so action must have been
      // deleted.
      if (!val.hasAction(code.getName())) {
        details.removeDetailWithName(code.getName());

      }
    }

    actions = val;
    for (Iterator iter = this.actions.iterator(); iter.hasNext();) {
      ActionCode code = (ActionCode) iter.next();
      code.setDefiningProducer(this);
    }
  }

  /**
   * Gets the ScheduleDetails (the schedule property) for this
   * NetworkNodeProducer. Note that this is the property accessor and
   * NOT the EvolverBean.schedule() implementation.
   */
  public ScheduleDetails getSchedule() {
    ArrayList list = new ArrayList();
    Iterator i = actions.iterator();
    while (i.hasNext()) {
      ActionCode code = (ActionCode) i.next();
      list.add(code.getName());
    }

    details.setActionNames(list);
    return details;
  }

  /**
   * Gets the ScheduleDetails (the schedule property) for this
   * NetworkNodeProducer. Note that this is the property accessor and
   * NOT the EvolverBean.schedule() implementation.
   */
  public void setSchedule(ScheduleDetails val) {
    details = val;
  }


  /**
   * Adds this EvolverBean to the specified EvolverParent. This method is
   * corresponds to the accept method in the visitor pattern. Classes which
   * implement this method should call the appropriate method of the
   * EvolverParent with themselves as the parameter.
   *
   * @throws uchicago.src.sim.util.RepastException
   *          if the
   *          parent cannot successfully add the child
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    parent.addCompileable(this);
  }

  /**
   * Removes this NetworkNodeProducer from the specified parent.
   */
  public void removeFromParent(EvolverParent parent) {
    // mark all the schedule details associated with this
    // agent as deleted, so that the master schedule
    // will remove them.
    for (Iterator iter = actions.iterator(); iter.hasNext();) {
      ActionCode action = (ActionCode) iter.next();
      details.removeDetailWithName(action.getName());
    }
    parent.removeCompileable(this);
  }

  /**
   * Implenentation of EvolverBean.schedule(). This returns the ScheduleDetails
   * used for compilation, as opposed to that which is the schedule property.
   */
  public ScheduleDetails schedule() {
    String stmt = "for __agent__ as " + EvolverBean.AGENT_NAME_TEMPLATE +
            " in self.agentList: \n";

    for (int i = 0; i < details.size(); i++) {
      ScheduleDetail d = details.getDetail(i);
      StringBuffer b = new StringBuffer(stmt);
      b.append("\t__agent__.");
      b.append(d.getActionName());
      b.append("()\n");
      d.setNQPy(b.toString());
      d.setTarget("for all " + name);
    }
    return details;
  }

  public String className() {
    return name;
  }


  /**
   * Creates ivar and set method for the EvolverModel associated with
   * this NetworkNodeProducer.
   */
  public void makeModelSet(ByteCodeGenerator generator) {
    String modelName = (String) generator.getDetail(EvolverBean.MODEL_NAME);
    generator.createPublicSetMethod("setModel", void.class, modelName, "model");
  }


  /**
   * Performs any post compilation cleanup.
   */
  public void postCompile() {
  }

  /**
   * Adds the specified component to this EvolverParent
   */
  public void addComponent(EvolverComponent child) throws RepastException {
    throw new RepastException("Cannot add components to Nodes");
  }

  /**
   * Adds the specified compileable to this EvolverParent
   */
  public void addCompileable(EvolverCompileable child) throws RepastException {
    if (!(child instanceof EvolverAgentCompileable)) {
      throw new RepastException("Cannot add " + child.className() +
              " to EvolverNodes");
    }

    EvolverAgentCompileable eac = (EvolverAgentCompileable) child;
    eac.setEvolverCompileable(this);
    compileables.add(child);
  }

  /**
   * Removes the specified component to this EvolverParent
   */
  public void removeComponent(EvolverComponent child) {
  }

  /**
   * Removes the specified compileable to this EvolverParent
   */
  public void removeCompileable(EvolverCompileable child) {
    compileables.remove(child);
  }

  public void setPackage(String pkg) {
    this.pkg = pkg;
    for (int i = 0, n = compileables.size(); i < n; i++) {
      EvolverCompileable ec = (EvolverCompileable) compileables.get(i);
      ec.setPackage(pkg);
    }
  }

  protected void implementAsCustomProbeable(NQPyDynamicClass dt) {
    String[] paramNames = new String[parameters.size()];
    dt.addField("propNames", paramNames.getClass());
    dt.addMethod("getProbedProperties", paramNames.getClass(), new Class[]{});
  }

  protected void compileAsCustomProbeable(ByteCodeGenerator generator,
                                          String[] otherParams) {

    ArrayList list = new ArrayList();
    for (Iterator iter = parameters.fields().iterator(); iter.hasNext();) {
      FieldData p = (FieldData) iter.next();
      if (p.isProbeable()) {
        list.add(p.getName());
      }
    }

    String[] names = new String[list.size() + otherParams.length];
    for (int i = 0, n = list.size(); i < n; i++) {
      names[i] = (String) list.get(i);
    }

    System.arraycopy(otherParams, 0, names, list.size(), otherParams.length);
    generator.createStringArrayField("propNames", names);
    generator.createGetMethodRetField("getProbedProperties", "propNames",
            names.getClass());
  }

  /**
   * Returns a list of the children of this EvolverBean.
   */
  public ArrayList children() {
    return compileables;
  }

  public void displayEditor(JFrame frame, AbstractCompilerException ex) {
    ActionsPropertyEditor editor = new ActionsPropertyEditor();
    editor.setValue(getActions());
    editor.addPropertyChangeListener(this);
    ActionsPropertyPanel panel = new ActionsPropertyPanel(actions);
    panel.addError(ex);
    panel.setVarNames(actions.getVarNames());
    panel.display(frame, editor);

    setActions((Actions) editor.getValue());
  }


  /**
   * Saves an xml representation of this NetworkNodeProducer to the specified
   * output stream.
   *
   * @param out the BufferedWriter to write the xml to.
   */
  public void saveXML(BufferedWriter out) throws IOException {
    out.write(actions.toXML());
    out.newLine();
    out.write(parameters.toXML());
    out.newLine();
    getSchedule();
    out.write(details.toXML());
    out.newLine();

    for (int i = 0; i < compileables.size(); i++) {
      EvolverBean bean = (EvolverBean) compileables.get(i);
      bean.saveXML(out);
      out.newLine();
    }
  }

  public void loadFromXML(Element nodeE) {
    setName(nodeE.getAttribute("name"));
    setActions(XMLLoader.getActions(nodeE));
    setParameters(XMLLoader.getParameters(nodeE));
    setSchedule(XMLLoader.getScheduleDetails(nodeE));

    try {
      ArrayList beanList = XMLLoader.getChildBeans(nodeE);
      for (int i = 0; i < beanList.size(); i++) {
        EvolverBean bean = (EvolverBean) beanList.get(i);
        bean.addToParent(this);
      }
    } catch (RepastException ex) {
      ex.printStackTrace();
    }
  }

  public void propertyChange(PropertyChangeEvent evt) {
    if (evt.getSource() instanceof EvolverPropertyEditor) {
      EvolverPropertyEditor editor = (EvolverPropertyEditor) evt.getSource();
      editor.setPropertyFromValue(this);
    }
  }
}
