package uchicago.src.simbuilder.beans;

import org.w3c.dom.Element;
import uchicago.src.nqpy.compiler.AbstractCompilerException;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.sim.engine.DefaultGroup;
import uchicago.src.sim.util.RepastException;
import uchicago.src.simbuilder.property.EvolverPropertyEditor;
import uchicago.src.simbuilder.util.BooleanMessage;
import uchicago.src.simbuilder.util.RepastPyUtilities;

import javax.swing.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyEditorManager;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.*;

/**
 * Abstract base class for all Evolver model type beans. This implements
 * most of the methods in EvolverModel and thus EvolverCompileable,
 * and EvolverBean. It handles interaction with the compiler and with
 * other EvolverBeans. Any model that extends this should only
 * have to implement behavoir specific to that model.<b>
 * <p/>
 * There are methods here that seem to do the same thing (e.g.
 * actions() and getActions() which both return an Actions object. The
 * semantics are different however actions() returns an Actions object
 * suitable for use by the EvolverModelCompiler while getActions returns
 * the bean property actions.
 * <p/>
 * $Revision: 1.5 $ $Date: 2004/12/14 22:10:48 $
 */
public abstract class AbstractEvolverModel extends AbstractEvolverCompileable
        implements EvolverModel, ActionCodeProducer {

  protected ArrayList compileables = new ArrayList();
  protected ArrayList components = new ArrayList();

  protected EvSchedule schedule = new EvSchedule();
  protected ScheduleDetails details = new ScheduleDetails();
  protected Fields parameters = new Fields();
  protected Actions actions = new Actions();

  protected String className = "MyModel";
  protected String dispName = "My Model";

  protected ArrayList xmlProps = new ArrayList();

  // ensures that the appropriate editor is used when displaying the
  // value of the Evolver specific properties.
  static {
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.Fields.class,
                    uchicago.src.simbuilder.beans.FieldsPropertyEditor.class);
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.Actions.class,
                    uchicago.src.simbuilder.beans.ActionsPropertyEditor.class);
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.EvSchedule.class,
                    uchicago.src.simbuilder.beans.EvScheduleEditor.class);
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.network.NetworkSource.class,
                    uchicago.src.simbuilder.beans.network.NetworkSourceEditor.class);
  }

  public AbstractEvolverModel() {
    xmlProps.add("ModelName");
    xmlProps.add("DisplayName");
    xmlProps.add("Class");
  }

  /**
   * Returns the master schedule, that is, the merged view
   * of all the schedules in the model components. Note that this
   * is a bean property accessor. modelSchedule() gets the master
   * schedule used by the compiler.
   */
  public EvSchedule getMasterSchedule() {
    ArrayList list = new ArrayList();
    list.addAll(compileables);
    list.add(this);
    list.addAll(components);
    schedule.setEvolverBeans(list);
    return schedule;
  }

  public void setPackage(String pkg) {
    super.setPackage(pkg);
    for (int i = 0; i < compileables.size(); i++) {
      EvolverCompileable bean = (EvolverCompileable) compileables.get(i);
      bean.setPackage(pkg);
    }
  }

  /**
   * Sets the master schedule. See getMasterSchedule()
   */
  public void setMasterSchedule(EvSchedule val) {
    schedule = val;
  }

  /**
   * Sets the model's parameters.
   */
  public void setParameters(Fields newParameters) {
    parameters = newParameters;
    Iterator i = parameters.namesIter();
    actions.clearVarNames();
    while (i.hasNext()) {
      actions.addVarName((String) i.next());
    }
  }

  /**
   * Gets this model's parameters.
   */
  public Fields getParameters() {
    return parameters;
  }

  /**
   * Sets the model name.
   */
  public void setModelName(String val) {
    //System.out.println("setting name: " + val);
    pcs.firePropertyChange(PropertyName.CLASS_NAME.getName(), className, val);
    className = val;

  }

  /**
   * Gets the model name.
   */
  public String getModelName() {
    //System.out.println("getting name");
    return className;
  }


  public Actions getActions() {
    actions.clearVarNames();
    for (Iterator i = parameters.namesIter(); i.hasNext();) {
      actions.addVarName((String) i.next());
    }

    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        String groupName = producer.getGroupName();
        actions.addVarName(groupName);
      }
    }

    return actions;
  }

  public void setActions(Actions val) {
    for (Iterator iter = this.actions.iterator(); iter.hasNext();) {
      ActionCode code = (ActionCode) iter.next();
      // actions has that code, but val doesn't so action must have been
      // deleted.
      if (!val.hasAction(code.getName())) {
        details.removeDetailWithName(code.getName());
      }
    }

    actions = val;
    for (Iterator iter = this.actions.iterator(); iter.hasNext();) {
      ActionCode code = (ActionCode) iter.next();
      code.setDefiningProducer(this);
    }


  }

  public ScheduleDetails getSchedule() {
    ArrayList list = new ArrayList();
    Iterator i = actions.iterator();
    while (i.hasNext()) {
      ActionCode code = (ActionCode) i.next();
      list.add(code.getName());
    }

    details.setActionNames(list);
    return details;
  }

  public void setSchedule(ScheduleDetails val) {
    details = val;
  }

  public void setDisplayName(String val) {
    //System.out.println("setting name: " + val);
    if (!val.equals(dispName)) {
      String tmp = dispName;
      dispName = val;
      pcs.firePropertyChange(PropertyName.DISPLAY_NAME.getName(), tmp, dispName);
    }
  }

  public String getDisplayName() {
    //System.out.println("getting name");
    return dispName;
  }


  /**
   * Adds the specified component to this EvolverParent
   */
  public void addComponent(EvolverComponent child) throws RepastException {
    components.add(child);
  }

  /**
   * Adds the specified compileable to this EvolverParent
   */
  public void addCompileable(EvolverCompileable child) throws RepastException {
    child.addPropertyChangeListener(this);
    // add this child to the list of possible super classes
    // for the other agent producers
    if (child instanceof AgentProducer) {
      addAgentProducer((AgentProducer) child);
    }

    compileables.add(child);
  }

  private void addAgentProducer(AgentProducer newProducer) {
    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        producer.addPossibleParentClass(newProducer);
        newProducer.addPossibleParentClass(producer);
      }
    }
  }

  /**
   * Removes the specified component to this EvolverParent
   */
  public void removeComponent(EvolverComponent child) {
    components.remove(child);
    parameters.removeImplicitFields(child);
  }

  /**
   * Removes the specified compileable to this EvolverParent
   */
  public void removeCompileable(EvolverCompileable child) {
    // we first reset beans in the master schedule with getMasterSchedule
    // then process them. The child here should have marked 
    // all its details as deleted as part of it being deleted and so
    // processEvolverBeans() should delete all references to any scheduled
    // methods of that child.
    getMasterSchedule();
    schedule.processEvolverBeans();
    compileables.remove(child);
    child.removePropertyChangeListener(this);
    parameters.removeImplicitFields(child);
    if (child instanceof AgentProducer) {
      AgentProducer removedProducer = (AgentProducer) child;
      for (int i = 0, n = compileables.size(); i < n; i++) {
        Object o = compileables.get(i);
        if (o instanceof AgentProducer) {
          AgentProducer producer = (AgentProducer) o;
          producer.removePossibleParentClass(removedProducer);
        }
      }
    }
  }

  /**
   * Gets the classname for this dynamically compiled class. Implements
   * EvolverCompileable.className();
   */
  public String className() {
    return className;
  }

  /**
   * Adds this EvolverBean to the specified EvolverParent. This method is
   * corresponds to the accept method in the visitor pattern. Classes which
   * implement this method should call the appropriate method of the
   * EvolverParent with themselves as the parameter.
   *
   * @throws uchicago.src.sim.util.RepastException
   *          if the
   *          parent cannot successfully add the child
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    parent.addCompileable(this);
  }

  /**
   * Removes this EvolverBean from the specified parent.
   */
  public void removeFromParent(EvolverParent parent) {
    parent.removeCompileable(this);
  }

  /**
   * Gets the ScheduleDetails for this AbstractEvolverModel. Implements
   * EvolverBean.schedule()
   */
  public ScheduleDetails schedule() {
    for (int i = 0; i < details.size(); i++) {
      ScheduleDetail d = details.getDetail(i);
      StringBuffer b = new StringBuffer("self.");
      b.append(d.getActionName());
      b.append("();");
      d.setNQPy(b.toString());
      d.setTarget(className);
    }
    return details;
  }

  /**
   * Returns a list of the children of this EvolverBean.
   */
  public ArrayList children() {
    ArrayList list = new ArrayList();
    list.addAll(compileables);
    list.addAll(components);

    return list;
  }

  /**
   * Returns the ModelParameters associated with this model.
   */
  public Fields parameters() {
    return getParameters();
  }

  /**
   * Returns the Actions associated with this model.
   */
  public Actions actions() {
    return actions;
  }

  /**
   * Gets the List of EvolverCompileables associated with this model.
   */
  public List compileables() {
    return Collections.unmodifiableList(compileables);
  }

  /**
   * Gets the List of EvolverComponents associated with this model.
   */
  public List components() {
    return Collections.unmodifiableList(components);
  }

  /**
   * Gets the EvSchedule (the master schedule) associated with this model.
   */
  public EvSchedule masterSchedule() {
    EvSchedule s = getMasterSchedule();
    s.processEvolverBeans();
    return s;
  }

  /**
   * Gets the display name for this model.
   */
  public String displayName() {
    return getDisplayName();
  }

  /**
   * Performs post compile cleanup.
   */
  public void postCompile() {
    actions.postCompile();
    for (int i = 0; i < compileables.size(); i++) {
      EvolverCompileable c = (EvolverCompileable) compileables.get(i);
      c.postCompile();
    }
  }

  public void addPropertyChangeListener(PropertyChangeListener l) {
    super.addPropertyChangeListener(l);
    for (int i = 0; i < compileables.size(); i++) {
      EvolverCompileable c = (EvolverCompileable) compileables.get(i);
      c.addPropertyChangeListener(l);
    }
  }

  public void displayEditor(JFrame frame, AbstractCompilerException ex) {
    ActionsPropertyEditor editor = new ActionsPropertyEditor();
    editor.setValue(getActions());
    editor.addPropertyChangeListener(this);
    ActionsPropertyPanel panel = new ActionsPropertyPanel(actions);
    panel.addError(ex);
    panel.setVarNames(actions.getVarNames());
    panel.display(frame, editor);

    setActions((Actions) editor.getValue());
  }

  /**
   * Gets nqpy statements that become the first part of the compiled class's
   * _buildStatement. These statements are guaranteed to execute before any
   * agent added code executed in _build.
   */
  public String preBuildStmts() {
    return "";
  }

  /**
   * Gets nqpy statements that become the first part of the compiled class's
   * _displayStatement. These statements are guaranteed to execute before any
   * agent or component added code executed in _display.
   *
   * @return
   */
  public String preDisplayStmts() {
    return "";
  }

  /**
   * Gets nqpy statements that become the first part of the compiled class's
   * _setup statements. These statements are guaranteed to execute before any
   * agent or component added code executed in _setup.
   *
   * @return
   */
  public String preSetupStmts() {
    return "";
  }

  public void saveXML(BufferedWriter out) throws IOException {
    out.write(actions.toXML());
    out.newLine();
    out.write(parameters.toXML());
    out.newLine();
    EvSchedule sch = getMasterSchedule();
    out.newLine();
    out.write(sch.toXML());
    getSchedule();
    out.write(details.toXML());
    out.newLine();
    saveBeans(compileables, out);
    saveBeans(components, out);
  }

  private void saveBeans(ArrayList list, BufferedWriter out)
          throws IOException {
    for (int i = 0; i < list.size(); i++) {
      EvolverBean bean = (EvolverBean) list.get(i);
      bean.saveXML(out);
      out.newLine();
    }
  }


  public void loadFromXML(Element modelE) {
    setModelName(modelE.getAttribute("ModelName"));
    setDisplayName(modelE.getAttribute("DisplayName"));
    setActions(XMLLoader.getActions(modelE));
    setParameters(XMLLoader.getParameters(modelE));
    setSchedule(XMLLoader.getScheduleDetails(modelE));
    setMasterSchedule(XMLLoader.getEvSchedule(modelE));

    try {
      ArrayList beanList = XMLLoader.getChildBeans(modelE);
      for (int i = 0; i < beanList.size(); i++) {
        EvolverBean bean = (EvolverBean) beanList.get(i);
        bean.addToParent(this);
      }
    } catch (RepastException ex) {
      ex.printStackTrace();
    }

    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        producer.setInitialSuperClass();
      }
    }
  }

  public void propertyChange(PropertyChangeEvent evt) {
    if (evt.getSource() instanceof EvolverPropertyEditor) {
      EvolverPropertyEditor editor = (EvolverPropertyEditor) evt.getSource();
      editor.setPropertyFromValue(this);
    }

    if (evt.getSource() != null && evt.getSource() instanceof AgentProducer &&
            evt.getPropertyName().equals(PropertyName.CLASS_NAME.getName())) {
      // need to update the name in the possible parents for other 
      AgentProducer sourceProducer = (AgentProducer) evt.getSource();
      for (int i = 0, n = compileables.size(); i < n; i++) {
        Object o = compileables.get(i);
        if (o instanceof AgentProducer) {
          AgentProducer producer = (AgentProducer) o;
          producer.updatePossibleParent(sourceProducer, evt.getNewValue().toString());
        }
      }
    }
  }

  // checks that the compileables do not conflict with each other
  // and produce bad code
  protected void sanityCheck() {
    Set groupNames = new HashSet();
    Set classNames = new HashSet();

    for (int i = 0, n = components.size(); i < n; i++) {
      EvolverComponent component = (EvolverComponent) components.get(i);
      BooleanMessage result = component.sanityCheck();
      if (!result.isValid()) throw new RuntimeException(result.getMessage());
      Hashtable fields = component.fields();
      for (Iterator iter = fields.keySet().iterator(); iter.hasNext();) {
        Object key = iter.next();
        if (classNames.contains(key)) {
          throw new RuntimeException("Two or more components have the same name '" + key.toString() + "'");
        }
        classNames.add(key);
      }
    }


    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        String groupName = producer.getGroupName();
        String agentName = producer.className();
        if (groupNames.contains(groupName)) {
          throw new RuntimeException("Two or more agent producers have the same group name '" + groupName + "'");
        }
        if (classNames.contains(agentName)) {
          throw new RuntimeException("Two or more agent producers have the same name '" + agentName + "'");
        }
        groupNames.add(groupName);
        classNames.add(agentName);

        BooleanMessage valid = producer.isValid();
        if (!valid.isValid()) {
          throw new RuntimeException(valid.getMessage());
        }

        BooleanMessage result = producer.getParameters().sanityCheck(producer.className());
        if (!result.isValid()) throw new RuntimeException(result.getMessage());
      }
    }

    BooleanMessage result = parameters.sanityCheck(dispName);
    if (!result.isValid()) throw new RuntimeException(result.getMessage());
    if (!RepastPyUtilities.isValidJavaIdentifier(className)) {
      throw new RuntimeException("Invalid model name. Valid models names may only contain letters, numbers, or '_'");
    }
  }

  /*
   * Processes the compileables that are AgentProducers, adding their
   * group names as ivars to the produced model class together with
   * a get accessor.
   */
  protected void processAgentProducers(NQPyDynamicClass modelClass) {
    List producers = new ArrayList();
    Set parents = new HashSet();
    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        producers.add(producer);
        producer.setIsParent(false);
        String groupName = producer.getGroupName();
        modelClass.addField(groupName, DefaultGroup.class);
        String capName = Character.toUpperCase(groupName.charAt(0)) + groupName.substring(1);
        modelClass.addMethod("get" + capName, DefaultGroup.class, new Class[]{});

        AgentProducer parent = producer.getParent();
        if (parent != null) parents.add(parent);
      }
    }

    if (parents.size() > 0) {
      for (int i = 0, n = producers.size(); i < n; i++) {
        // set the producers as parents if necessary
        AgentProducer producer = (AgentProducer) producers.get(i);
        if (parents.contains(producer)) producer.setIsParent(true);
      }
    }
  }
}
