package uchicago.src.simbuilder.beans;

import java.io.Serializable;
import java.util.ArrayList;

/**
 * A *ScheduleUnit stores ScheduleDetails in some way and is responsible
 * for creating nqpy statements out of those ScheduleDetails.
 * AbstractScheduleUnit implements methods primarily for the creation of
 * rscript statements, that is, methods for turning collections of
 * ScheduleDetails into Actions. These Actions typically become _* methods
 * when the model is compiled. In the case of schedule Actions, they
 * are turned into a _schedule method as well as other methods
 * scheduled for execution in the _schedule method.
 * <p/>
 * $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:25 $
 */

public abstract class AbstractScheduleUnit implements Serializable {

  protected String name;
  protected Integer id;

  private String baseName = "_schedule";
  protected String xmlString = EvolverConstants.XMLNS + ":schedule";

  /**
   * Turns ScheduleDetails into Actions.
   */
  class Method implements Serializable {
    String name;
    int type;
    boolean last = false;
    long tick = 1;
    StringBuffer rscript = new StringBuffer();

    public Method(ScheduleDetail detail) {
      this.last = detail.getLast();
      this.tick = detail.getTick();
      type = detail.getType();
    }

    public void addRscript(String script) {
      rscript.append(script);
      rscript.append("\n");
    }

    public String getName(Actions actions) {
      if (name == null) {
        name = makeMethodName(actions);
      }
      return name;
    }

    public Actions addToSchedule(Actions actions) {
      StringBuffer stmt = new StringBuffer("self.schedule.scheduleAction");
      String name = getName(actions);

      switch (type) {
        case ScheduleDetails.EVERY_TICK:
          stmt.append("Beginning(");
          stmt.append(tick);
          stmt.append(", ");
          break;

        case ScheduleDetails.AT:
          stmt.append("At(");
          stmt.append(tick);
          stmt.append(", ");
          break;

        case ScheduleDetails.AT_INTERVAL:
          stmt.append("AtInterval(");
          stmt.append(tick);
          //stmt.append(")");
          stmt.append(", ");
          break;

        case ScheduleDetails.END:
          stmt.append("AtEnd(");
          break;

        case ScheduleDetails.PAUSE:
          stmt.append("AtPause(");
          break;

        default:
          throw new RuntimeException("invalid schedule type");
      }


      stmt.append("self, \"");
      stmt.append(name);
      stmt.append("\"");
      if (last) stmt.append(", Schedule.LAST");
      stmt.append(")\n");

      actions.addToAction("_schedule", stmt.toString());
      return actions;
    }

    public Actions makeMethod(Actions actions) {
      getName(actions);
      //System.out.println("rscript.toString() = " + rscript.toString());
      // hack!!!
      if (!rscript.toString().equals("null\n")) {
        actions.addToAction(name, rscript.toString());
      }
      return actions;
    }
  }

  public AbstractScheduleUnit(String name, Integer id) {
    this.name = name;
    this.id = id;
  }

  private String makeMethodName(Actions actions) {
    int index = 1;
    String name = baseName + String.valueOf(index);
    while (actions.hasAction(name)) {
      index++;
      name = baseName + String.valueOf(index);
    }

    return name;
  }

  /**
   * Turns the specified list of ScheduleDetails into Actions. The
   * list of ScheduleDetails is ActionCode objects and these ActionCode
   * objects are added to the specified Actions object. This new object
   * is returned.
   */
  protected Actions detailsToActions(ArrayList details, Actions actions) {
    Method normal = null;
    Method last = null;

    for (int i = 0; i < details.size(); i++) {
      ScheduleDetail d = (ScheduleDetail) details.get(i);
      if (d.getLast()) {
        if (last == null) last = new Method(d);
        last.addRscript(d.getNQPy());
      } else {
        if (normal == null) normal = new Method(d);
        normal.addRscript(d.getNQPy());
      }
    }

    // want to create one ActionCode objects in Actions for the RSscript
    // in normal and last, and add approprate schedule rscript to _schedule
    // in Actions.
    if (normal != null) {
      actions = normal.makeMethod(actions);
      actions = normal.addToSchedule(actions);
    }

    if (last != null) {
      actions = last.makeMethod(actions);
      actions = last.addToSchedule(actions);
    }

    return actions;
  }

  /**
   * Makes an Actions object out of the ScheduleDetails contained this
   * ScheduleUnit. The ScheduleDetails are turned into ActionCode objects.
   * These objects are added to the passed-in Actions object which is then
   * returned.
   */
  public abstract Actions makeActions(Actions actions);

  /**
   * Is this ScheduleUnit empty of ScheduleDetails.
   */
  public abstract boolean isEmpty();

  /**
   * Adds a ScheduleDetail to the ScheduleUnit.
   */
  public abstract void addScheduleDetail(ScheduleDetail detail);

  /**
   * Makes a SchTreeNode from this ScheduleUnit.
   */
  public abstract SchTreeNode makeNode();

  /**
   * Deletes the specified ScheduleDetail from this ScheduleUnit.
   */
  public abstract void delete(ScheduleDetail detail);


  /**
   * Returns an XML representation of this ScheduleUnit
   */
  public abstract String toXML();

  void showDetails() {
  }
}
