package uchicago.src.simbuilder.beans;

import java.io.Serializable;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;

/**
 * Stores the possible values of a property as well as the current value.
 * ContrainedProperty is used as a bean property that should be displayed
 * as a combo box. Either the possible objects themselves are displayed
 * in the combobox or the user specifiable String representation of those
 * objects. For example, the bean property X may have 3 possible int
 * values: 0, 1, 2. These values may be format flags for DL FORMAT,
 * ASCII FORMAT, and EXCEL FORMAT. So possible objects would be 3
 * Integers while the String representation of these would be the format
 * flags DL FORMAT, ASCII FORMAT, and EXCEL FORMAT.
 * <p/>
 * $Revision: 1.2 $ $Date: 2004/12/10 21:59:42 $
 */

public class ConstrainedProperty implements Serializable {

  private List elements = new ArrayList();
  private int selectedIndex;

  


  class Element implements Comparable  {
    Object possible;
    String view;

    public Element(Object possible, String view) {
      this.possible = possible;
      this.view = view;
    }
    
    public Element(Element element) {
      this.possible = element.possible;
      this.view = element.view;
    }

    public int compareTo(Object o) {
      Element other = (Element) o;
      return this.view.compareTo(other.view);
    }
  }


  /**
   * Creates a new ConstrainedProperty.
   *
   * @param possibles     the list of possible values for this property
   * @param selectedIndex the index of the currently selected value
   */
  public ConstrainedProperty(Object[] possibles, int selectedIndex) {
    for (int i = 0; i < possibles.length; i++) {
      elements.add(new Element(possibles[i], possibles[i].toString()));
    }
    this.selectedIndex = selectedIndex;
  }

  /**
   * Creates a new ConstrainedProperty.
   *
   * @param possibles     the list of possible values for this property
   * @param selectedIndex the index of the currently selected value
   * @param views         String representations of the possible values
   */
  public ConstrainedProperty(Object[] possibles, int selectedIndex,
                             String[] views) {
    for (int i = 0; i < possibles.length; i++) {
      Object possible = possibles[i];
      String view = views[i];
      elements.add(new Element(possible, view));
    }
    this.selectedIndex = selectedIndex;
  }
  

  /**
   * Creates a new ConstrainedProperty as copy of the specified
   * ConstrainedProperty.
   */
  public ConstrainedProperty(ConstrainedProperty prop) {
    for (int i = 0, n = prop.elements.size(); i < n; i++) {
      Element element = (Element) prop.elements.get(i);
      elements.add(new Element(element));
    }
    this.selectedIndex = prop.selectedIndex;
  }

  public void setSelected(Object o) {
    for (int i = 0, n = elements.size(); i < n; i++) {
      Element element = (Element) elements.get(i);
      if (element.possible.equals(o)) {
        selectedIndex = i;
        break;
      }
    }
  }

  public void setSelectedView(String view) {
    selectedIndex = 0;
    for (int i = 0, n = elements.size(); i < n; i++) {
      Element element = (Element) elements.get(i);
      if (element.view.equals(view)) {
        selectedIndex = i;
        break;
      }
    }
  }
  
  /**
   * Changes the view name for the specified possible.
   */
  public void edit(Object possible, String newView) {
    for (int i = 0, n = elements.size(); i < n; i++) {
      Element element = (Element) elements.get(i);
      if (element.possible.equals(possible)) {
        element.view = newView;
        break;
      }
    }
  }
  
  public void removePossible(AgentProducer producer) {
    for (Iterator iter = elements.iterator(); iter.hasNext(); ) {
      Element element = (Element) iter.next();
      if (element.possible.equals(producer)) { 
        iter.remove();
        break;
      }
    }
  }

  public Object getSelectedView() {
    Element element = (Element) elements.get(selectedIndex);
    return element.view;
  }

  /**
   * Sets the selected value as index of the possible values.
   */
  public void setSelected(int index) {
    if (index != -1) selectedIndex = index;
  }

  /**
   * Gets the currently selected object.
   */
  public Object getSelected() {
    Element element = (Element) elements.get(selectedIndex);
    return element.possible;
  }

  /**
   * Gets the index of the currently selected item.
   */
  public int getSelectedIndex() {
    return selectedIndex;
  }
  
  public void add(String view, Object possible) {
    elements.add(new Element(possible, view));
  }
  
  public void sort() {
    Object obj = getSelected();
    Collections.sort(elements);
    setSelected(obj);
  }

  /**
   * Gets a the views of this object. If no String representations were
   * specified in the constructor this returns the possible objects
   * themselves.
   */
  public Object[] getViews() {
    Object[] views = new Object[elements.size()];
    for (int i = 0, n = elements.size(); i < n; i++) {
      Element element = (Element) elements.get(i);
      views[i] = element.view;
    }
    return views;
  }
}
