package uchicago.src.simbuilder.beans;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

/**
 * Stores ScheduleDetails in *ScheduleUnits, allowing for the manipulation
 * of the ScheduleDetails relative to one another. Turns ScheduleDetails
 * into the appropriate Actions via their *ScheduleUnits. This the type
 * of the so-called master schedule in an EvolverModel that displays
 * how all the individual scheduled actions in the model relate to
 * one another.
 * <p/>
 * $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:26 $
 */

public class EvSchedule implements Serializable {

  private ArrayList beans = new ArrayList();
  // key is type (EVERY etc.), val is HashMap where
  // key is tick, and val is arrayList of ScheduleDetails
  private HashMap masterMap = new HashMap();

  private final Integer EVERY = new Integer(ScheduleDataConstants.EVERY_TICK);
  private final Integer AT = new Integer(ScheduleDataConstants.AT);
  private final Integer INTERVAL = new Integer(ScheduleDataConstants.AT_INTERVAL);
  private final Integer PAUSE = new Integer(ScheduleDataConstants.PAUSE);
  private final Integer END = new Integer(ScheduleDataConstants.END);

  public EvSchedule() {
    masterMap.put(EVERY, new MapScheduleUnit("EVERY TICK", EVERY));
    masterMap.put(AT, new MapScheduleUnit("AT", AT));
    masterMap.put(INTERVAL, new MapScheduleUnit("AT INTERVAL", INTERVAL));
    masterMap.put(PAUSE, new ListScheduleUnit("PAUSE", PAUSE));
    masterMap.put(END, new ListScheduleUnit("END", END));
  }

  public EvSchedule copy() {
    EvSchedule copy = new EvSchedule();
    copy.masterMap.put(EVERY, ((MapScheduleUnit) masterMap.get(EVERY)).copy(copy));
    copy.masterMap.put(AT, ((MapScheduleUnit) masterMap.get(AT)).copy(copy));
    copy.masterMap.put(INTERVAL, ((MapScheduleUnit) masterMap.get(INTERVAL)).copy(copy));
    copy.masterMap.put(PAUSE, ((ListScheduleUnit) masterMap.get(PAUSE)).copy(copy));
    copy.masterMap.put(END, ((ListScheduleUnit) masterMap.get(END)).copy(copy));
    copy.beans = beans;

    return copy;
  }

  /**
   * Sets the list of all the EvolverBean in the current model to
   * the specific val.
   */
  public void setEvolverBeans(ArrayList val) {
    beans = val;
  }

  /**
   * Adds the ScheduleDetails from the EvolverBeans to this EvSchedule.
   * Calls schedule() on each EvolverBean.
   */
  public void processEvolverBeans() {
    for (int i = 0; i < beans.size(); i++) {
      EvolverBean bean = (EvolverBean) beans.get(i);
      ScheduleDetails dets = bean.schedule();
      ArrayList deletedDetails = dets.getDeletedDetails();
      for (int j = 0, n = deletedDetails.size(); j < n; j++) {
        delete((ScheduleDetail) deletedDetails.get(j));
      }
      dets.clearDeletedDetails();
      addScheduleDetails(dets);
    }
  }

  /**
   * Deletes the specified ScheduleDetail from this EvSchedule. This
   * will also delete it from the list of ScheduleDetails in the
   * appropriate EvolverBean.
   */
  public void delete(ScheduleDetail detail) {
    AbstractScheduleUnit unit =
            (AbstractScheduleUnit) masterMap.get(new Integer(detail.getType()));
    unit.delete(detail);
  }

  /**
   * Turns the ScheduleDetails in to SchTreeNodes for display in the
   * EvScheduleEditorPanel.
   */
  public SchTreeNode makeNode() {
    SchTreeNode root = new SchTreeNode("root", new Integer(0), true);
    root.add(((AbstractScheduleUnit) masterMap.get(EVERY)).makeNode());
    root.add(((AbstractScheduleUnit) masterMap.get(INTERVAL)).makeNode());
    root.add(((AbstractScheduleUnit) masterMap.get(AT)).makeNode());
    root.add(((AbstractScheduleUnit) masterMap.get(PAUSE)).makeNode());
    root.add(((AbstractScheduleUnit) masterMap.get(END)).makeNode());

    return root;
  }

  /**
   * Clears this EvSchedule of ScheduleDetails.
   */
  public void clear() {
    masterMap.clear();
    masterMap.put(EVERY, new MapScheduleUnit("EVERY TICK", EVERY));
    masterMap.put(AT, new MapScheduleUnit("AT", AT));
    masterMap.put(INTERVAL, new MapScheduleUnit("AT INTERVAL", INTERVAL));
    masterMap.put(PAUSE, new ListScheduleUnit("PAUSE", PAUSE));
    masterMap.put(END, new ListScheduleUnit("END", END));
  }

  /**
   * Adds the specified ScheduleDetail to this EvSchedule.
   */
  public void addScheduleDetail(ScheduleDetail detail) {
    AbstractScheduleUnit unit =
            (AbstractScheduleUnit) masterMap.get(new Integer(detail.getType()));
    unit.addScheduleDetail(detail);
  }

  /**
   * Adds the ScheduleDetail(s) in the specified ScheduleDetails object
   * to this EvSchedule.
   */
  public void addScheduleDetails(ScheduleDetails details) {

    for (int i = 0; i < details.size(); i++) {
      ScheduleDetail d = details.getDetail(i);
      d.setEvSchedule(this);
      AbstractScheduleUnit unit =
              (AbstractScheduleUnit) masterMap.get(new Integer(d.getType()));
      unit.addScheduleDetail(d);
    }
  }

  public String toXML() {

    // assigns an id to each ScheduleDetail. This id is used to
    // track order the ScheduleDetail-s.
    int id = 0;
    for (int i = 0; i < beans.size(); i++) {
      EvolverBean bean = (EvolverBean) beans.get(i);
      ScheduleDetails dets = bean.schedule();
      id = dets.setID(id);
      addScheduleDetails(dets);
    }

    StringBuffer b = new StringBuffer();

    AbstractScheduleUnit unit = (AbstractScheduleUnit) masterMap.get(EVERY);
    b.append(unit.toXML());

    unit = (AbstractScheduleUnit) masterMap.get(INTERVAL);
    b.append(unit.toXML());

    unit = (AbstractScheduleUnit) masterMap.get(AT);
    b.append(unit.toXML());

    unit = (AbstractScheduleUnit) masterMap.get(PAUSE);
    b.append(unit.toXML());

    unit = (AbstractScheduleUnit) masterMap.get(END);
    b.append(unit.toXML());

    return b.toString();
  }

  /**
   * Creates ActionCode objects from the ScheduleDetails. These
   * ActionCode objects are then compiled as methods in the class
   * compiled from the current EvolverModel. These methods are then
   * schedule for execution in the _schedule method in the class
   * compiled from the current EvolverModel.
   * <p/>
   * Initial implementation is to create individual ActionCode objects for
   * each cluster of ScheduleDetails. ScheduleDetails are clustered by type
   * and then by tick. So all the nqpy scheduled to be executed every tick
   * beginning at tick 3 will be encapsulated in an ActionCode object. This
   * object will then become some method (e.g. "_schedule3") in the model.
   */
  public Actions createActions(Actions actions) {
    //System.out.println(actions.toXML());

    Iterator i = masterMap.keySet().iterator();
    while (i.hasNext()) {
      Integer key = (Integer) i.next();
      AbstractScheduleUnit unit = (AbstractScheduleUnit) masterMap.get(key);
      actions = unit.makeActions(actions);
    }
    return actions;
  }

  public void showDetails() {
    Iterator i = masterMap.keySet().iterator();
    while (i.hasNext()) {
      Integer key = (Integer) i.next();
      AbstractScheduleUnit unit = (AbstractScheduleUnit) masterMap.get(key);
      unit.showDetails();
    }
  }

}
