package uchicago.src.simbuilder.beans;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.simbuilder.EvolverGui;
import uchicago.src.nqpy.compiler.NQPyClassFactory;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.sim.util.RepastException;
import uchicago.src.sim.engine.AbstractGUIController;
import uchicago.src.xml.XMLFactory;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyEditorManager;
import java.beans.PropertyChangeEvent;
import java.io.*;
import java.util.*;

/**
 * The top level EvolverBean. All other EvolverBeans are added to
 * a project. This is not semantically an EvolverBean in the sense
 * that it doesn't produce RePast objects. It does act as the entry
 * point for all actions on the project tree, including compilation
 * and so forth.
 * <p/>
 * $Revision: 1.4 $ $Date: 2004/11/18 21:52:17 $
 */

public class EvolverProject implements EvolverParent, EvolverConstants, PropertyChangeListener {

  private EvolverModel model;
  private String projDir = "output";
  private String classPath = "";
  private String projName = "Environment";
  private String packageName = "default_package";
  protected PropertyChangeSupport pcs;
  private String[] xmlPropNames = {"ProjectDirectory", "PackageName", "ProjectName"};
  private Description description = new Description();
  private boolean outputToConsole = true;
  private boolean errorsToConsole = true;

  static {
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.Description.class,
                    uchicago.src.simbuilder.beans.DescriptionEditor.class);
  }

  public EvolverProject() {
    createProjDir("");
  }

  private void createProjDir(String oldDir) {
    File f = new File(".");
    File[] roots = File.listRoots();
    boolean rootStart = false;
    for (int i = 0; i < roots.length; i++) {
      String root = roots[i].getAbsolutePath();
      if (projDir.startsWith(root)) {
        rootStart = true;
        break;
      }
    }
    if (projDir.startsWith(File.separator)) {
      // create the project dir starting with root
      File root = File.listRoots()[0];
      f = new File(root.getAbsolutePath() + projDir);
    } else if (rootStart) {
      f = new File(projDir);
    } else {
      String tmp = f.getAbsolutePath();
      tmp = tmp.substring(0, tmp.length() - 1) + projDir;
      f = new File(tmp);
    }
    if (f.exists()) {
      if (!f.isDirectory()) {
        f.mkdir();
      }
    } else {
      f.mkdir();
    }

    try {
      if (!f.exists()) {
        // couldn't create f for some reason
        throw new IOException("Unable to create directory: " + f.getCanonicalPath());
      }
      projDir = new File(projDir).getCanonicalPath();
    } catch (IOException e) {
      if (oldDir.length() > 0) projDir = oldDir;
      EvolverGui.showError("Error creating project directory", e);
      e.printStackTrace();
    }
    
    pcs = new PropertyChangeSupport(this);
  }

  public Description getDescription() {
    return description;
  }

  public void setDescription(Description description) {
    this.description = description;
  }

  public String getProjectDirectory() {
    return projDir;
  }

  public void setProjectDirectory(String dir) {
    String oldProjDir = projDir;
    projDir = dir;
    File file = new File(projDir);
    if (!file.exists() || !file.isDirectory())
      createProjDir(oldProjDir);
    else {
      try {
        projDir = new File(projDir).getCanonicalPath();
      } catch (IOException e) {
        projDir = oldProjDir;
        EvolverGui.showError("Error creating project directory", e);
        e.printStackTrace();
      }
    }
  }

  public String getClassPath() {
    return classPath;
  }

  /**
   * Paths should be separated by semi-colon (windows style).
   *
   * @param classPath
   */
  public void setClassPath(String classPath) {
    if (classPath != null) this.classPath = classPath;
  }

  public boolean isErrorsToConsole() {
    return errorsToConsole;
  }

  public void setErrorsToConsole(boolean errorsToConsole) {
    this.errorsToConsole = errorsToConsole;
    AbstractGUIController.CONSOLE_ERR = errorsToConsole;
  }

  public boolean isOutputToConsole() {
    return outputToConsole;
  }

  public void setOutputToConsole(boolean outputToConsole) {
    this.outputToConsole = outputToConsole;
    AbstractGUIController.CONSOLE_OUT = outputToConsole;
  }

  /**
   * Returns the class path property parsed into a list of individual paths.
   *
   * @return
   */
  public ArrayList getParsedClassPath() {
    ArrayList list = new ArrayList();
    StringTokenizer tok = new StringTokenizer(classPath.trim(), ";");
    while (tok.hasMoreTokens()) {
      list.add(tok.nextToken().trim());
    }

    return list;
  }


  public String getProjectName() {
    return projName;
  }

  public void setProjectName(String val) {
    String oldName = projName;
    projName = val;
    pcs.firePropertyChange(PropertyName.DISPLAY_NAME.getName(), oldName, val);
  }

  /**
   * Adds the specified component to this EvolverParent
   */
  public void addComponent(EvolverComponent child) throws RepastException {
  }

  /**
   * Adds the specified compileable to this EvolverParent
   */
  public void addCompileable(EvolverCompileable child) throws RepastException {
    if (!(child instanceof EvolverModel)) {
      throw new RepastException("Only model objects can be added directly to projects");
    } else if (model != null) {
      throw new RepastException("You must delete the current model before adding a new one");
    } else {
      model = (EvolverModel) child;
    }
  }

  /**
   * Removes the specified component to this EvolverParent
   */
  public void removeComponent(EvolverComponent child) {
  }

  /**
   * Removes the specified compileable to this EvolverParent
   */
  public void removeCompileable(EvolverCompileable child) {
    if (child.equals(model)) {
      model = null;
    }
  }

  /**
   * Throws an exception if called. EvolverProject should be the top most
   * object in the tree.
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    throw new RepastException("Environment cannot be a child");
  }

  /**
   * Doesn't do anything as this is the top most object in the project tree.
   */
  public void removeFromParent(EvolverParent parent) {
  }

  /**
   * Returns null as this is not a true EvolverBean.
   */
  public ScheduleDetails schedule() {
    return null;
  }

  /**
   * Returns an ArrayList of this EvolverProjects' children. This list
   * should have a single member, the EvolverModel for this project.
   */
  public ArrayList children() {
    ArrayList list = new ArrayList();
    if (model != null) list.add(model);
    return list;
  }

  /**
   * Compiles all the members of the project tree of which this EvolverProject
   * is the top.
   */
  public void compile(ByteCodeGenerator generator) throws RepastException {
    if (model == null) throw new RepastException("Nothing to compile!");
    String pkg = packageName.trim().toLowerCase().replace(' ', '_');
    pkg = pkg.replace('\'', '_');
    pkg = pkg.replace('\"', '_');
    String pkgDir = projDir + File.separator + pkg + File.separator;
    File f = new File(pkgDir);
    if (!f.exists()) {
      f.mkdir();
    }
    compile(pkgDir, generator);
  }

  /**
   * Performs post compilation clean up.
   */
  public void postCompile() {
    if (model != null) model.postCompile();
  }

  /**
   * Gets the class name of the model in this project.
   */
  public String modelName() {
    if (model == null) return null;
    return model.className();
  }

  /**
   * Builds Type objects for all the EvolverBeans etc. in this project.
   */
  public Hashtable buildType(Hashtable h, ByteCodeGenerator generator) {
    h = model.buildType(h, generator);
    return h;
  }


  /**
   * Compiles all the members of this EvolverProject.
   */
  public void compile(String dir, ByteCodeGenerator generator) {
    String pkg = packageName.trim().toLowerCase().replace(' ', '_');
    // recursively sets the package for model components.
    model.setPackage(pkg);
    generator.putDetail(EvolverBean.MODEL_NAME, pkg + "." + modelName());
    Hashtable h = buildType(new Hashtable(), generator);
    generator.putDetail(EvolverBean.DYN_TYPE_LIST, h);
    Enumeration e = h.keys();
    while (e.hasMoreElements()) {
      String name = (String) e.nextElement();
      NQPyDynamicClass dc = (NQPyDynamicClass) h.get(name);
      NQPyClassFactory.addDynamicType(dc.getJClassName(), dc);
    }

    model.compile(dir, generator);
  }

  public void addPropertyChangeListener(PropertyChangeListener l) {
    pcs.addPropertyChangeListener(l);
    if (model != null) model.addPropertyChangeListener(l);
  }

  public void removePropertyChangeListener(PropertyChangeListener l) {
    pcs.removePropertyChangeListener(l);
  }

  private String toXML(XMLFactory fac) {

    try {
      return fac.getXMLStartAttrib();
    } catch (Exception ex) {
      ex.printStackTrace();
    }
    return "";
  }

  public void saveXML(BufferedWriter out) throws IOException {
    XMLFactory fac = new XMLFactory(XMLNS + ":EvolverProject", this,
            new ArrayList(Arrays.asList(xmlPropNames)));

    out.write(XML_HEADER);
    out.newLine();
    out.newLine();
    out.write("<!-- Evolver Model Specification File -->");
    out.newLine();
    out.write(XML_ROOT_START);
    out.newLine();
    out.write(toXML(fac));
    out.write("ClassPath=\"" + classPath + "\" ");
    out.write("consoleErr=\"");
    out.write(String.valueOf(isErrorsToConsole()));
    out.write("\" consoleOut=\"");
    out.write(String.valueOf(isOutputToConsole()));
    out.write("\" >\n");
    out.write("<Description>");
    out.write("\n<![CDATA[\n");
    out.write(description.getText().trim());
    out.write("\n]]>\n</Description>");
    out.newLine();

    if (model != null) model.saveXML(out);

    out.write(fac.getXMLEnd());
    out.newLine();
    out.write(XML_ROOT_END);
    out.flush();
    out.close();
  }

  public void loadFromXML(Element proj) {
    File file = new File(proj.getAttribute("ProjectDirectory"));
    if (!file.exists()) file = new File("./output");
    try {
      if (!file.exists()) file.createNewFile();
      setProjectDirectory(file.getCanonicalPath());
    } catch (IOException ex) {
      ex.printStackTrace();
    }
    setProjectName(proj.getAttribute("ProjectName"));
    setPackageName(proj.getAttribute("PackageName"));
    setClassPath(proj.getAttribute("ClassPath"));
    String consoleErr = proj.getAttribute("consoleErr");
    boolean errs = Boolean.valueOf(consoleErr).booleanValue();
    if (consoleErr == null || consoleErr.length() == 0) errs = true;
    setErrorsToConsole(errs);
    String consoleOut = proj.getAttribute("consoleOut");
    boolean out = Boolean.valueOf(consoleOut).booleanValue();
    if (consoleOut == null || consoleOut.length() == 0) out = true;
    setOutputToConsole(out);
    NodeList list = proj.getElementsByTagName("Description");
    for (int i = 0; i < list.getLength(); i++) {
      Node n = list.item(i);
      setDescription(new Description(n.getChildNodes().item(1).getNodeValue().trim()));
    }

    model = XMLLoader.getModel(proj);
  }

  public EvolverModel getModel() {
    return model;
  }

  public String getPackageName() {
    return packageName;
  }

  public void setPackageName(String packageName) {
    this.packageName = packageName;
  }

  public void saveXMLToString(StringWriter writer) {
    BufferedWriter out = null;

    try {
      out = new BufferedWriter(writer);
      saveXML(out);
    } catch (IOException ex) {
      ex.printStackTrace();
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  public void propertyChange(PropertyChangeEvent evt) {}

  public void saveXMLToFile(String fileName) throws IOException {
    BufferedWriter out = null;
    PrintWriter writer = null;

    try {
      out = new BufferedWriter(new FileWriter(fileName));
      saveXML(out);
      // note that out is closed in saveXML
     
      if (model != null) {
        String pfFile = fileName.substring(0, fileName.length() - 3) + "pf";
        writer = new PrintWriter(new FileOutputStream(pfFile));
        model.getParameters().writePropertiesFile(writer);
        writer.flush();
        writer.close();
      }

    } finally {
      if (writer != null) writer.close();
    }
  }
}
