package uchicago.src.simbuilder.beans;

import org.w3c.dom.*;
import uchicago.src.simbuilder.beans.network.LayoutProperty;
import uchicago.src.simbuilder.beans.network.MatrixFileNetworkSource;
import uchicago.src.simbuilder.beans.network.NetworkSource;
import uchicago.src.simbuilder.beans.network.RandomDensityNetworkSource;
import uchicago.src.simbuilder.beans.network.SquareLatticeNetworkSource;
import uchicago.src.simbuilder.beans.network.UnlinkedNodeNetworkSource;
import uchicago.src.simbuilder.beans.network.WattsStrogatzNetworkSource;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

/**
 * Creates an simbuilder project from an xml specification file.
 *
 * @version $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:27 $
 */

public class XMLLoader {

  private static final String MATRIX_SOURCE = "uchicago.src.simbuilder.beans.network.MatrixFileNetworkSource";
  private static final String RANDOM_DENSITY = "uchicago.src.simbuilder.beans.network.RandomDensityNetworkSource";
  private static final String WATTS_STROGATZ = "uchicago.src.simbuilder.beans.network.WattsStrogatzNetworkSource";
  private static final String SQUARE_LATTICE = "uchicago.src.simbuilder.beans.network.SquareLatticeNetworkSource";
  private static final String UNLINKED_NODES = "uchicago.src.simbuilder.beans.network.UnlinkedNodeNetworkSource";

  private DocumentBuilder parser;
  Document doc;
  private static HashMap detailsMap = new HashMap();
  private static ClassLoader loader;

  public XMLLoader(String fileName) throws IOException {
    FileInputStream file = null;
    try {
      DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
      parser = dbf.newDocumentBuilder();

      file = new FileInputStream(fileName);
      doc = parser.parse(file, fileName);
    } catch (Exception ex) {
      throw new IOException(ex.getMessage());
    } finally {
      if (file != null) file.close();
    }
  }

  public EvolverProject load(ClassLoader loader) {
    if (loader == null)
      XMLLoader.loader = this.getClass().getClassLoader();
    else
      XMLLoader.loader = loader;

    Element root = doc.getDocumentElement();
    NodeList children = root.getChildNodes();
    EvolverProject proj = null;
    for (int i = 0; i < children.getLength(); i++) {
      Node child = children.item(i);
      if (child.getNodeType() == Node.ELEMENT_NODE) {
        proj = makeProject((Element) child);
        break;
      }
    }

    return proj;
  }

  public static Element getNextElement(Element elem) {
    NodeList children = elem.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
      Node child = children.item(i);
      if (child.getNodeType() == Node.ELEMENT_NODE) {
        return (Element) child;
      }
    }

    return null;
  }

  public static ScheduleDetails getScheduleDetails(Element elem) {
    ScheduleDetails sch = new ScheduleDetails();
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":" + "schedule_item");
    for (int i = 0; i < list.size(); i++) {
      Integer id = new Integer(((Element) list.get(i)).getAttribute("id"));
      sch.addDetail((ScheduleDetail) detailsMap.get(id));
    }

    return sch;
  }

  public static EvSchedule getEvSchedule(Element elem) {
    EvSchedule sch = new EvSchedule();
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":" + "schedule");
    for (int i = 0; i < list.size(); i++) {
      String order = ((Element) list.get(i)).getAttribute("order");
      StringTokenizer t = new StringTokenizer(order, " ");
      while (t.hasMoreTokens()) {
        Integer id = new Integer(t.nextToken());
        sch.addScheduleDetail((ScheduleDetail) detailsMap.get(id));
      }
    }

    return sch;

  }

  public static DataSourceActions getDSActions(Element elem) {
    DataSourceActions actions = new DataSourceActions();
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":" + "dsAction");
    for (int i = 0; i < list.size(); i++) {
      Element action = (Element) list.get(i);
      DataSourceActionCode code = new DataSourceActionCode(action.getAttribute("name"));
      NodeList children = action.getChildNodes();
      String rscode = children.item(1).getNodeValue();
      code.setCode(rscode.trim());
      actions.addDataSource(code);

      ArrayList l = new ArrayList();
      StringTokenizer t = new StringTokenizer(action.getAttribute("imports"),
              " ");
      while (t.hasMoreTokens()) {
        l.add(t.nextToken());
      }
      code.setImports(l);

      l = new ArrayList();
      t = new StringTokenizer(action.getAttribute("paramNames"), " ");
      while (t.hasMoreTokens()) {
        l.add(t.nextToken());
      }
      code.setParameterNames(l);
      code.setRetType(action.getAttribute("retVal"));

      code.setMaxFracDigits(Integer.valueOf(action.getAttribute("maxFracDigits")).intValue());
      code.setMaxIntDigits(Integer.valueOf(action.getAttribute("maxIntDigits")).intValue());
      code.setType(Integer.valueOf(action.getAttribute("type")).intValue());
    }

    return actions;
  }

  public static Actions getActions(Element elem) {
    Actions actions = new Actions();
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":" + "action");
    for (int i = 0; i < list.size(); i++) {
      Element action = (Element) list.get(i);
      ActionCode code = new ActionCode(action.getAttribute("name"));
      NodeList children = action.getChildNodes();
      String rscode = children.item(1).getNodeValue();
      code.setCode(rscode.trim());
      actions.addAction(code);

      ArrayList l = new ArrayList();
      StringTokenizer t = new StringTokenizer(action.getAttribute("imports"),
              " ");
      while (t.hasMoreTokens()) {
        l.add(t.nextToken());
      }
      code.setImports(l);
      
      String retType = action.getAttribute("retType");
      if (retType != null) {
        code.setRetType(retType);
      }

      l = new ArrayList();
      t = new StringTokenizer(action.getAttribute("paramNames"),
              " ");
      while (t.hasMoreTokens()) {
        l.add(t.nextToken());
      }
      code.setParameterNames(l);
    }


    return actions;
  }

  public static Fields getParameters(Element elem) {
    Fields ps = new Fields();
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":" + "parameter");
    for (int i = 0; i < list.size(); i++) {
      Element e = (Element) list.get(i);
      String name = e.getAttribute("name");
      String type = e.getAttribute("type");
      String defVal = e.getAttribute("defaultValue");
      String accessible = e.getAttribute("accessible");

      boolean probe = Boolean.valueOf(e.getAttribute("probe")).booleanValue();
      if (accessible == null || accessible.trim().length() == 0) {
        // if accessible attribute doesn't exist then this is a version 1.0
        // sbp file and were accessible was implicitly true.
        ps.addField(name, type, defVal, probe, true);
      } else {
        boolean a = Boolean.valueOf(accessible).booleanValue();
        ps.addField(name, type, defVal, probe, a);
      }
    }

    return ps;

  }

  public static ArrayList getChildBeans(Element elem) {
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":EvolverBean");
    ArrayList beanList = new ArrayList(list.size());
    for (int i = 0; i < list.size(); i++) {
      Element e = (Element) list.get(i);
      String beanClass = e.getAttribute("class");
      beanClass = beanClass.substring(beanClass.indexOf(" ") + 1,
              beanClass.length());

      EvolverBean bean = (EvolverBean) createObject(beanClass);
      bean.loadFromXML(e);
      beanList.add(bean);
    }

    return beanList;
  }

  public static LayoutProperty getGraphLayout(Element elem) {
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":network_display");
    LayoutProperty l = null;

    if (list.size() > 0) {
      Element ns = (Element) list.get(0);
      String cls = ns.getAttribute("class");
      l = (uchicago.src.simbuilder.beans.network.LayoutProperty) createObject(cls);
      l.fromMap(createMap(ns));
    }

    return l;
  }

  public static NetworkSource getNetworkSource(Element elem) {
    ArrayList list = getChildElementsByName(elem, EvolverConstants.XMLNS +
            ":network_source");
    if (list.size() > 0) {
      Element ns = (Element) list.get(0);
      String cls = ns.getAttribute("class");
      Object o = createObject(cls);
      if (cls.equals(MATRIX_SOURCE)) {
        MatrixFileNetworkSource source = (MatrixFileNetworkSource) o;
        source.setFileName(ns.getAttribute("fileName"));
        source.setType(Integer.valueOf(ns.getAttribute("type")).intValue());
        return source;
      } else if (cls.equals(RANDOM_DENSITY)) {
        RandomDensityNetworkSource source = (RandomDensityNetworkSource) o;
        source.setSize(Integer.valueOf(ns.getAttribute("size")).intValue());
        source.setDensity(Double.valueOf(ns.getAttribute("density")).
                doubleValue());
        source.setAllowLoops(Boolean.valueOf(ns.getAttribute("allowLoops")).
                booleanValue());
        source.setSymmetric(Boolean.valueOf(ns.getAttribute("symmetric")).
                booleanValue());
        return source;
      } else if (cls.equals(SQUARE_LATTICE)) {
        SquareLatticeNetworkSource source = (SquareLatticeNetworkSource) o;
        source.setCols(Integer.valueOf(ns.getAttribute("cols")).intValue());
        source.setRows(Integer.valueOf(ns.getAttribute("rows")).intValue());
        source.setConnectRadius(Integer.
                valueOf(ns.getAttribute("connectRadius")).
                intValue());

        source.setWrapAround(Boolean.valueOf(ns.getAttribute("wrapAround")).
                booleanValue());
        return source;
      } else if (cls.equals(WATTS_STROGATZ)) {
        WattsStrogatzNetworkSource source = (WattsStrogatzNetworkSource) o;
        source.setSize(Integer.valueOf(ns.getAttribute("size")).intValue());
        source.setConnectRadius(Integer.
                valueOf(ns.getAttribute("connectRadius")).
                intValue());
        source.setRewireProb(Double.valueOf(ns.getAttribute("rewireProb")).
                doubleValue());
        return source;
      } else if (cls.equals(UNLINKED_NODES)) {
        UnlinkedNodeNetworkSource source = (UnlinkedNodeNetworkSource) o;
        source.setNumberOfNodes(Integer.valueOf(ns.
                getAttribute("numberOfNodes")).
                intValue());
        return source;
      }
    }

    return null;
  }

  public static ArrayList getChildElementsByName(Element elem, String name) {
    NodeList children = elem.getChildNodes();
    ArrayList list = new ArrayList();
    for (int i = 0; i < children.getLength(); i++) {
      Node n = children.item(i);
      if (n.getNodeType() == Node.ELEMENT_NODE) {
        if (n.getNodeName().equals(name)) {
          list.add(n);
        }
      }
    }

    return list;
  }


  public static EvolverModel getModel(Element projElement) {
    Element modelE = (Element) projElement.getElementsByTagName(EvolverConstants.XMLNS + ":EvolverModel").item(0);
    if (modelE != null) {
      //System.out.println(modelE);

      String className = modelE.getAttribute("Class");
      // remove "class " prefix
      className = className.substring(className.indexOf(" ") + 1,
              className.length());
      EvolverModel model = (EvolverModel) createObject(className);
      model.loadFromXML(modelE);
      return model;
    }

    return null;
  }

  private static Object createObject(String name) {
    Object o = null;
    try {
      Class clazz = Class.forName(name, true, loader);
      Constructor con = clazz.getConstructor(new Class[]{});
      o = con.newInstance(new Object[]{});
    } catch (ClassNotFoundException ex) {
      ex.printStackTrace();
    } catch (NoSuchMethodException ex) {
      ex.printStackTrace();
    } catch (InstantiationException ex) {
      ex.printStackTrace();
    } catch (IllegalAccessException ex) {
      ex.printStackTrace();
    } catch (InvocationTargetException ex) {
      ex.printStackTrace();
    }

    return o;
  }


  private EvolverProject makeProject(Element projE) {
    fillHashMap(projE);
    EvolverProject proj = new EvolverProject();
    proj.loadFromXML(projE);

    //System.out.println(proj.getProjectDirectory());
    //System.out.println(proj.getProjectName());

    return proj;
  }

  private void fillHashMap(Element e) {
    NodeList list = e.getElementsByTagName(EvolverConstants.XMLNS + ":" +
            "schedule_item");
    for (int i = 0; i < list.getLength(); i++) {
      Element detail = (Element) list.item(i);
      String action = detail.getAttribute("action");
      int type = Integer.valueOf(detail.getAttribute("type")).intValue();
      long tick = Long.valueOf(detail.getAttribute("tick")).longValue();
      boolean last = Boolean.valueOf(detail.getAttribute("last")).
              booleanValue();
      Integer id = new Integer(detail.getAttribute("id"));
      String target = detail.getAttribute("target");
      ScheduleDetail d = new ScheduleDetail(action, tick, type, last);
      d.setTarget(target);
      detailsMap.put(id, d);
    }
  }

  public static int getIntVal(Element e, String attribName) {
    return Integer.valueOf(e.getAttribute(attribName)).intValue();
  }

  public static boolean getBooleanVal(Element e, String attribName) {
    return Boolean.valueOf(e.getAttribute(attribName)).booleanValue();
  }

  public static double getDoubleVal(Element e, String attribName) {
    return Double.valueOf(e.getAttribute(attribName)).doubleValue();
  }

  public static HashMap createMap(Element e) {
    NamedNodeMap m = e.getAttributes();
    HashMap map = new HashMap();
    for (int i = 0; i < m.getLength(); i++) {
      Node n = m.item(i);
      String name = n.getNodeName();
      map.put(name, e.getAttribute(name));
    }

    return map;
  }
}
