package uchicago.src.simbuilder.beans.gis;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.w3c.dom.Element;
import uchicago.src.nqpy.compiler.CompilerException;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.simbuilder.EvolverModelCompiler;
import uchicago.src.simbuilder.base.gis.GISSimModelImpl;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.xml.XMLFactory;

import java.beans.PropertyChangeSupport;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

/**
 * A default model EvolverModel for network simulations constructed in
 * Evolver. This loads the agents from a dl format file. Most of the
 * EvovlerModel functionality is provided in the base class,
 * AbstractEvolverModel.
 * <p/>
 * $Revision: 1.4 $ $Date: 2004/12/10 22:04:44 $
 */

public class DefaultGISModelProducer extends AbstractEvolverModel {

  private ConstrainedProperty gisPackage;

  public DefaultGISModelProducer() {
    className = "GISModel";
    dispName = "GIS Model";
    pcs = new PropertyChangeSupport(this);
    setParameters(parameters);

    GISPackage[] packages = {GISPackage.OPEN_MAP, GISPackage.ARCGIS};
    String[] packagesViews = {packages[0].toString(), packages[1].toString()};
    gisPackage = new ConstrainedProperty(packages, 0, packagesViews);

    actions.createDefaultAction("initAgents", "");
    actions.createDefaultAction("updateDisplay", "self.updateGISDisplay()\n");
    actions.createDefaultAction("writeAgents", "#self.writeAgents(your_agentGroup_here, your_shapefile_here)");
    
    /*
				"for agent as " + EvolverBean.AGENT_UNQ_NAME_TEMPLATE + " in " +
				"self.agentList:\n  agent.setModel(self)\n" +
				"  agent.init()\n");
        */
  }

  /**
   * Returns nqpy statements that become part of the _build method in
   * the class compiled from this EvolverModel.
   */
  public String buildStmts() {
    StringBuffer stmt = new StringBuffer();
    if (compileables.size() > 0) {
      stmt.append("self.initAgents()\n");
    }

    return stmt.toString();
  }

  /**
   * Gets nqpy statements that become the first part of the compiled class's
   * _displayStatement. These statements are guaranteed to execute before any
   * agent or component added code executed in _display.
   *
   * @return
   */
  public String preDisplayStmts() {
    GISPackage gPackage = (GISPackage) gisPackage.getSelected();
    String disp = gPackage.getDisplaySetupStatements(getDisplayName() + " Display");
    VelocityContext context = new VelocityContext();
    context.put("displaySetup", disp + "\n" + super.preDisplayStmts());
    boolean isOpenMap = gPackage == GISPackage.OPEN_MAP;
    context.put("openMap", new Boolean(isOpenMap));
    if (isOpenMap) {
      Map groupTitleMap = new HashMap();
      for (int i = 0, n = compileables.size(); i < n; i++) {
        Object o = compileables.get(i);
        if (o instanceof VectorAgentProducer) {
          VectorAgentProducer producer = (VectorAgentProducer) o;
          String groupName = producer.getGroupName();
          if (producer.getDatasource().isValid()) {
            groupTitleMap.put(groupName, groupName);
            String shapefile = producer.getDatasource().getShapefile().replace('\\', '/');
            context.put("shapefile", shapefile);
          }
        }
      }
      context.put("layers", groupTitleMap);
      
    }
    
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();
    try {
      String template = getClass().getName().replace('.', '/') + "PreDisplay.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    //System.out.println("stmt = \n" + stmt);
    return stmt.toString();
  }

  // incorporates the EvolverComponents that are children of this model
  // into the DynamicType the represents the compilied class for this
  // model. In general EvolverComponents add instance variables
  // (fields) into the compiled class, and the DynamicType needs to
  // represent these fields during compilation.
  private void processComponents(NQPyDynamicClass me) {
    for (int i = 0; i < components.size(); i++) {
      EvolverComponent comp = (EvolverComponent) components.get(i);
      Hashtable h = comp.fields();
      Enumeration e = h.keys();
      while (e.hasMoreElements()) {
        String compName = (String) e.nextElement();
        String className = (String) h.get(compName);
        me.addField(compName, className);
      }
    }
  }

  /**
   * Creates a DynamicType that represents the compiled class produced
   * by this NetworkModelProducer. This DynamicType is added to the
   * passed-in Hashtable (key is class name, value is DynamicType
   * which is then returned.
   */
  public Hashtable buildType(Hashtable h, ByteCodeGenerator generator) {
    // check that we don't have conflicting names and groups names in our agent producers.
    sanityCheck();

    String fqn = pkg + "." + className;
    NQPyDynamicClass me = new NQPyDynamicClass(fqn,
            GISSimModelImpl.class);
    me.addMethod(className, fqn, new String[]{});
    processComponents(me);
    processAgentProducers(me);
    me = parameters.buildType(me);
    me = actions.buildType(me);
    h.put(me.getJClassName(), me);

    // this must come after processAgentProducers
    for (int i = 0; i < compileables.size(); i++) {
      EvolverCompileable bean = (EvolverCompileable) compileables.get(i);
      h = bean.buildType(h, generator);
    }

    generator.putDetail(EvolverBean.MODEL_NAME, fqn);
    return h;
  }

  public Class getDynamicSuperclass() {
    return GISSimModelImpl.class;
  }

  public ConstrainedProperty getGisPackage() {
    return gisPackage;
  }

  public void setGisPackage(ConstrainedProperty gisPackage) {
    this.gisPackage = gisPackage;
    pcs.firePropertyChange(PropertyName.GIS_PACKAGE_NAME.getName(), null, this.gisPackage.getSelected());
  }


  /**
   * Compile this NetworkModelProducer. Most of the actual compilation
   * work is done in the EvolverModelCompiler.
   */
  public void compile(String dir, ByteCodeGenerator generator) {


    for (int i = 0; i < compileables.size(); i++) {
      EvolverBean bean = (EvolverBean) compileables.get(i);
      bean.compile(dir, generator);
    }

    EvolverModelCompiler c = new EvolverModelCompiler();
    c.init(this, pkg);
    c.compile(generator, dir);
  }

  public void loadFromXML(Element modelE) {
    super.loadFromXML(modelE);
    gisPackage.setSelectedView(modelE.getAttribute("gisPackage"));
    setGisPackage(gisPackage);
  }


  public void saveXML(BufferedWriter out) throws IOException {
    XMLFactory fac = new XMLFactory(EvolverConstants.XML_MODEL_FULL,
            this, xmlProps);

    try {
      out.write(fac.getXMLStartAttrib());
      out.write("gisPackage=\"");
      out.write(gisPackage.getSelectedView().toString());
      out.write("\" >");
      out.newLine();
      super.saveXML(out);

      out.newLine();
      out.write(fac.getXMLEnd());
      out.newLine();
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }
}


