package uchicago.src.simbuilder.beans.gis;

import layout.TableLayout;

import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.Vector;

/**
 * @author Nick Collier
 * @version $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:27 $
 */
public class VectorDataSourcePropertyPanel extends JPanel {

  private VectorDataSource dataSource;
  private VectorDataSourcePropertyEditor editor;
  private Vector shpFiles = new Vector();
  private JComboBox fileBox = new JComboBox(shpFiles);
  private JButton browseBtn = new JButton("Browse");
  private JButton okBtn = new JButton("OK");
  private JButton cancelBtn = new JButton("Cancel");
  private JTable fieldsTbl = new JTable();
  private JDialog dialog;
  private SymbologyDialog symbology;
  private JTabbedPane tabs = new JTabbedPane();

  private FileFilter shpFilter = new FileFilter() {
    public boolean accept(File f) {
      return f.isDirectory() ? true : f.getAbsolutePath().endsWith(".shp");
    }

    public String getDescription() {
      return "Shape File (*.shp)";
    }
  };

  private VectorFieldsTableModel dataSourceTableModel;

  public VectorDataSourcePropertyPanel(VectorDataSource dataSource) {
    super(new BorderLayout());
    this.dataSource = dataSource;
    shpFiles.addAll(dataSource.getShapefiles());
    fileBox.setRenderer(new FileCellRenderer());
    guiInit();
    addListeners();
    if (shpFiles.size() > 0) {
      fileBox.setSelectedItem(new File(dataSource.getShapefile()));
    }
  }

  private void guiInit() {
    JPanel top = new JPanel();
    top.setBorder(BorderFactory.createTitledBorder("Name"));
    double[][] sizes = {{4, .70, 4, .3, 4}, {4, TableLayout.FILL, 4}};
    top.setLayout(new TableLayout(sizes));
    top.add(fileBox, "1, 1");
    top.add(browseBtn, "3, 1");
    add(top, BorderLayout.NORTH);

    JPanel center = new JPanel(new BorderLayout());
    center.add(new JScrollPane(fieldsTbl), BorderLayout.CENTER);
    center.setBorder(BorderFactory.createTitledBorder("Data Source Fields"));
    tabs.add("Fields", center);
    this.add(tabs, BorderLayout.CENTER);
    
    symbology = new SymbologyDialog(dataSource);
    center = new JPanel(new BorderLayout());
    center.add(symbology, BorderLayout.CENTER);
    tabs.add("Symbology", center);

    JPanel bottom = new JPanel();
    bottom.add(okBtn);
    bottom.add(cancelBtn);
    add(bottom, BorderLayout.SOUTH);

    dataSourceTableModel = new VectorFieldsTableModel(dataSource);
    fieldsTbl.setModel(dataSourceTableModel);
  }

  private void addListeners() {
    browseBtn.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        openDataSource();
      }
    });

    fileBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        fileBox.setToolTipText(fileBox.getSelectedItem().toString());
        checkShapefile();
      }
    });

    okBtn.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataSource.isValid() && dataSource.getShapefile().trim().length() > 0) {
          JOptionPane.showMessageDialog(dialog, "Invalid shapefile. Please select another.", "Invalid Shapefile",
                  JOptionPane.ERROR_MESSAGE);
          return;
        }
        if (dialog != null) dialog.dispose();
        symbology.setDisplayDataOnDataSource();
        editor.setValue(dataSource);
        editor.firePropertyChange();
      }
    });

    cancelBtn.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (dialog != null) dialog.dispose();
      }
    });
  }

  private void checkShapefile() {
    String shpFile = fileBox.getSelectedItem().toString();
    if (!dataSource.getShapefile().equals(shpFile)) {
      try {
        dataSource.setShapefile(shpFile);
        dataSourceTableModel.fireTableDataChanged();
        symbology.reset();
      } catch (IOException ex) {
        JOptionPane.showMessageDialog(dialog, "Invalid shapefile: " + ex.getMessage(), "Invalid Shapefile",
                JOptionPane.ERROR_MESSAGE);
        shpFiles.remove(shpFile);
        dataSource.getShapefiles().remove(shpFile);
        dataSourceTableModel.update();
        if (shpFiles.size() > 0) {
          fileBox.setSelectedIndex(0);
        } else {
          fileBox.setSelectedIndex(-1);
        }
        checkShapefile();
      }
    }
  }

  private void openDataSource() {
    JFileChooser chooser = new JFileChooser("./");

    if (dataSource.isValid()) {
      chooser.setSelectedFile(new File(dataSource.getShapefile()));
    }
    chooser.setFileFilter(shpFilter);
    if (chooser.showOpenDialog(dialog) != JFileChooser.CANCEL_OPTION) {
      String shpFile = chooser.getSelectedFile().getAbsolutePath();
      try {
        dataSource.setShapefile(shpFile);
      } catch (IOException e) {
        JOptionPane.showMessageDialog(dialog, "Invalid shapefile: " + e.getMessage(), "Invalid Shapefile",
                JOptionPane.ERROR_MESSAGE);
        if (fileBox.getSelectedItem() != null)
          try {
            dataSource.setShapefile(fileBox.getSelectedItem().toString());
          } catch (IOException e1) {
            e1.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
          }
        dataSourceTableModel.update();
        return;

      }

      String shapefile = dataSource.getShapefile();
      dataSourceTableModel.update();
      dataSource.addShapefile(shapefile);
      shpFiles.clear();
      shpFiles.addAll(dataSource.getShapefiles());
      fileBox.setSelectedItem(new File(shapefile));
      symbology.reset();
    }
  }

  public void display(JFrame frame, VectorDataSourcePropertyEditor editor) {
    this.editor = editor;
    dialog = new JDialog(frame, false);
    dialog.setTitle("Data Source Editor");
    dialog.getContentPane().setLayout(new BorderLayout());
    dialog.getContentPane().add(this, BorderLayout.CENTER);
    dialog.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent evt) {
        dialog.dispose();
      }
    });

    dialog.pack();
    dialog.setSize(350, 350);
    uchicago.src.guiUtils.GuiUtilities.centerComponentOnScreen(dialog);
    dialog.setVisible(true);
  }

  class FileCellRenderer extends DefaultListCellRenderer {
    public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
      if (value != null) {
        String name = value.toString();
        int sepIndex = name.lastIndexOf(File.separator);
        if (sepIndex != -1) {
          name = name.substring(sepIndex + 1, name.length());
          value = name;
        }
      }
      return super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);    //To change body of overridden methods use File | Settings | File Templates.
    }
  }


  public static void main(String[] args) {
    VectorDataSourcePropertyPanel panel = new VectorDataSourcePropertyPanel(new VectorDataSource());
    panel.display(new JFrame(), null);
  }

}
