package uchicago.src.simbuilder.beans.grid;


import org.w3c.dom.Element;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.nqpy.compiler.CompilerException;
import uchicago.src.simbuilder.base.DefaultGridAgent;
import uchicago.src.simbuilder.base.GridGameAgent;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.sim.util.RepastException;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Hashtable;

public class GridAgentProducer extends AbstractAgentProducer {

  private static final String[] plainAgentParams = {"x", "y"};
  private static final String[] gameAgentParams = {"x", "y", "payoff"};
  
  private static final Integer RANDOM_LOCATION = new Integer(0);
  private static final Integer MANUAL_LOCATION = new Integer(1);
  private static final Integer DUMMY_NUM_AGENTS = new Integer(-1);

  private ConstrainedProperty type, location;
  private int numAgents = 50;

  public GridAgentProducer() {
    super();
    name = "GridAgent";
    groupName = "GridAgentGroup";
    actions.addVarName("space");

    Class[] vals = {DefaultGridAgent.class, GridGameAgent.class};
    String[] views = {"Plain", "Game"};
    type = new ConstrainedProperty(vals, 0, views);
    
    Integer[] locationVals = {RANDOM_LOCATION, MANUAL_LOCATION};
    String[] locationViews = {"Random", "Manual"};
    location = new ConstrainedProperty(locationVals, 0, locationViews);
  }
  
  /**
   * Sets the group name for NetworkNodeProducer.
   * 
   * Overrides the default to change the network source parameter names in the
   * parent model.
   * 
   * @param groupName the name of the group
   */ 
  public void setGroupName(String groupName) {
    super.setGroupName(groupName); 
    pcs.firePropertyChange(PropertyName.GRID_NUM_AGENTS.getName(), DUMMY_NUM_AGENTS, new Integer(numAgents));
    
  }
  
  public int getNumAgents() {
    return numAgents;
  }

  public void setNumAgents(int val) {
    numAgents = val;
    pcs.firePropertyChange(PropertyName.GRID_NUM_AGENTS.getName(), DUMMY_NUM_AGENTS, new Integer(numAgents));
  }
  
  /**
   * Adds this EvolverBean to the specified EvolverParent. This method is
   * corresponds to the accept method in the visitor pattern. Classes which
   * implement this method should call the appropriate method of the
   * EvolverParent with themselves as the parameter.
   * 
   * Overrides parent implementation in order to fire property change of
   * num agents so the model will be setup correctly from the start.
   *
   * @throws uchicago.src.sim.util.RepastException
   *          if the
   *          parent cannot successfully add the child
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    super.addToParent(parent);
    // old value has to be different from the new value in order for the
    // event to be fired.
    pcs.firePropertyChange(PropertyName.GRID_NUM_AGENTS.getName(), DUMMY_NUM_AGENTS, new Integer(numAgents));
  }
  
  public ConstrainedProperty getLocation() {
    return location;
  }
  
  public void setLocation(ConstrainedProperty location) {
    this.location = location;
  }

  /**
   * Returns the nqpy that becomes part of the model's build (not the produced agent's)
   * method.
   *
   * @return
   */
  public String getNQPyInit() {
    VelocityContext context = new VelocityContext();
    context.put("groupName", getGroupName());
    context.put("name", name);
    context.put("random", new Boolean(location.getSelected().equals(RANDOM_LOCATION)));
    context.put("numAgentsParam", "self." + getGroupName() + "NumAgents");
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "NQPyInit.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    return stmt.toString();
  }

  /*
  public String getNQPySetup() {
    VelocityContext context = new VelocityContext();
    context.put("groupName", getGroupName());
    context.put("name", name);
    context.put("superSetup", super.getNQPySetup());
    context.put("numAgentsParam", "self." + getGroupName() + "NumAgents");
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "NQPySetup.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    return stmt.toString();
  }
  */

  /**
   * Builds a DynamicType that represents the class compiled from this
   * NetworkNodeProducer.
   */
  public Hashtable buildType(Hashtable h, ByteCodeGenerator generator) {

    for (int i = 0; i < compileables.size(); i++) {
      EvolverAgentCompileable comp =
              (EvolverAgentCompileable) compileables.get(i);
      h = comp.buildType(h, generator);
    }

    String fqn = pkg + "." + name;
    NQPyDynamicClass me = new NQPyDynamicClass(fqn, (Class) type.getSelected());

    // add a no-op constructor
    me.addMethod(name, fqn, new String[]{});
    me = parameters.buildType(me);
    implementAsCustomProbeable(me);
    me = actions.buildType(me);
    String modelName = (String) generator.getDetail(EvolverBean.MODEL_NAME);
    me.addField("model", modelName);
    me.addMethod("setModel", "void", new String[]{modelName});
    h.put(me.getJClassName(), me);
    generator.putDetail(EvolverBean.AGENT_NAME, fqn);

    return h;
  }

  /**
   * Compile the class described by this GridAgentProducer into
   * bytecode.
   */
  public void compile(String dir, ByteCodeGenerator generator) {
    for (int i = 0; i < compileables.size(); i++) {
      EvolverBean bean = (EvolverBean) compileables.get(i);
      bean.compile(dir, generator);
    }

    String fqn = pkg + "." + name;
    ArrayList list = new ArrayList();
    list.add("model");
    String[] params = new String[1];
    String modelName = (String) generator.getDetail(EvolverBean.MODEL_NAME);
    params[0] = modelName;

    generator.createClass(fqn, (Class) type.getSelected());
    generator.createFieldsConstructor(list, params);

    if (type.getSelectedIndex() == 0)
      compileAsCustomProbeable(generator,
              plainAgentParams);
    else
      compileAsCustomProbeable(generator, gameAgentParams);

    parameters.compile(generator, null);
    actions.compile(generator, fqn);
    generator.finishCtor();
    makeModelSet(generator);
    generator.finishClass();
    generator.writeToFile(fqn, dir);
  }

  public void setParameters(Fields newParameters) {
    super.setParameters(newParameters);
    actions.addVarName("space");
  }

  public ConstrainedProperty getType() {
    return type;
  }

  public void setType(ConstrainedProperty val) {
    type = val;
  }


  /**
   * Saves an xml representation of this NetworkNodeProducer to the specified
   * output stream.
   *
   * @param out the BufferedWriter to write the xml to.
   */
  public void saveXML(BufferedWriter out) throws IOException {
    VelocityContext context = new VelocityContext();
    context.put("class", getClass());
    context.put("type", type.getSelectedView());
    context.put("numAgents", String.valueOf(getNumAgents()));
    context.put("groupName", getGroupName());
    context.put("location", location.getSelectedView());
    context.put("name", name);
    
    StringWriter superData = new StringWriter();
    BufferedWriter writer = new BufferedWriter(superData);
    super.saveXML(writer);
    writer.close();
    
    context.put("superData", superData.toString());
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "XML.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    out.write(stmt.toString());
    out.newLine();
  }

  public void loadFromXML(Element nodeE) {
    // loads name, actions, parameters and schedule
    super.loadFromXML(nodeE);
    type.setSelectedView(nodeE.getAttribute("agentType"));
    setType(type);
    location.setSelectedView(nodeE.getAttribute("location"));
    setLocation(location);
    setNumAgents(Integer.parseInt(nodeE.getAttribute("numAgents")));
    setGroupName(nodeE.getAttribute("groupName"));
  }
}
