package uchicago.src.simbuilder.beans.grid;

import org.w3c.dom.Element;
import uchicago.src.sim.util.RepastException;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.simbuilder.util.BooleanMessage;

import java.awt.*;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * EvolverBean for adding network display capability to an Evolver-
 * created simulation.
 * <p/>
 * $Revision: 1.2 $ $Date: 2004/10/18 20:18:11 $
 */

public class GridDisplayProducer extends AbstractEvolverComponent implements Schedulable {

  private String title = "Grid Display";
  private ScheduleDetails details = new ScheduleDetails();
  private static final String UPDATE_DISPLAY = "update_display";
  private Color background = Color.black;

  static {
    java.beans.PropertyEditorManager.registerEditor(java.awt.Color.class,
            uchicago.src.simbuilder.beans.
            ColorPropertyEditor.class);
  }


  public GridDisplayProducer() {
    ArrayList a = new ArrayList();
    a.add(UPDATE_DISPLAY);
    details.setActionNames(a);

    details.addDetail(new ScheduleDetail("update_display", 1,
            ScheduleDetails.EVERY_TICK, false));
  }

  /**
   * Performs a precompile check making sure that the component is ready to compile.
   *
   * @return
   */
  public BooleanMessage sanityCheck() {
    return BooleanMessage.TRUE;
  }

  /**
   * Gets the ScheduleDetails representing the schedule property for
   * this GridDisplayProducer These details are for display in an editor
   * and not for compilation. schedule() returns ScheduleDetails for
   * compilation.
   */
  public ScheduleDetails getSchedule() {
    return details;
  }

  /**
   * Sets the ScheduleDetails representing the schedule property for
   * this GridDisplayProducer.
   */
  public void setSchedule(ScheduleDetails val) {
    details = val;
  }

  public Color getBackgroundColor() {
    return background;
  }

  public void setBackgroundColor(Color val) {
    background = val;
  }


  /**
   * Gets the title of the display.
   */
  public String getTitle() {
    return title;
  }

  /**
   * Sets the title of the display.
   */
  public void setTitle(String val) {
    title = val;
  }

  /**
   * Adds this EvolverBean to the specified EvolverParent. This method is
   * corresponds to the accept method in the visitor pattern. Classes which
   * implement this method should call the appropriate method of the
   * EvolverParent with themselves as the parameter.
   *
   * @throws uchicago.src.sim.util.RepastException
   *          if the
   *          parent cannot successfully add the child
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    parent.addComponent(this);
  }

  /**
   * Removes this EvolverBean from the specified parent.
   */
  public void removeFromParent(EvolverParent parent) {
    parent.removeComponent(this);
  }

  /**
   * Compile this GridDisplayProducer. As EvolverComponents typically do not
   * need compilation, this method does nothing.
   */
  public void compile(String dir, ByteCodeGenerator generator) {
  }

  /**
   * Returns a String of rscript statements to be added to an
   * EvolverModel's _build method. Can return null if nothing is to be added.
   */
  public String buildStmts() {
    return null;
  }

  /**
   * Returns a String of rscript statements to be added to an
   * EvolverModel's _setup method. Can return null if nothing is to be
   * added.
   */
  public String setupStmts() {

    StringBuffer b = new StringBuffer("if (self._surface):\n\tself._surface.dispose()\n");
    b.append("self._surface = DisplaySurface(self, \"");
    b.append(title);
    b.append("\")\n");
    b.append("self.registerDisplaySurface(self.name, self._surface)\n");
    return b.toString();
  }

  /**
   * Returns a String of nqpy statements to be added to an
   * EvolverModel's _display method. Can return null if nothing is to be
   * added.
   */
  public String displayStmts() {
    StringBuffer b = new StringBuffer();
    b.append("color = Color(");
    b.append(background.getRed());
    b.append(", ");
    b.append(background.getGreen());
    b.append(", ");
    b.append(background.getBlue());
    b.append(")\n");
    b.append("self._surface.setBackground(color)\n");
    b.append("self.addSimEventListener(self._surface)\n");
    b.append("self._surface.display()\n");
    return b.toString();
  }

  /**
   * Returns the ScheduleDetails used during compilation.
   */
  public ScheduleDetails schedule() {
    if (details.size() > 0) {
      ScheduleDetail d = details.getDetail(0);
      d.setTarget("Grid Display");
      d.setNQPy("self._surface.updateDisplay()\n");
    }

    return details;
  }


  /**
   * Returns a hashtable containing the fields that become fields in
   * class compiled from the EvolverModel associated with this
   * GridDisplayProducer. The key is the field name, and the value is the
   * fully-qualified type (e.g. uchicago.src.sim.gui.DisplaySurface) of
   * that field.
   */
  public Hashtable fields() {
    Hashtable h = new Hashtable();
    h.put("_surface", "uchicago.src.sim.gui.DisplaySurface");
    return h;
  }

  public ArrayList methods() {
    return null;
  }

  /**
   * Saves an xml representation of this GridDisplayProducer to the specified
   * output stream.
   *
   * @param out the BufferedWriter to write the xml to.
   */
  public void saveXML(BufferedWriter out) throws IOException {
    StringBuffer b = new StringBuffer("<");
    b.append(EvolverConstants.XML_BEAN_FULL);
    b.append(" type=\"Component\" class=\"");
    b.append(getClass());
    b.append("\" title=\"");
    b.append(title);
    b.append("\" backgroundColor=\"");
    b.append(background.getRed());
    b.append(" ");
    b.append(background.getGreen());
    b.append(" ");
    b.append(background.getBlue());
    b.append("\" >");
    out.write(b.toString());
    out.newLine();
    getSchedule();
    out.write(details.toXML());
    out.newLine();
    out.write("</" + EvolverConstants.XML_BEAN_FULL + ">");
  }

  public void loadFromXML(Element dispE) {
    setTitle(dispE.getAttribute("title").trim());

    String colorStr = dispE.getAttribute("backgroundColor");
    String red = ViolinStrings.Strings.word(colorStr, 0);
    String green = ViolinStrings.Strings.word(colorStr, 1);
    String blue = ViolinStrings.Strings.word(colorStr, 2);
    Color c = new Color(Integer.parseInt(red), Integer.parseInt(green),
            Integer.parseInt(blue));
    setBackgroundColor(c);

    setSchedule(XMLLoader.getScheduleDetails(dispE));
    ArrayList a = new ArrayList();
    a.add(UPDATE_DISPLAY);
    details.setActionNames(a);
  }
}
