package uchicago.src.simbuilder.beans.grid;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.w3c.dom.Element;
import uchicago.src.nqpy.compiler.CompilerException;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.sim.util.RepastException;
import uchicago.src.simbuilder.EvolverModelCompiler;
import uchicago.src.simbuilder.base.DefaultGridModel;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.xml.XMLFactory;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeSupport;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.*;


public class GridModelProducer extends AbstractEvolverModel {

  private int width = 100;
  private int height = 100;
  private ConstrainedProperty spaceType;
  private boolean displayOn = false;

  public GridModelProducer() {
    pcs = new PropertyChangeSupport(this);
    className = "GridModel";
    dispName = "Grid Model";
    String[] vals = {"Object2DGrid", "Object2DTorus"};
    String[] views = {"Grid", "Torus"};

    spaceType = new ConstrainedProperty(vals, 0, views);
    actions.createDefaultAction("initAgents", "");

  }

  /**
   * Gets the width of the space.
   */
  public int getSpaceWidth() {
    return width;
  }

  /**
   * Sets the width of the space.
   */
  public void setSpaceWidth(int val) {
    width = val;
  }

  /**
   * Gets the height of the space.
   */
  public int getSpaceHeight() {
    return height;
  }

  /**
   * Sets the height of the space.
   */
  public void setSpaceHeight(int val) {
    height = val;
  }

  public ConstrainedProperty getSpaceType() {
    return spaceType;
  }

  public void setSpaceType(ConstrainedProperty val) {
    spaceType = val;
  }

  /**
   * Adds the specified compileable to this EvolverParent.
   * Overrides parent to add property change support.
   *
   * @param child the child to add to this parent
   * @throws uchicago.src.sim.util.RepastException
   *
   *
  public void addCompileable(EvolverCompileable child) throws RepastException {
    super.addCompileable(child);
    //child.addPropertyChangeListener(this);
  }
   */

  /**
   * Removes the specified compileable to this EvolverParent. Overrides parent
   * to remove property change support.
   *
  public void removeCompileable(EvolverCompileable child) {
    super.removeCompileable(child);
    //child.removePropertyChangeListener(this);
  }
   */

  /**
   * Gets nqpy statements that become the first part of the compiled class's
   * _buildStatement. These statements are guaranteed to execute before any
   * agent added code executed in _build.
   */
  public String preBuildStmts() {
    List groupNames = getGroupNames();
    VelocityContext context = new VelocityContext();
    context.put("groupNames", groupNames);
    context.put("type", spaceType.getSelected());
    context.put("width", String.valueOf(width));
    context.put("height", String.valueOf(height));
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "PreBuild.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    return stmt.toString();
  }

  /**
   * Gets nqpy statements that become the first part of the compiled class's
   * _displayStatement. These statements are guaranteed to execute before any
   * agent or component added code executed in _display.
   *
   * @return
   */
  public String preDisplayStmts() {
    if (!displayOn) return "";
    List groupNames = getGroupNames();
    VelocityContext context = new VelocityContext();
    context.put("groupNames", groupNames);
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "PreDisplay.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    return stmt.toString();
  }

  private List getGroupNames() {
    List groupNames = new ArrayList();
    for (Iterator iter = compileables.iterator(); iter.hasNext();) {
      Object o = iter.next();
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        groupNames.add(producer.getGroupName());

      }
    }
    return groupNames;
  }


  /**
   * Returns NQPy statements that become part of the _build method in
   * the class compiled from this EvolverModel.
   */
  public String buildStmts() {
    if (compileables.size() > 0) return "self.initAgents()\n\n";
    return "";
  }

  // incorporates the EvolverComponents that are children of this model
  // into the DynamicType the represents the compilied class for this
  // model. In general EvolverComponents add instance variables
  // (fields) into the compiled class, and the DynamicType needs to
  // represent these fields during compilation.
  private void processComponents(NQPyDynamicClass me) {
    for (int i = 0, n = components.size(); i < n; i++) {
      EvolverComponent comp = (EvolverComponent) components.get(i);
      Hashtable h = comp.fields();
      Enumeration e = h.keys();
      while (e.hasMoreElements()) {
        String compName = (String) e.nextElement();
        String className = (String) h.get(compName);
        me.addField(compName, className);
      }
    }

    displayOn = me.hasDynField("_surface");

  }

  /*
   * Processes the compileables that are AgentProducers, adding their
   * group names as ivars to the produced model class together with
   * a get accessor.
   *
   * Overrides superclass to add a "space" field for each agent producer.
   */
  protected void processAgentProducers(NQPyDynamicClass modelClass) {
    super.processAgentProducers(modelClass);
    /*
    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        modelClass.addField(producer.getGroupName() + "Space", spaceType.getSelected().toString());
      }
    }
    */
  }


  /**
   * Creates a DynamicType that represents the compiled class produced
   * by this NetworkModelProducer. This DynamicType is added to the
   * passed-in Hashtable (key is class name, value is DynamicType
   * which is then returned.
   */
  public Hashtable buildType(Hashtable h, ByteCodeGenerator generator) {
    sanityCheck();
    String fqn = pkg + "." + className;
    NQPyDynamicClass me = new NQPyDynamicClass(fqn,
            DefaultGridModel.class);

    me.addMethod(className, fqn, new String[]{});
    processComponents(me);
    processAgentProducers(me);
    me = parameters.buildType(me);
    me = actions.buildType(me);
    h.put(me.getJClassName(), me);

    for (int i = 0, n = compileables.size(); i < n; i++) {
      EvolverCompileable bean = (EvolverCompileable) compileables.get(i);
      h = bean.buildType(h, generator);
    }

    generator.putDetail(EvolverBean.MODEL_NAME, fqn);
    return h;

  }

  public Class getDynamicSuperclass() {
    return DefaultGridModel.class;
  }

  /**
   * Compile this GridModelProducer. Most of the actual compilation
   * work is done in the EvolverModelCompiler.
   */
  public void compile(String dir, ByteCodeGenerator generator) {
    for (int i = 0, n = compileables.size(); i < n; i++) {
      EvolverBean bean = (EvolverBean) compileables.get(i);
      bean.compile(dir, generator);
    }
    
    // todo remove this when space parameters are added as real parameters.
    /*
    for (int i = 0, n = compileables.size(); i < n; i++) {
      Object o = compileables.get(i);
      if (o instanceof AgentProducer) {
        AgentProducer producer = (AgentProducer) o;
        parameters.addImplicitField(this, producer.getGroupName() + "Space", "uchicago.src.sim.space." + 
                spaceType.getSelected().toString(), "", false, true);
      }
    }
    */

    try {
      EvolverModelCompiler c = new EvolverModelCompiler();
      c.init(this, pkg);
      c.compile(generator, dir);
    } finally {
      parameters.removeImplicitFields(this);
    }
  }

  public void loadFromXML(Element modelE) {
    // loads actions, model name, display name, parameters,
    // schedule and master schedule
    super.loadFromXML(modelE);
    setSpaceWidth(XMLLoader.getIntVal(modelE, "spaceWidth"));
    setSpaceHeight(XMLLoader.getIntVal(modelE, "spaceHeight"));

    spaceType.setSelected(modelE.getAttribute("spaceType").trim());
    setSpaceType(spaceType);
  }

  public void saveXML(BufferedWriter out) throws IOException {
    XMLFactory fac = new XMLFactory(EvolverConstants.XML_MODEL_FULL,
            this, xmlProps);
    try {
      out.write(fac.getXMLStartAttrib());
      StringBuffer b = new StringBuffer();
      b.append(" spaceWidth=\"");
      b.append(getSpaceWidth());
      b.append("\" spaceHeight=\"");
      b.append(getSpaceHeight());
      b.append("\" spaceType=\"");
      b.append(spaceType.getSelected());
      b.append("\" >");
      out.write(b.toString());
      out.newLine();
      super.saveXML(out);

      out.newLine();
      out.write(fac.getXMLEnd());
      out.newLine();
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  public void showSchDetails() {
    schedule.showDetails();
  }

  // checks that the compileables etc. do not conflict with each other
  // and produce bad code
  protected void sanityCheck() {
    super.sanityCheck();
    int limit = width * height;
    String badGroup = "";
    for (Iterator iter = parameters.fields().iterator(); iter.hasNext();) {
      FieldData data = (FieldData) iter.next();
      if (data.getName().endsWith("NumAgents")) {
        int numAgents = Integer.parseInt(data.getDefaultValue());
        if (numAgents > limit) {
          // -9 as that is the length of "NumAgents"
          badGroup = data.getName().substring(0, data.getName().length() - 9);
        }
      }
    }

    if (badGroup.length() > 0) {
      // find the agentproducer for that as this group name
      for (int i = 0, n = compileables.size(); i < n; i++) {
        Object o = compileables.get(i);
        if (o instanceof AgentProducer) {
          AgentProducer producer = (AgentProducer) o;
          if (producer.getGroupName().equalsIgnoreCase(badGroup)) {
            throw new RuntimeException("Too many '" + producer.className() + "' agents for specified size of the space.\n" +
                    "Change the 'Number of Agents' property or the default value of the model field");
          }
        }
      }
      
      // this is a last ditch attempt to warn when we can't find the agent name for the
      // specified group
      throw new RuntimeException("Too many agents in group '" + badGroup +  "' agents for the specified size of the space.\n" +
                    "Change the 'Number of Agents' property or the default value of the model field");
    }
  }

  private void addNumAgentImpilicitParameter(GridAgentProducer producer, Integer numAgents) {
    String groupName = producer.getGroupName();
    parameters.removeImplicitFields(producer);
    parameters.addImplicitField(producer, groupName + "NumAgents", "int",
            numAgents.toString(), true);
    parameters.addImplicitField(producer, groupName + "Space", "uchicago.src.sim.space." +
            spaceType.getSelected().toString(), "", false, true);
    setParameters(parameters);
  }

  public void propertyChange(PropertyChangeEvent evt) {
    // let super class handle the events it is interested in.
    super.propertyChange(evt);
    if (evt.getPropertyName() != null && evt.getPropertyName().equals(PropertyName.GRID_NUM_AGENTS.getName())) {
      Integer newValue = (Integer) evt.getNewValue();
      addNumAgentImpilicitParameter((GridAgentProducer) evt.getSource(), newValue);
    }
  }
}
