package uchicago.src.simbuilder.beans.network;

import org.w3c.dom.Element;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.sim.util.RepastException;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.simbuilder.util.BooleanMessage;
import uchicago.src.nqpy.compiler.CompilerException;

import java.beans.PropertyEditorManager;
import java.beans.PropertyChangeEvent;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * EvolverBean for adding network display capability to an Evolver-
 * created simulation.
 * <p/>
 * $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:28 $
 */

public class NetworkDisplayProducer extends AbstractEvolverComponent implements Schedulable {
  
  private ScheduleDetails details = new ScheduleDetails();
  private static final String UPDATE_DISPLAY = "update_display";
  private static final String UPDATE_LAYOUT = "update_layout";
  private LayoutProperty layout = new FruchLayoutProperty();
  private ConstrainedProperty nodeShape;
  private int width = 400;
  private int height = 400;

  static {
    PropertyEditorManager.
            registerEditor(uchicago.src.simbuilder.beans.network.LayoutProperty.class,
                    uchicago.src.simbuilder.beans.network.LayoutEditor.class);
  }

  public NetworkDisplayProducer() {
    ArrayList a = new ArrayList();
    a.add(UPDATE_DISPLAY);
    a.add(UPDATE_LAYOUT);
    details.setActionNames(a);

    String[] vals1 = {"OvalNetworkItem", "RectNetworkItem"};
    String[] views1 = {"Oval", "Rectangle"};

    nodeShape = new ConstrainedProperty(vals1, 0, views1);
  }
  
  /**
   * Performs a precompile check making sure that the component is ready to compile.
   *
   * @return
   */
  public BooleanMessage sanityCheck() {
    return BooleanMessage.TRUE;
  }

  /**
   * Gets the ScheduleDetails representing the schedule property for
   * this NetworkDisplayProducer These details are for display in an editor
   * and not for compilation. schedule() returns ScheduleDetails for
   * compilation.
   */
  public ScheduleDetails getSchedule() {
    return details;
  }

  /**
   * Sets the ScheduleDetails representing the schedule property for
   * this NetworkDisplayProducer.
   */
  public void setSchedule(ScheduleDetails val) {
    details = val;
  }


  /**
   * Gets the width of the display.
   */
  public int getWidth() {
    return width;
  }

  /**
   * Sets the width of the display.
   */
  public void setWidth(int val) {
    width = val;
  }

  /**
   * Gets the height of the display.
   */
  public int getHeight() {
    return height;
  }

  /**
   * Sets the height of the display.
   */
  public void setHeight(int val) {
    height = val;
  }

  public LayoutProperty getLayout() {
    return layout;
  }

  public void setLayout(LayoutProperty val) {
    layout = val;
  }

  public ConstrainedProperty getNodeShape() {
    return nodeShape;
  }

  public void setNodeShape(ConstrainedProperty val) {
    nodeShape = val;
  }

  /**
   * Adds this EvolverBean to the specified EvolverParent. This method is
   * corresponds to the accept method in the visitor pattern. Classes which
   * implement this method should call the appropriate method of the
   * EvolverParent with themselves as the parameter.
   *
   * @throws uchicago.src.sim.util.RepastException
   *          if the
   *          parent cannot successfully add the child
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    parent.addComponent(this);
  }

  /**
   * Removes this EvolverBean from the specified parent.
   */
  public void removeFromParent(EvolverParent parent) {
    parent.removeComponent(this);
  }

  /**
   * Compile this NetworkDisplayProducer. As EvolverComponents typically do not
   * need compilation, this method does nothing.
   */
  public void compile(String dir, ByteCodeGenerator generator) {
  }

  /**
   * Returns a String of rscript statements to be added to an
   * EvolverModel's _build method. Can return null if nothing is to be added.
   */
  public String buildStmts() {
    return null;
  }

  /**
   * Returns a String of nqpy statements to be added to an
   * EvolverModel's _setup method. Can return null if nothing is to be
   * added.
   */
  public String setupStmts() {

    StringBuffer b = new StringBuffer("if (self._surface):\n\tself._surface.dispose()\n");
    //b.append("print 'dispsurf:', DisplaySurface(self, self.name)\n");
    b.append("self._surface = DisplaySurface(self, self.name + ' Display')\n");
    b.append("self.registerDisplaySurface(self.name, self._surface)\n");

    // b.append("print '!!!!!!!!surface:', self._surface\n");
    return b.toString();
  }

  /**
   * Returns a String of nqpy statements to be added to an
   * EvolverModel's _display method. Can return null if nothing is to be
   * added.
   */
  public String displayStmts() {
    VelocityContext context = new VelocityContext();
    context.put("nodeShape", nodeShape.getSelected());
    String ctorArgs = "self.agentList, " + width + ", " + height;
    context.put("layout", layout.getInitScript("self._layout", ctorArgs, "self._surface"));
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "Display.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    return stmt.toString();
    
    /*
    StringBuffer b = new StringBuffer();

    b.append("for agent as EvolverNodeImpl in self.agentList:\n");
    b.append("\titem = ");
    b.append(nodeShape.getSelected());
    b.append("(0, 0)\n");
    b.append("\tagent.setDrawable(item)\n");
    String ctorArgs = "self.agentList, " + width + ", " + height;
    b.append(layout.getInitScript("self._layout", ctorArgs, "self._surface"));
    b.append("display = Network2DDisplay(self._layout)\n");
    b.append("self._surface.addDisplayableProbeable(display, \"node\")\n");
    b.append("self.addSimEventListener(self._surface)\n");

    b.append("self._layout.updateLayout()\n");
    b.append("self._surface.display()\n");

    return b.toString();
    */
  }

  /**
   * Returns the ScheduleDetails used during compilation.
   */
  public ScheduleDetails schedule() {
    for (int i = 0; i < details.size(); i++) {
      ScheduleDetail d = details.getDetail(i);
      d.setTarget("Network Display");
      StringBuffer b = new StringBuffer("self.");
      if (d.getActionName().equals(UPDATE_DISPLAY)) {
        b.append("_surface.updateDisplay();\n");
      } else {
        b.append("_layout.updateLayout();\n");
      }

      d.setNQPy(b.toString());
    }

    return details;
  }

  /**
   * Returns a hashtable containing the fields that become fields in
   * class compiled from the EvolverModel associated with this
   * NetworkDisplayProducer. The key is the field name, and the value is the
   * fully-qualified type (e.g. uchicago.src.sim.gui.DisplaySurface) of
   * that field.
   */
  public Hashtable fields() {
    Hashtable h = new Hashtable();
    h.put("_surface", "uchicago.src.sim.gui.DisplaySurface");
    h.put("_layout", layout.getClassName());
    return h;
  }

  public ArrayList methods() {
    return null;
  }

  /**
   * Saves an xml representation of this NetworkDisplayProducer to the specified
   * output stream.
   *
   * @param out the BufferedWriter to write the xml to.
   */
  public void saveXML(BufferedWriter out) throws IOException {
    StringBuffer b = new StringBuffer("<");
    b.append(EvolverConstants.XML_BEAN_FULL);
    b.append(" type=\"Component\" class=\"");
    b.append(getClass());
    b.append("\" nodeShape=\"");
    b.append(nodeShape.getSelected());
    b.append("\" >");
    out.write(b.toString());
    out.newLine();
    out.write(layout.toXML());
    out.newLine();
    getSchedule();
    out.write(details.toXML());
    out.newLine();
    out.write("</" + EvolverConstants.XML_BEAN_FULL + ">");
  }

  public void loadFromXML(Element dispE) {
    //layout.setSelected(dispE.getAttribute("layout").trim());
    //setLayout(layout);
    nodeShape.setSelected(dispE.getAttribute("nodeShape").trim());
    setNodeShape(nodeShape);
    setSchedule(XMLLoader.getScheduleDetails(dispE));
    setLayout(XMLLoader.getGraphLayout(dispE));
    ArrayList a = new ArrayList();
    a.add(UPDATE_DISPLAY);
    a.add(UPDATE_LAYOUT);
    details.setActionNames(a);
  }
}
