package uchicago.src.simbuilder.beans.network;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.w3c.dom.Element;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.nqpy.compiler.CompilerException;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.sim.util.RepastException;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * Default agent class for network models created with Evolver. When this
 * is compiled a default step method is created as well as ivar and
 * accessor methods for the EvolverModel for which this is an agent.
 * <p/>
 * $Revision: 1.2 $ $Date: 2004/11/22 21:30:07 $
 */
public class NetworkNodeProducer extends AbstractAgentProducer {

  private static final String[] OTHER_PARAMS = {"width", "height", "x", "y",
                                                "nodeLabel"};

  private NetworkSource source = new UnlinkedNodeNetworkSource();

  public NetworkNodeProducer() {
    super();
    name = "NetworkNodeAgent";
    groupName = "nodes";
  }

  /**
   * Sets the group name for NetworkNodeProducer.
   * 
   * Overrides the default to change the network source parameter names in the
   * parent model.
   * 
   * @param groupName the name of the group
   */ 
  public void setGroupName(String groupName) {
    super.setGroupName(groupName); 
    pcs.firePropertyChange(PropertyName.NETWORK_SOURCE.getName(), new DummyNetworkSource(), source);
  }

  public NetworkSource getNetworkSource() {
    return source;
  }

  public void setNetworkSource(NetworkSource val) {
    NetworkSource oldSource = source;
    source = val;
    pcs.firePropertyChange(PropertyName.NETWORK_SOURCE.getName(), oldSource, source);
  }

  public String getNQPyInit() {
    VelocityContext context = new VelocityContext();
    context.put("groupName", getGroupName());
    context.put("netSetup", source.getInitScript(getFQName(), getGroupName()));
    context.put("superInit", super.getNQPyInit());
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "NQPyInit.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    return stmt.toString();
  }

  /**
   * Builds a DynamicType that represents the class compiled from this
   * NetworkNodeProducer.
   */
  public Hashtable buildType(Hashtable h, ByteCodeGenerator generator) {

    String fqn = pkg + "." + name;
    for (int i = 0; i < compileables.size(); i++) {
      EvolverAgentCompileable comp = (EvolverAgentCompileable) compileables.get(i);
      h = comp.buildType(h, generator);
    }

    NQPyDynamicClass me = new NQPyDynamicClass(fqn,
            EvolverNodeImpl.class);
    // create a no-op constructor.
    me.addMethod(name, fqn, new String[]{});
    me = parameters.buildType(me);
    implementAsCustomProbeable(me);

    me = actions.buildType(me);
    String modelName = (String) generator.getDetail(EvolverBean.MODEL_NAME);
    me.addField("model", modelName);
    me.addMethod("setModel", "void", new String[]{modelName});
    h.put(me.getJClassName(), me);
    generator.putDetail(EvolverBean.AGENT_NAME, fqn);

    return h;
  }

  /**
   * Adds this EvolverBean to the specified EvolverParent. This method is
   * corresponds to the accept method in the visitor pattern. Classes which
   * implement this method should call the appropriate method of the
   * EvolverParent with themselves as the parameter.
   * 
   * Overrides parent implementation in order to fire property change of
   * network source so the model will be setup correctly from the start.
   *
   * @throws uchicago.src.sim.util.RepastException
   *          if the
   *          parent cannot successfully add the child
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    super.addToParent(parent);
    // old value has to be different from the new value in order for the
    // event to be fired.
    pcs.firePropertyChange(PropertyName.NETWORK_SOURCE.getName(), new DummyNetworkSource(), source);
  }

  /**
   * Compile the class described by this NetworkNodeProducer into
   * bytecode.
   */
  public void compile(String dir, ByteCodeGenerator generator) {
    for (int i = 0; i < compileables.size(); i++) {
      EvolverBean bean = (EvolverBean) compileables.get(i);
      bean.compile(dir, generator);
    }

    ArrayList list = new ArrayList();
    list.add("model");
    String[] params = new String[1];
    String modelName = (String) generator.getDetail(EvolverBean.MODEL_NAME);
    params[0] = modelName;

    String fqn = pkg + "." + name;
    generator.createClass(fqn, EvolverNodeImpl.class);
    generator.createFieldsConstructor(list, params);
    compileAsCustomProbeable(generator, OTHER_PARAMS);

    parameters.compile(generator, null);
    actions.compile(generator, fqn);
    generator.finishCtor();
    makeModelSet(generator);
    generator.finishClass();
    generator.writeToFile(fqn, dir);
  }

  /**
   * Saves an xml representation of this NetworkNodeProducer to the specified
   * output stream.
   *
   * @param out the BufferedWriter to write the xml to.
   */
  public void saveXML(BufferedWriter out) throws IOException {
    VelocityContext context = new VelocityContext();
    context.put("class", getClass());
    context.put("groupName", getGroupName());
    context.put("name", name);
    context.put("source", source.toXML());
    
    StringWriter superData = new StringWriter();
    BufferedWriter writer = new BufferedWriter(superData);
    super.saveXML(writer);
    writer.close();
    
    context.put("superData", superData.toString());
     
    // render a template (in the classpath) giving fully qualified
    // path to the template file
    StringWriter stmt = new StringWriter();

    try {
      String template = getClass().getName().replace('.', '/') + "XML.vt";
      Velocity.mergeTemplate(template, "UTF-8", context, stmt);
    } catch (Exception e) {
      throw new CompilerException("Compiler Error" + e);
    }

    out.write(stmt.toString());
    out.newLine();
  }

  public void loadFromXML(Element nodeE) {
    super.loadFromXML(nodeE);
    setNetworkSource(XMLLoader.getNetworkSource(nodeE));
    setGroupName(nodeE.getAttribute("groupName"));
  }
}
