package uchicago.src.simbuilder.beans.network;

import org.w3c.dom.Element;
import uchicago.src.simbuilder.codegen.ByteCodeGenerator;
import uchicago.src.sim.network.NetworkRecorder;
import uchicago.src.sim.util.RepastException;
import uchicago.src.simbuilder.beans.*;
import uchicago.src.simbuilder.util.BooleanMessage;
import uchicago.src.simbuilder.util.RepastPyUtilities;

import java.beans.PropertyEditorManager;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * EvolverBean for adding network recorder capability to an Evolver-
 * created simulation.  Modified by Michael J. North.
 * <p/>
 * $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:29 $
 */

public class NetworkRecorderProducer extends AbstractEvolverComponent implements Schedulable {

  private String fileName = "";
  private String name = "networkRecorder";
  private ScheduleDetails details = new ScheduleDetails();
  private ConstrainedProperty format;
  private ConstrainedProperty netType;

  static {
    PropertyEditorManager.registerEditor(uchicago.src.simbuilder.beans.ConstrainedProperty.class,
            uchicago.src.simbuilder.beans.ConstrainedPropertyEditor.class);
    PropertyEditorManager.registerEditor(uchicago.src.simbuilder.beans.ScheduleDetails.class,
            uchicago.src.simbuilder.beans.ScheduleDetailsPropertyEditor.class);

  }

  public NetworkRecorderProducer() {
    Integer[] vals = {new Integer(NetworkRecorder.DL),
                      new Integer(NetworkRecorder.EXCEL),
                      new Integer(NetworkRecorder.ASCII)};
    String[] views = {"DL Format", "EXCEL", "ASCII"};

    format = new ConstrainedProperty(vals, 0, views);

    Integer[] types = {new Integer(NetworkRecorder.BINARY),
                       new Integer(NetworkRecorder.SMALL),
                       new Integer(NetworkRecorder.LARGE)};
    String[] netViews = {"Binary", "Small", "Large"};
    netType = new ConstrainedProperty(types, 0, netViews);

    ArrayList a = new ArrayList();
    a.add("record");
    a.add("write");
    details.setActionNames(a);
  }
  
  /**
   * Performs a precompile check making sure that the component is ready to compile.
   *
   * @return
   */
  public BooleanMessage sanityCheck() {
    if (!RepastPyUtilities.isValidJavaIdentifier(name)) {
      return new BooleanMessage("Invalid name '" + name + "'" +
              ". Valid names contain only letters, numbers, or '_'",
                false);
    }
    
    if (fileName == null || fileName.length() == 0) {
      return new BooleanMessage("Invalid file name in Network Recorder", false);
    }
    
    return BooleanMessage.TRUE;
  }

  public String getName() {
    return name;
  }

  public void setName(String val) {
    name = val;
  }

  public String getFileName() {
    return fileName;
  }

  public void setFileName(String val) {
    fileName = val;
  }

  /**
   * Gets the ScheduleDetails representing the schedule property for
   * this NetworkRecorderProducer. These details are for display in an editor
   * and not for compilation. schedule() returns ScheduleDetails for
   * compilation.
   */
  public ScheduleDetails getSchedule() {
    return details;
  }

  /**
   * Sets the ScheduleDetails representing the schedule property for
   * this NetworkRecorderProducer.
   */
  public void setSchedule(ScheduleDetails val) {
    details = val;
  }

  public ConstrainedProperty getFormat() {
    return format;
  }

  public void setFormat(ConstrainedProperty val) {
    format = val;
  }

  public ConstrainedProperty getMatrixType() {
    return netType;
  }

  public void setMatrixType(ConstrainedProperty val) {
    netType = val;
  }

  /**
   * Adds this EvolverBean to the specified EvolverParent. This method is
   * corresponds to the accept method in the visitor pattern. Classes which
   * implement this method should call the appropriate method of the
   * EvolverParent with themselves as the parameter.
   *
   * @throws uchicago.src.sim.util.RepastException
   *          if the
   *          parent cannot successfully add the child
   */
  public void addToParent(EvolverParent parent) throws RepastException {
    parent.addComponent(this);
  }

  /**
   * Removes this EvolverBean from the specified parent.
   */
  public void removeFromParent(EvolverParent parent) {
    parent.removeComponent(this);
  }

  /**
   * Compile this NetworkDisplayProducer. As EvolverComponents typically do not
   * need compilation, this method does nothing.
   */
  public void compile(String dir, ByteCodeGenerator generator) {
  }

  /**
   * The instance name to be used for this component.
   */
  public String name() {
    return name;
  }

  /**
   * Returns a String of rscript statements to be added to an
   * EvolverModel's _build method. Can return null if nothing is to be added.
   */
  public String buildStmts() {
    return null;
  }

  /**
   * Returns a String of rscript statements to be added to an
   * EvolverModel's _display method. Can return null if nothing is to be
   * added.
   */
  public String displayStmts() {
    return null;
  }

  /**
   * Returns a String of rscript statements to be added to an
   * EvolverModel's _setup method. Can return null if nothing is to be
   * added.
   */
  public String setupStmts() {

    StringBuffer b = new StringBuffer("self.");
    b.append(name);
    b.append(" = NetworkRecorder(");
    b.append(format.getSelected());
    b.append(", \"");
    b.append(fileName);
    b.append("\", self)");

    return b.toString();
  }

  /**
   * Returns the ScheduleDetails used during compilation.
   */
  public ScheduleDetails schedule() {
    for (int i = 0; i < details.size(); i++) {
      ScheduleDetail d = details.getDetail(i);
      StringBuffer b = new StringBuffer("self.");
      b.append(name);
      if (d.getActionName().equals("record")) {
        b.append(".record(self.agentList, String.valueOf(self.getTickCount()), ");
        b.append(((Integer) format.getSelected()).intValue());
        b.append(")");
      } else {
        b.append(".write()");
      }

      //System.out.println(b);
      d.setNQPy(b.toString());
      d.setTarget(name);
    }
    return details;
  }

  /**
   * Returns a hashtable containing the fields that become fields in
   * class compiled from the EvolverModel associated with this
   * NetworkDisplayProducer. The key is the field name, and the value is the
   * fully-qualified type (e.g. uchicago.src.sim.gui.DisplaySurface) of
   * that field.
   */
  public Hashtable fields() {
    Hashtable h = new Hashtable();
    h.put(name, "uchicago.src.sim.network.NetworkRecorder");
    return h;
  }

  public ArrayList methods() {
    return null;
  }

  /**
   * Saves an xml representation of this NetworkRecorderProducer to the specified
   * output stream.
   *
   * @param out the BufferedWriter to write the xml to.
   */
  public void saveXML(BufferedWriter out) throws IOException {
    StringBuffer b = new StringBuffer("<");
    b.append(EvolverConstants.XML_BEAN_FULL);
    b.append(" type=\"Component\" class=\"");
    b.append(getClass());
    b.append("\" name=\"");
    b.append(name);
    b.append("\" fileName=\"");
    b.append(fileName);
    b.append("\" format=\"");
    b.append(format.getSelected());
    b.append("\" matrixType=\"");
    b.append(netType.getSelected());
    b.append("\" >");
    out.write(b.toString());
    out.newLine();
    getSchedule();
    out.write(details.toXML());
    out.newLine();
    out.write("</" + EvolverConstants.XML_BEAN_FULL + ">");
  }

  public void loadFromXML(Element recE) {
    setName(recE.getAttribute("name"));
    setFileName(recE.getAttribute("fileName"));
    format.setSelected(new Integer(XMLLoader.getIntVal(recE, "format")));
    setFormat(format);
    netType.setSelected(new Integer(XMLLoader.getIntVal(recE, "matrixType")));
    setMatrixType(netType);
    setSchedule(XMLLoader.getScheduleDetails(recE));
    ArrayList a = new ArrayList();
    a.add("write");
    a.add("record");
    details.setActionNames(a);
  }
}
