package uchicago.src.simbuilder.codegen;

import com.go.trove.classfile.*;
import uchicago.src.codegen.RSClassLoader;
import uchicago.src.nqpy.compiler.CompilerException;
import uchicago.src.nqpy.compiler.NQPyClassFactory;
import uchicago.src.nqpy.compiler.NQPyClass;
import uchicago.src.nqpy.compiler.NQPyDynamicClass;
import uchicago.src.simbuilder.beans.EvolverBean;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;


/**
 * Wrapper around trove.classfile.ClassFile methods. ByteCode generator
 * works using the following pattern. There are various createX methods.
 * All the emit methods will operate on the
 * relevant "opened" object. So having created a method, any subsequent
 * emitX calls will add code to that method.
 *
 * @author Nick Collier
 * @version $Revision: 1.2 $ $Date: 2004/12/10 21:57:39 $
 */

public class ByteCodeGenerator {

  public static final String INT = "int";
  public static final String DOUBLE = "double";
  public static final String FLOAT = "float";
  public static final String LONG = "long";
  public static final String BOOLEAN = "boolean";
  public static final String STRING = "java.lang.String";
  public static final String SHORT_STRING = "String";

  protected ArrayList classFiles = new ArrayList();
  protected Hashtable classes = new Hashtable();
  protected Hashtable localVars = new Hashtable();
  protected Hashtable fields = new Hashtable();

  private Hashtable details = new Hashtable();

  protected ClassFile cf;
  protected CodeBuilder builder;
  protected CodeBuilder ctorBuilder;

  protected RSClassLoader loader;

  protected static Set allowableInitFields = new HashSet();

  static {
    allowableInitFields.add(int.class);
    allowableInitFields.add(double.class);
    allowableInitFields.add(float.class);
    allowableInitFields.add(long.class);
    allowableInitFields.add(boolean.class);
    allowableInitFields.add(String.class);
  }


  public ByteCodeGenerator() {
    reInit();
  }

  public void reInit() {
    Class c = uchicago.src.sim.engine.SimModel.class;
    Class b = uchicago.src.sim.engine.BasicAction.class;
    Class a = uchicago.src.sim.util.ByteCodeBuilder.class;
    loader = new RSClassLoader(this.getClass().getClassLoader());
    NQPyClassFactory.setClassLoader(loader);
  }

  /**
   * Adds a jar or class path to search when compiling.
   *
   * @param path
   */
  public void addClassPath(String path) {
    try {
      loader.addPath(path);
      // we just swallow the error, mimicing javac's behavior.
    } catch (MalformedURLException ex) {
    }
  }

  public RSClassLoader getLoader() {
    return loader;
  }

  protected TypeDescriptor[] makeTypeDescriptors(Class[] cls) {
    if (cls == null) return null;

    TypeDescriptor[] d = new TypeDescriptor[cls.length];
    for (int i = 0; i < cls.length; i++) {
      d[i] = new TypeDescriptor(cls[i]);
    }

    return d;
  }

  protected TypeDescriptor[] makeTypeDescriptors(String[] cls) {
    if (cls == null) return null;

    TypeDescriptor[] d = new TypeDescriptor[cls.length];
    for (int i = 0; i < cls.length; i++) {
      d[i] = new TypeDescriptor(cls[i]);
    }

    return d;
  }

  protected Class classFromString(String cName) throws ClassNotFoundException {
    if (cName.equals(INT))
      return int.class;
    else if (cName.equals(DOUBLE))
      return double.class;
    else if (cName.equals(STRING) || cName.equals(SHORT_STRING))
      return String.class;
    else if (cName.equals(FLOAT))
      return float.class;
    else if (cName.equals(LONG))
      return long.class;
    else if (cName.equals(BOOLEAN))
      return boolean.class;
    else {
      try {
        Class clazz = Class.forName(cName);
        return clazz;
      } catch (ClassNotFoundException ex) {
        if (cName.indexOf(".") == -1) {
          // name is not qualified. Try to make the name using the
          // standard imports
          NQPyClass jClass = NQPyClassFactory.makeJavaClass(cName);
          if (jClass != null && jClass.getJClass() != null) return jClass.getJClass();
        }
        throw ex;
      }
    }
  }

  public ClassFile getClassFile() {
    return cf;
  }

  public void setWorkingClass(String name) {
    builder = null;
    ctorBuilder = null;
    cf = (ClassFile) classes.get(name);
  }


  /**
   * Create class with specified name.
   */
  public void createClass(String name) {
    if (cf != null) classFiles.add(cf);
    fields = new Hashtable();
    cf = new ClassFile(name);
    classes.put(name, cf);
  }

  /**
   * Create class with specified name, and specified superClass.
   */
  public void createClass(String name, String superClass) {
    if (cf != null) classFiles.add(cf);
    fields = new Hashtable();
    cf = new ClassFile(name, superClass);
    classes.put(name, cf);
  }

  /**
   * Create class with specified name, and specified superClass.
   */
  public void createClass(String name, Class superClass) {
    if (cf != null) classFiles.add(cf);
    fields = new Hashtable();
    cf = new ClassFile(name, superClass);
    classes.put(name, cf);
  }

  public void addInterface(String interfaceName) {
    cf.addInterface(interfaceName);
  }

  public void createMethod(Method method) {
    localVars = new Hashtable();
    MethodInfo mi = cf.addMethod(method);
    //mi.markSynthetic();
    builder = new CodeBuilder(mi);

  }

  public void finishClass() {
    if (cf != null) {
      try {
        // unused write just to initialize the constant pool 
        // so we can test each method.
        cf.getConstantPool().writeTo(new DataOutputStream(new ByteArrayOutputStream()));
      } catch (IOException ex) {
      }

      // this is a bit of hack to test for missing return statements
      // Trove will throw a RuntimeException with the message
      // Execution flows through end of method if when it tries to get
      // the byte codes for a method that is missing return statement(s).
      for (int i = 0; i < cf.getMethods().length; i++) {
        MethodInfo info = cf.getMethods()[i];
        try {
          info.getCodeAttr().getCodeBuffer().getByteCodes();
        } catch (RuntimeException ex) {
          String msg = "Internal compiler error";
          if (ex.getMessage().equals("Execution flows through end of method")) {
            msg = "Missing return statement";
          }
          CompilerException compEx = new CompilerException(msg, ex);
          compEx.setLineNum(-1);
          compEx.setBlockName(cf.getClassName() + "." + info.getName());
          throw compEx;
        }
      }
      classFiles.add(cf);
    }
    
    fields = new Hashtable();
    cf = null;
    builder = null;
    ctorBuilder = null;
  }

  public void createPublicSetMethod(String name, Class retVal, String arg,
                                    String fieldName) {
    localVars = new Hashtable();
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);

    TypeDescriptor ret = null;
    if (retVal != null) ret = new TypeDescriptor(retVal);
    TypeDescriptor[] tds = new TypeDescriptor[]{new TypeDescriptor(arg)};
    MethodInfo mi = cf.addMethod(pub, name, ret, tds);
    builder = new CodeBuilder(mi);
    builder.loadThis();
    builder.loadLocal(builder.getParameters()[0]);
    builder.storeField(fieldName, new TypeDescriptor(arg));
    builder.returnVoid();
  }

  public void createPublicMethod(String name, Class retVal, Class[] args) {
    localVars = new Hashtable();
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);

    TypeDescriptor ret = null;
    if (retVal != null) ret = new TypeDescriptor(retVal);

    MethodInfo mi = cf.addMethod(pub, name, ret, makeTypeDescriptors(args));
    builder = new CodeBuilder(mi);
  }

  private void createPublicMethod(String name, TypeDescriptor retVal, TypeDescriptor[] args) {
    localVars = new Hashtable();
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);

    MethodInfo mi = cf.addMethod(pub, name, retVal, args);
    builder = new CodeBuilder(mi);
  }

  public void createStringArrayField(String name, String[] defValues) {
    TypeDescriptor td = new TypeDescriptor(defValues.getClass());
    AccessFlags priv = new AccessFlags();
    priv.setPrivate(true);
    fields.put(name, td);
    cf.addField(priv, name, td);

    // creates the array e.g new String[10];
    ctorBuilder.loadThis();
    ctorBuilder.loadConstant(defValues.length);
    ctorBuilder.newObject(td);
    ctorBuilder.storeField(name, td);

    for (int i = 0, n = defValues.length; i < n; i++) {
      ctorBuilder.loadThis();
      ctorBuilder.loadField(name, td);
      ctorBuilder.loadConstant(i);
      ctorBuilder.loadConstant(defValues[i]);
      ctorBuilder.storeToArray(String.class);
    }
  }


  /**
   * Creates a default constuctor that calls a superclasses no-arg constructor
   * including the necessary Object();
   */
  public void createDefaultConstructor(Class[] params) {
    localVars = new Hashtable();
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);

    MethodInfo ctor = cf.addConstructor(pub, makeTypeDescriptors(params));
    //ctor.markSynthetic();
    ctorBuilder = new CodeBuilder(ctor);
    ctorBuilder.loadThis();
    ctorBuilder.invokeSuperConstructor(null);
  }

  public void createConstructorWithSuper(Class[] params) {
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);
    TypeDescriptor[] tds = makeTypeDescriptors(params);
    MethodInfo ctor = cf.addConstructor(pub, tds);
    ctorBuilder = new CodeBuilder(ctor);
    ctorBuilder.loadThis();
    LocalVariable[] ps = ctorBuilder.getParameters();
    for (int i = 0; i < ps.length; i++) {
      ctorBuilder.loadLocal(ps[i]);
    }

    ctorBuilder.invokeSuperConstructor(tds);
  }


  /**
   * Creates fields and a contructor that sets the value of those fields.
   * Also creates a no arg constructor. This no arg constructor will
   * be the target for any future constructor calls.
   */
  public void createFieldsConstructor(ArrayList fieldNames, String[] params) {
    AccessFlags priv = new AccessFlags();
    priv.setPrivate(true);

    localVars = new Hashtable();
    AccessFlags pub = new AccessFlags();
    pub.setPublic(true);

    MethodInfo ctor = cf.addConstructor(pub, makeTypeDescriptors(params));
    ctorBuilder = new CodeBuilder(ctor);
    //ctorBuilder.loadThis();
    //ctorBuilder.invokeSuperConstructor(null);
    ctorBuilder.loadThis();
    ctorBuilder.invokeConstructor(null);

    for (int i = 0; i < params.length; i++) {
      String name = (String) fieldNames.get(i);
      String c = params[i];
      TypeDescriptor td = (TypeDescriptor) fields.get(name);
      if (td == null) {
        td = new TypeDescriptor(c);
        fields.put(name, td);
        cf.addField(priv, name, td);
      }

      ctorBuilder.loadThis();
      LocalVariable[] vars = ctorBuilder.getParameters();
      ctorBuilder.loadLocal(vars[i]);
      ctorBuilder.storeField(name, td);
    }

    finishCtor();
    createDefaultConstructor(new Class[]{});

  }

  public void finishCtor() {
    ctorBuilder.returnVoid();
    ctorBuilder = null;
  }

  public void emitConstString(String str) {
    builder.loadConstant(str);
  }

  public void emitNew(Class c) {
    builder.newObject(new TypeDescriptor(c));
  }

  public void emitNew(String name) {
    builder.newObject(new TypeDescriptor(name));
  }

  public void emitDup() {
    builder.dup();
  }

  public void emitPop() {
    builder.pop();
  }

  public void createAccessor(String name, Class type, String defValue, Modifier varMod) {
    createAccessor(name, type, varMod);
    setInitialFieldValue(name, type, defValue);
  }

  /**
   * Creates the field if not already created.
   *
   * @param name
   * @param type
   * @param defValue
   */
  public void setInitialFieldValue(String name, String type, String defValue, Modifier varMod) {
    try {
      if (!fields.containsKey(name)) {
        if (varMod == Modifier.PRIVATE) createPrivateField(name, type);
        else createProtectedField(name, type);
      }
      setInitialFieldValue(name, classFromString(type), defValue);
    } catch (ClassNotFoundException ex) {
      throw new RuntimeException("Error setting field value: class " + type +
              " not found");
    }
  }

  /**
   * Assumes the field has already been created.
   *
   * @param name
   * @param type
   * @param defValue
   */
  public void setInitialFieldValue(String name, Class type, String defValue) {
    // set default value in the constructor
    if (allowableInitFields.contains(type)) {
      ctorBuilder.loadThis();
      if (type.equals(int.class)) {
        ctorBuilder.loadConstant(Integer.parseInt(defValue));
      } else if (type.equals(float.class)) {
        ctorBuilder.loadConstant(Float.parseFloat(defValue));
      } else if (type.equals(double.class)) {
        ctorBuilder.loadConstant(Double.parseDouble(defValue));
      } else if (type.equals(long.class)) {
        ctorBuilder.loadConstant(Long.parseLong(defValue));
      } else if (type.equals(boolean.class)) {
        ctorBuilder.loadConstant(Boolean.getBoolean(defValue));
      } else if (type.equals(String.class)) {
        ctorBuilder.loadConstant(defValue);
      }
    } else {
      System.out.println("Type: " + type.getName() + " not supported as default value");
      return;
    }

    ctorBuilder.storeField(name, (TypeDescriptor) fields.get(name));
  }

  public void createAccessor(String name, String type, String defValue, Modifier varMod) {
    try {
      createAccessor(name, classFromString(type), defValue, varMod);
    } catch (ClassNotFoundException ex) {
      throw new RuntimeException("Error creating accessor: class " + type +
              " not found");
    }
  }

  public void createGetMethodRetField(String name, String fieldName, Class retType) {
    createPublicMethod(name, retType, new Class[]{});
    builder.loadThis();
    builder.loadField(fieldName, new TypeDescriptor(retType));
    builder.returnValue(retType);
    builder = null;
  }

  /**
   * Creates a class field with the specified name, and get and set methods
   * for that field.
   */
  public void createAccessor(String name, Class type, Modifier ivarMod) {
    if (ivarMod == Modifier.PRIVATE) createPrivateField(name, type);
    else createProtectedField(name, type);
    String capName = Character.toUpperCase(name.charAt(0)) + name.substring(1);
    this.createPublicMethod("set" + capName, null, new Class[]{type});
    builder.loadThis();
    builder.loadLocal(builder.getParameters()[0]);
    builder.storeField(name, new TypeDescriptor(type));
    builder.returnVoid();

    this.createPublicMethod("get" + capName, type, null);
    builder.loadThis();
    builder.loadField(name, new TypeDescriptor(type));
    builder.returnValue(type);

    builder = null;
  }

  private void createAccessor(String name, TypeDescriptor type, Modifier varMod) {
    if (varMod == Modifier.PRIVATE) createPrivateField(name, type);
    else createProtectedField(name, type);
    String capName = Character.toUpperCase(name.charAt(0)) + name.substring(1);
    this.createPublicMethod("set" + capName, null, new TypeDescriptor[]{type});
    builder.loadThis();
    builder.loadLocal(builder.getParameters()[0]);
    builder.storeField(name, type);
    builder.returnVoid();

    this.createPublicMethod("get" + capName, type, new TypeDescriptor[]{});
    builder.loadThis();
    builder.loadField(name, type);
    builder.returnValue(Object.class);

    builder = null;
  }

  public void createAccessor(String name, String type, Modifier varMod) {
    try {
      createAccessor(name, classFromString(type), varMod);
    } catch (ClassNotFoundException ex) {
      String fqnType = getDynamicClassType(type);
      if (fqnType != null) {
        // create with only strings not class objects for type info
        createAccessor(name, new TypeDescriptor(fqnType), varMod);
      } else {
        throw new RuntimeException("Error creating accessor: class " + type +
                " not found");
      }
    }
  }

  public String getDynamicClassType(String type) {
    Map dynTypes = (Hashtable) details.get(EvolverBean.DYN_TYPE_LIST);
    if (!dynTypes.containsKey(type)) {
      // can we make a dynamic class from this if its is not
      // fully qualified
      NQPyClass nqpyClass = NQPyClassFactory.makeJavaClass(type);
      if (nqpyClass != null && nqpyClass instanceof NQPyDynamicClass) {
        return nqpyClass.getJClassName();
      }
      
      return null;
    }
    
    return type;
  }

  public void createPrivateField(String name, Class type) {
    AccessFlags priv = new AccessFlags();
    priv.setPrivate(true);

    TypeDescriptor td = new TypeDescriptor(type);
    fields.put(name, td);
    cf.addField(priv, name, td);
  }

  /**
   * Type must be fully qualified name.
   *
   * @param name
   * @param type
   */
  public void createPrivateField(String name, String type) {
    try {
      TypeDescriptor typeDescriptor;
      String fqnType = getDynamicClassType(type);
      if (fqnType != null) {
        typeDescriptor = new TypeDescriptor(fqnType);
      } else {
        typeDescriptor = new TypeDescriptor(classFromString(type));
      }
      createPrivateField(name, typeDescriptor);
    } catch (ClassNotFoundException ex) {
      throw new RuntimeException("Error creating field: Unable to find class for '" +
              type + "'");
    }
  }

  private void createPrivateField(String name, TypeDescriptor type) {
    AccessFlags priv = new AccessFlags();
    priv.setPrivate(true);
    fields.put(name, type);
    cf.addField(priv, name, type);
  }
  
  /**
   * Type must be fully qualified name.
   *
   * @param name
   * @param type
   */
  public void createProtectedField(String name, String type) {
    try {
      TypeDescriptor typeDescriptor;
      String fqnType = getDynamicClassType(type);
      if (fqnType != null) {
        typeDescriptor = new TypeDescriptor(fqnType);
      } else {
        typeDescriptor = new TypeDescriptor(classFromString(type));
      }
      createProtectedField(name, typeDescriptor);
    } catch (ClassNotFoundException ex) {
      throw new RuntimeException("Error creating field: Unable to find class for '" +
              type + "'");
    }
  }
  
  private void createProtectedField(String name, TypeDescriptor type) {
    AccessFlags prot = new AccessFlags();
    prot.setProtected(true);
    fields.put(name, type);
    cf.addField(prot, name, type); 
  }

  public void initStringField(String name, String val, boolean inSuper) {
    ctorBuilder.loadThis();
    ctorBuilder.loadConstant(val);
    TypeDescriptor td = new TypeDescriptor(String.class);
    if (inSuper)
      ctorBuilder.storeField(cf.getSuperClassName(), name, td);
    else
      ctorBuilder.storeField(name, td);
  }

  public void createProtectedField(String name, Class type) {
    AccessFlags prot = new AccessFlags();
    prot.setProtected(true);

    TypeDescriptor td = new TypeDescriptor(type);
    fields.put(name, td);
    cf.addField(prot, name, td);
  }

  public void emitInvokeConstructor(Class clazz, Class[] argTypes) {
    TypeDescriptor[] types = makeTypeDescriptors(argTypes);
    builder.invokeConstructor(clazz.getName(), types);
  }

  public void emitInvokeConstructor(String name, Class[] argTypes) {
    TypeDescriptor[] types = makeTypeDescriptors(argTypes);
    builder.invokeConstructor(name, types);
  }

  public void emitInvokeConstructor(String name, String[] argTypes) {
    TypeDescriptor[] types = makeTypeDescriptors(argTypes);
    builder.invokeConstructor(name, types);
  }

  public void emitThis() {
    builder.loadThis();
  }

  public void emitInvokeSuper(Method m) {
    builder.loadThis();
    builder.invokeSuper(m);
  }

  /**
   * Assumes field has already been defined and can be found in
   * the fields hashtable.
   */
  public void emitStoreValInField(String name, String defValue) {
    Class type = ((TypeDescriptor) fields.get(name)).getClassArg();
    builder.loadThis();

    if (type.equals(int.class)) {
      builder.loadConstant(Integer.parseInt(defValue));
    } else if (type.equals(float.class)) {
      builder.loadConstant(Float.parseFloat(defValue));
    } else if (type.equals(double.class)) {
      builder.loadConstant(Double.parseDouble(defValue));
    } else if (type.equals(long.class)) {
      builder.loadConstant(Long.parseLong(defValue));
    } else if (type.equals(boolean.class)) {
      builder.loadConstant(Boolean.getBoolean(defValue));
    } else if (type.equals(String.class)) {
      builder.loadConstant(defValue);
    } else {
      System.out.println("Type: " + type.getName() + " not supported as default value");
      builder.pop();
      return;
    }

    builder.storeField(name, (TypeDescriptor) fields.get(name));
  }

  public void emitInvokeVirtual(String methodName, Class retType, Class[] args) {
    TypeDescriptor[] tds = makeTypeDescriptors(args);
    builder.invokeVirtual(methodName, new TypeDescriptor(retType), tds);
  }

  public void emitInvokeVirtual(Class clazz, String methodName, Class retType,
                                Class[] argTypes) {
    TypeDescriptor[] tds = makeTypeDescriptors(argTypes);
    if (retType == null) {
      builder.invokeVirtual(clazz.getName(), methodName, null, tds);
    } else {
      builder.invokeVirtual(clazz.getName(), methodName,
              new TypeDescriptor(retType), tds);
    }
  }

  public void emitInvokeStatic(Class clazz, String methodName, Class retType,
                               Class[] argTypes) {
    TypeDescriptor[] tds = makeTypeDescriptors(argTypes);
    builder.invokeStatic(clazz.getName(), methodName,
            new TypeDescriptor(retType), tds);
  }

  public void emitLoadFieldFromVar(String varName, String fieldName, Class fieldType) {
    LocalVariable variable = (LocalVariable) localVars.get(varName);
    builder.loadLocal(variable);
    builder.loadField(variable.getType().getClassName(), fieldName,
            new TypeDescriptor(fieldType));
  }

  /**
   * Generates code that stores a value into a field of the class
   * of the localVariable. Assumes that the class and the new value
   * are already on the stack.
   */
  public void emitStoreFieldFromVar(String varName, String fieldName, Class fieldType)
          throws NoSuchFieldException {
    LocalVariable variable = (LocalVariable) localVars.get(varName);
    builder.storeField(variable.getType().getClassName(),
            fieldName, new TypeDescriptor(fieldType));
  }

  /**
   * Generates code that pushes the value of the specified class'
   * static field on the stack.
   */
  public void emitLoadStaticField(Class jclass, String fieldName, Class fieldType) {
    builder.loadStaticField(jclass.getName(), fieldName, new TypeDescriptor(fieldType));
  }

  /**
   * Push System.out on the stack. This must be followed by an emit*
   * that places the object on the stack, and then an emitPrint*End() call
   * which will invoke println on the System.out.
   */
  public void emitPrintStart() {

    builder.loadStaticField("java.lang.System", "out",
            new TypeDescriptor(java.io.PrintStream.class));
  }


  /**
   * Calls println on System.out. Expects System.out and a stack entry with
   * the specified type.
   */
  public void emitPrintEnd(Class clazz) {
    TypeDescriptor td = new TypeDescriptor(clazz);
    builder.invokeVirtual("java.io.PrintStream", "print", null,
            new TypeDescriptor[]{td});
  }

  /**
   * Calls println on System.out. Expects System.out and a stack entry with
   * the specified type.
   */
  public void emitPrintlnEnd(Class clazz) {
    TypeDescriptor td = new TypeDescriptor(clazz);
    builder.invokeVirtual("java.io.PrintStream", "println", null,
            new TypeDescriptor[]{td});
  }

  /**
   * Creates a local variable with the specified name, if no such
   * variable has been created
   */
  public void emitLocalVar(String name, Class clazz) {
    // TO DO: some warning if already defined?
    if (!localVars.containsKey(name)) {
      LocalVariable var =
              builder.createLocalVariable(name, new TypeDescriptor(clazz));
      localVars.put(name, var);
    }
  }

  public void emitStoreLocal(String name) {
    LocalVariable v = (LocalVariable) localVars.get(name);
    builder.storeLocal(v);
  }

  public void emitLoadLocal(String name) {
    LocalVariable v = (LocalVariable) localVars.get(name);
    builder.loadLocal(v);
  }

  public void emitIntToDouble() {
    builder.convert(int.class, double.class);
  }

  public void emitConstInt(int val) {
    builder.loadConstant(val);
  }

  public void emitConstDouble(double val) {
    builder.loadConstant(val);
  }

  /**
   * Creates a label.
   */
  public Label createLabel() {
    return builder.createLabel();
  }

  /**
   * Emits an unconditional jump to the specified branch.
   */
  public void emitGoto(Label branch) {
    builder.branch(branch);
  }

  /**
   * Emits bytecode that compares int on stack to zero. If int != 0
   * branch to specified label.
   */
  public void emitIfIntNotEqZero(Label branch) {
    builder.ifZeroComparisonBranch(branch, "!=");
  }

  /**
   * Emits bytecode that compares int on stack to zero. If int != 0
   * branch to specified label.
   */
  public void emitIfDoubleNotEqZero(Label branch) {
    builder.loadConstant(0d);

    // val == 0 then int 0 on stack, so branch if val on stack
    // != 0
    builder.math(Opcode.DCMPG);
    builder.ifZeroComparisonBranch(branch, "!=");
  }

  /**
   * Two doubles on stack. Compare them. If value1 > value2, 1 pushed on
   * stack. If equal, 0 pushed on stack. If value1 < value2, -1 pushed on
   * stack.
   */
  public void emitDoubleLessThan() {
    builder.math(Opcode.DCMPL);
  }

  public void emitIfNotNull(Label branch) {
    builder.ifNullBranch(branch, false);
  }

  public void emitIfIntEquals(Label branch) {
    builder.ifComparisonBranch(branch, "==");
  }

  /**
   * Pop two nums from the stack, adds them, and leaves the result.
   */
  public void emitAdd(String type) {
    if (type.equals(INT)) {
      builder.math(Opcode.IADD);
    } else {
      builder.math(Opcode.DADD);
    }
  }

  public void emitLoadField(String fname, Class fieldClass, boolean inSuper) {
    builder.loadThis();
    if (inSuper) {
      builder.loadField(cf.getSuperClassName(), fname,
              new TypeDescriptor(fieldClass));
    } else {
      builder.loadField(fname, new TypeDescriptor(fieldClass));
    }
  }

  public void emitIntInc(String varName, int amount) {
    builder.integerIncrement((LocalVariable) localVars.get(varName), amount);
  }

  public void emitIfIntLt(Label branch) {
    builder.ifComparisonBranch(branch, "<");
  }

  public void emitInitStringtArrayInCtor(int length, String arrayName) {
    TypeDescriptor t = new TypeDescriptor(String[].class);
    ctorBuilder.loadThis();
    ctorBuilder.loadConstant(length);
    ctorBuilder.newObject(t);
    //System.out.println(arrayName);
    ctorBuilder.storeField(cf.getSuperClassName(), arrayName, t);
  }

  /**
   * Adds the specified String value to an array with the specified name.
   */
  public void emitAddStringToCtorArray(String val, int index, String arrayName) {
    TypeDescriptor t = new TypeDescriptor(String[].class);
    ctorBuilder.loadThis();
    ctorBuilder.loadField(cf.getSuperClassName(), arrayName, t);
    ctorBuilder.loadConstant(index);
    ctorBuilder.loadConstant(val);
    ctorBuilder.storeToArray(String.class);
  }

  public void emitReturn(Class clazz) {
    if (clazz != null) {
      builder.returnValue(clazz);
    } else {
      builder.returnVoid();
    }
  }

  public void emitReturnField(String name) {
    TypeDescriptor td = (TypeDescriptor) fields.get(name);
    builder.loadThis();
    builder.loadField(name, td);
    builder.returnValue(td.getClassArg());
  }

  /**
   * Creates classes from the ClassFiles and adds them to the loader.
   */
  public void addClasses() {
    for (int i = 0; i < classFiles.size(); i++) {
      ClassFile cf = (ClassFile) classFiles.get(i);
      loader.addClass(cf.getClassName(), cf);
    }
  }

  public void writeToFile(String name, String dir) {
    addClasses();
    loader.writeClass(name, dir);
  }

  public Class loadClass(String name, String dir) throws
          ClassNotFoundException {
    String cpath = System.getProperty("java.class.path");
    String cSep = System.getProperty("path.separator");
    String fSep = java.io.File.separator;

    StringTokenizer tok = new StringTokenizer(cpath, cSep);
    URL[] urls = new URL[tok.countTokens() + 1];
    int count = 0;
    while (tok.hasMoreTokens()) {
      String path = tok.nextToken();
      //System.out.println(path);
      if (!path.endsWith("jar") && !path.endsWith(fSep)) {
        path = path + fSep;
      }
      URL url = null;
      try {
        url = new URL("file://" + path);
        urls[count] = url;
      } catch (MalformedURLException ex) {
        ex.printStackTrace();
      }
      count++;
    }

    URL url = null;
    try {
      url = new URL("file://" + dir + fSep);
      urls[count] = url;
    } catch (MalformedURLException ex) {
      ex.printStackTrace();
    }

    URLClassLoader uLoader = new URLClassLoader(urls, loader);

    Class c = null;
    try {
      c = uLoader.loadClass(name);
    } catch (ClassNotFoundException ex) {
      ex.printStackTrace();
      throw ex;
    }

    return c;
  }

  public Object getDetail(Object key) {
    return details.get(key);
  }

  public void putDetail(Object key, Object val) {
    details.put(key, val);
  }
}
