package uchicago.src.simbuilder.io;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;

/**
 * Parses an nqpy template file into its build, setup and display templates.
 * A template file has the following format<p>
 * <pre>
 * ## BUILD ##
 * nqpy build statments
 * <p/>
 * ## DISPLAY ##
 * nqpy display statements
 * <p/>
 * ## SETUP ##
 * nqpy setup statments
 * </pre>
 * <p/>
 * The order of the build, display and statement headers here is not fixed
 * and the presences of the header and its accompanying nqpy statements
 * is optional.<p>
 * <p/>
 * This is used by SimBuilder components to load templates for the
 * nqpy code that they produce.
 *
 * @version $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:50 $
 */

public class BeanDataLoader {

  private static final String BUILD_HEADER = "## BUILD ##";
  private static final String DISPLAY_HEADER = "## DISPLAY ##";
  private static final String SETUP_HEADER = "## SETUP ##";
  private static final String XML_OPEN_HEADER = "## XML OPEN ##";
  private static final String XML_CLOSE_HEADER = "## XML CLOSE ##";

  private Map headerMap = new HashMap();

  /**
   * Creates a BeanDataLoader by loading the  bean template file for the specified
   * Class. The template file has the same fully qualified name as the class with .data suffix.
   * For example, the class uchicago.src.simbuilder.beans.SequenceGraph
   * would have a corresponding nqpy template file at
   * /uchicago/src/simbuilder/beans/SequenceGraph.data<p>
   * <p/>
   * The BeanDataLoader will look for the following default headers and
   * their template statements:<p>
   * <ul>
   * <li>## BUILD ##
   * <li>## DISPLAY ##
   * <li>## SETUP ##"
   * <li>## XML OPEN ##
   * <li>## XML CLOSE ##
   * </ul>
   *
   * @param clazz the class to load the nqpy template file for
   * @throws IOException if the file cannot be found
   */
  public BeanDataLoader(Class clazz) throws IOException {
    this(clazz, new String[0]);
  }

  /**
   * Creates a BeanDataLoader by loading the  bean template file for the specified
   * Class. The template file has the same fully qualified name as the class with .data suffix.
   * For example, the class uchicago.src.simbuilder.beans.SequenceGraph
   * would have a corresponding nqpy template file at
   * /uchicago/src/simbuilder/beans/SequenceGraph.data<p>
   * <p/>
   * The BeanDataLoader will look for the following default headers and
   * their template statements:<p>
   * <ul>
   * <li>## BUILD ##
   * <li>## DISPLAY ##
   * <li>## SETUP ##"
   * <li>## XML OPEN ##
   * <li>## XML CLOSE ##
   * </ul>
   * <p/>
   * together with any headers specified in the headerList. Note that
   * the Strings in the headerList should not match any of the template
   * lines.
   *
   * @param clazz      the class to load the nqpy template file for
   * @param headerList an array of user specifiable headers to find template
   *                   statements for
   * @throws IOException if the file cannot be found
   */
  public BeanDataLoader(Class clazz, String[] headerList) throws IOException {
    headerMap.put(BUILD_HEADER, "");
    headerMap.put(DISPLAY_HEADER, "");
    headerMap.put(SETUP_HEADER, "");
    headerMap.put(XML_OPEN_HEADER, "");
    headerMap.put(XML_CLOSE_HEADER, "");

    for (int i = 0; i < headerList.length; i++) {
      headerMap.put(headerList[i], "");
    }

    String path = clazz.getName();
    path = "/" + path.replace('.', '/') + ".data";
    BufferedReader reader = new BufferedReader(new InputStreamReader(clazz.getResourceAsStream(path)));

    String line = reader.readLine();
    if (line != null) line = line.trim();

    // remove any comments at start
    while (line != null && !headerMap.containsKey(line)) {
      line = reader.readLine();
      if (line != null) line = line.trim();
    }
    // now after some header
    String curHeader = line;
    line = reader.readLine();
    StringBuffer b = new StringBuffer();
    while (line != null) {
      String header = line.trim();
      if (headerMap.containsKey(header)) {
        headerMap.put(curHeader, b.toString());
        curHeader = header;
        b = new StringBuffer();
      } else {
        b.append(line);
        b.append("\n");
      }

      line = reader.readLine();
    }

    headerMap.put(curHeader, b.toString());
    reader.close();
  }

  /**
   * Gets the build statements from the loaded  template file.
   */
  public String getBuild() {
    return (String) headerMap.get(BUILD_HEADER);
  }

  /**
   * Gets the setup statements from the loaded  template file.
   */
  public String getSetup() {
    return (String) headerMap.get(SETUP_HEADER);
  }

  /**
   * Gets the display statements from the loaded  template file.
   */
  public String getDisplay() {
    return (String) headerMap.get(DISPLAY_HEADER);
  }

  /**
   * Gets the XML open statements from the loaded  template file.
   */
  public String getXMLOpen() {
    return (String) headerMap.get(XML_OPEN_HEADER);
  }

  /**
   * Gets the XML close statements from the loaded  template file.
   */
  public String getXMLClose() {
    return (String) headerMap.get(XML_CLOSE_HEADER);
  }

  /**
   * Gets the template statements for the specified header.
   *
   * @param header the header
   */
  public String getTemplateFor(String header) {
    return (String) headerMap.get(header);
  }
}
