package uchicago.src.simbuilder.property;

import uchicago.src.simbuilder.beans.ConstrainedPropertyEditor;

import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.table.*;
import java.awt.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyDescriptor;
import java.beans.PropertyEditor;
import java.util.Hashtable;

/**
 * A Swing ColumnModel for displaying bean propeties via editor.
 * <p/>
 * $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:50 $
 */

public class PropertyColumnModel extends DefaultTableColumnModel {

  private static final String PROP_COL_LABEL = "Property";
  private static final String VAL_COL_LABEL = "Value";

  private static final int MIN_COL_WIDTH = 150;

  public PropertyColumnModel() {
    TableColumn column;

    column = new TableColumn(0);
    column.setHeaderValue(PROP_COL_LABEL);
    column.setPreferredWidth(MIN_COL_WIDTH);
    column.setCellRenderer(new PropertyNameRenderer());
    addColumn(column);

    column = new TableColumn(1);
    column.setHeaderValue(VAL_COL_LABEL);
    column.setPreferredWidth(MIN_COL_WIDTH * 2);
    PropertyValueCell cell = new PropertyValueCell();
    column.setCellRenderer(cell);
    column.setCellEditor(cell);
    addColumn(column);
  }

  class PropertyNameRenderer extends DefaultTableCellRenderer {

    public Component getTableCellRendererComponent(JTable table, Object value,
                                                   boolean isSelected, boolean hasFocus, int row, int col) {
      PropertyTableModel model = (PropertyTableModel) table.getModel();
      PropertyDescriptor desc = model.getPropertyDescriptor(row);
      setToolTipText(desc.getShortDescription());
      //setBackground(UIManager.getColor("control"));
      return super.getTableCellRendererComponent(table, value,
              isSelected, hasFocus,
              row, col);
    }
  }

  class PropertyValueCell extends AbstractCellEditor implements
          TableCellRenderer, TableCellEditor, PropertyChangeListener {
    private PropertyEditor curEditor;
    private DefaultTableCellRenderer renderer;
    private DefaultCellEditor cellEditor;
    private Hashtable editors = new Hashtable();
    private Hashtable pEditors = new Hashtable();
    private Border selectedBorder, emptyBorder;

    public PropertyValueCell() {
      renderer = new DefaultTableCellRenderer();
      cellEditor = new DefaultCellEditor(new JTextField());
    }

    private void showError(String message, Exception ex) {
      ex.printStackTrace();
    }

    // TableCellRenderer methods

    public Component getTableCellRendererComponent(JTable table, Object value,
                                                   boolean isSelected, boolean hasFocus, int row, int col) {
      
      PropertyEditor editor = getEditor(table, row, true);
      if (editor != null) {
        if (editor instanceof ConstrainedPropertyEditor) {
          ((ConstrainedPropertyEditor) editor).reInit();
        }

        if (editor instanceof TextFieldEditor) {
          //renderer.setBackground(UIManager.getColor("control"));
          renderer.setHorizontalAlignment(SwingConstants.RIGHT);
          JLabel label = (JLabel) renderer.getTableCellRendererComponent(table, value,
                  isSelected,
                  hasFocus, row, col);
          label.setFont(label.getFont().deriveFont(Font.BOLD));
          return label;
        } else {

          editor.setValue(value);
          Component comp = getEditorComponent(editor, isSelected);
          if (comp != null) {
            return comp;
          }
        }
        
      }
     
      return renderer.getTableCellRendererComponent(table, value, isSelected,
              hasFocus, row, col);
    }

    // TableCellEditor methods
    public Component getTableCellEditorComponent(JTable table, Object value,
                                                 boolean isSelected, int row, int col) {
      if (selectedBorder == null) {
        createBorders(table);
      }

      PropertyTableModel model = (PropertyTableModel) table.getModel();
      Class type = model.getPropertyType(row);
      curEditor = null;
      if (type != null) {
        curEditor = (PropertyEditor) pEditors.get(type);
        if (curEditor == null) {
          PropertyEditor ed = getEditor(table, row, false);
          if (ed != null) {
            try {
              curEditor = (PropertyEditor) ed.getClass().newInstance();
              curEditor.addPropertyChangeListener(this);
              curEditor.addPropertyChangeListener((PropertyChangeListener) model.getBean());
              //pEditors.put(type, curEditor);
            } catch (IllegalAccessException ex) {
              showError("Error instantiating editor", ex);
            } catch (InstantiationException ex) {
              showError("Error instantiating editor", ex);
            }
          }
        }
      }

      if (curEditor != null) {
        if (curEditor instanceof ConstrainedPropertyEditor) {
          ((ConstrainedPropertyEditor) curEditor).reInit();
        }
        curEditor.setValue(value);
        Component comp = getEditorComponent(curEditor, isSelected);
        if (comp != null) {
          return comp;
        }
      }

      // couldn't get a handle to an editor so use default cellEditor
      return cellEditor.getTableCellEditorComponent(table, value, isSelected,
              row, col);
    }

    private Component getEditorComponent(PropertyEditor editor,
                                         boolean isSelected) {
      Component comp = null;
      comp = editor.getCustomEditor();
      if (comp != null) {
        comp.setEnabled(isSelected);
        /*
if (comp instanceof JComponent) {
if (isSelected) ((JComponent)comp).setBorder(selectedBorder);
else ((JComponent)comp).setBorder(emptyBorder);
}
        */
      }

      return comp;
    }

    private PropertyEditor getEditor(JTable table, int row,
                                     boolean forRenderer) {

      PropertyTableModel model = (PropertyTableModel) table.getModel();
      Class type = model.getPropertyType(row);
      PropertyEditor editor = null;
      if (type != null) {
        editor = (PropertyEditor) editors.get(type);

        if (editor == null) editor = model.getPropertyEditor(row);
        if (editor != null && forRenderer) editors.put(type, editor);
      }

      return editor;
    }

    private void createBorders(JTable table) {
      selectedBorder = BorderFactory.createLineBorder(table.getSelectionBackground(), 2);
      emptyBorder = BorderFactory.createEmptyBorder(1, 1, 1, 1);
    }

    // AbstractCellEditor methods
    public Object getCellEditorValue() {
      Object obj = null;
      try {
        if (curEditor != null) {
          obj = curEditor.getValue();
        } else {
          obj = cellEditor.getCellEditorValue();
        }
      } catch (IllegalArgumentException ex) {
        //System.out.println(ex.getMessage());
        //uchicago.src.sim.util.SimUtilities.showMessage(ex.getMessage());
        cancelCellEditing();
      }

      //System.out.println("obj from getCellEditorValue: " + obj);
      //System.out.println("editor: " + curEditor);
      return obj;
    }

    // property change handler
    public void propertyChange(PropertyChangeEvent evt) {
      //boolean bad = listenerList.getListenerCount() == 0;
      stopCellEditing();
      /*
      if (bad) {
        // we will be here the user has selected another component 
        // while the editor was open
        System.out.println("curEditor = " + curEditor);
        //EvolverPropertyEditor editor = (EvolverPropertyEditor)curEditor
      }
      */
    }
  }
}
