package uchicago.src.simbuilder.property;

import uchicago.src.simbuilder.beans.EvolverBean;

import javax.swing.table.AbstractTableModel;
import java.beans.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Hashtable;

/**
 * A TableModel for displaying and editing java bean properties. The
 * general idea here is that a the bean is introspected and its
 * PropertyDescriptors are index to rows in the table. The first
 * column displays the property name via PropertyDescriptor.getDisplayName()
 * while the second column calls the appropriate get method on the bean
 * and displays the result via the appropriate editor.<p>
 * <p/>
 * For example, bean X has a String property "myName" with the appropriate
 * accessor methods getMyName, and setMyName. The value of this
 * property is displayed by calling getMyName and passing the result
 * to a StringPropertyEditor which displays this result in a JTextField.
 * Note that the actual display via an editor is done through
 * PropertyColumnModel
 * <p/>
 * $Revision: 1.1.1.1 $ $Date: 2004/09/30 03:40:50 $
 */
public class PropertyTableModel extends AbstractTableModel {

  private static final int NAME_COL = 0;
  private static final int VAL_COL = 1;
  private static final int NUM_COL = 2;

  private String[] colNames = {"Property", "Value"};

  private PropertyDescriptor[] descriptors;
  private Object bean;
  private BeanInfo info;

  private Hashtable editors = new Hashtable();

  public PropertyTableModel() {
    super();
    editors.put(Object.class, new ObjectPropertyEditor());
  }

  public void setObject(Object bean) {
    this.bean = bean;
    try {
      info = Introspector.getBeanInfo(bean.getClass());
      if (info != null) {
        descriptors = info.getPropertyDescriptors();

        // all this up to makeTable() is to remove write-only properties
        // from the array of PropertyDescriptors.
        int nullCount = 0;
        for (int i = 0; i < descriptors.length; i++) {
          if (descriptors[i].getReadMethod() == null) {
            nullCount++;
            descriptors[i] = null;
          }
        }

        if (nullCount != 0) {
          int index = 0;
          PropertyDescriptor[] tmp =
                  new PropertyDescriptor[descriptors.length - nullCount];
          for (int i = 0; i < descriptors.length; i++) {
            if (descriptors[i] != null) tmp[index++] = descriptors[i];
          }

          descriptors = tmp;
        }

        makeTable();
      }
    } catch (IntrospectionException ex) {
      showError("Bean error", ex);
    }
  }

  private void makeTable() {
    // sorting and filtering of properties would go here
    fireTableDataChanged();
  }

  public PropertyDescriptor getPropertyDescriptor(int row) {
    return descriptors[row];
  }

  public Class getPropertyType(int row) {
    return descriptors[row].getPropertyType();
  }
  
  public EvolverBean getBean() {
    return (EvolverBean)bean;
  }

  public PropertyEditor getPropertyEditor(int row) {

    Class cls = descriptors[row].getPropertyEditorClass();
    PropertyEditor editor = null;

    if (cls != null) {
      try {
        editor = (PropertyEditor) cls.newInstance();
      } catch (IllegalAccessException ex) {
        showError("Error creating editor", ex);
      } catch (InstantiationException ex) {
        showError("Error creating editor", ex);
      }
    } else {
      // no editor associated via the descriptor
      Class type = getPropertyType(row);
      if (type != null) {
        // check cache
        editor = (PropertyEditor) editors.get(type);
        if (editor == null) {
          editor = PropertyEditorManager.findEditor(type);
          if (editor != null) {
            editors.put(type, editor);
          }
        }

        if (editor == null) {
          editor = (PropertyEditor) editors.get(Object.class);
        }
      }
    }

    return editor;
  }


  /**
   * This should wrap an error producer and fire an error event here.
   */
  private void showError(String message, Exception ex) {
    System.out.println(message);
    ex.printStackTrace();
  }

  /**
   * This should wrap an error producer and fire an error event here.
   */
  private void showError(String message, Error ex) {
    System.out.println(message);
    ex.printStackTrace();
  }

  // AbstractTableModel methods
  public boolean isCellEditable(int row, int col) {
    if (col == VAL_COL && descriptors != null) {
      return descriptors[row].getWriteMethod() != null;
    }

    return false;
  }

  public int getRowCount() {
    return descriptors == null ? 0 : descriptors.length;
  }

  public int getColumnCount() {
    return NUM_COL;
  }

  public String getColumnName(int col) {
    return colNames[col];
  }

  public Object getValueAt(int row, int col) {

    Object val = null;

    if (col == NAME_COL) {
      val = descriptors[row].getDisplayName();
    } else {
      Method getter = descriptors[row].getReadMethod();
      if (getter != null) {
        Class[] pTypes = getter.getParameterTypes();
        Object[] args = new Object[pTypes.length];

        if (pTypes.length > 0) {
          System.out.println("get method with parameters!");
        }

        try {
          val = getter.invoke(bean, args);
        } catch (IllegalArgumentException ex) {
          showError("getter error", ex);
        } catch (IllegalAccessException ex) {
          showError("getter error", ex);
        } catch (InvocationTargetException ex) {
          showError("getter error", ex);
        } catch (NoSuchMethodError ex) {
          showError("getter error", ex);
        }
      }
    }

    return val;
  }

  public void setValueAt(Object value, int row, int col) {
    if (col == VAL_COL && descriptors != null && row < descriptors.length) {

      Method setter = descriptors[row].getWriteMethod();
      if (setter != null) {
        try {
          setter.invoke(bean, new Object[]{value});
        } catch (IllegalArgumentException ex) {
          String message = "Setter: " + setter + "\nArgument: " +
                  value.getClass().toString() + ", " + value + "\nrow: " + row +
                  "\ncol: " + col;
          showError(message, ex);
        } catch (IllegalAccessException ex) {
          String message = "Setter: " + setter + "\nArgument: " +
                  value.getClass().toString() + ", " + value + "\nrow: " + row +
                  "\ncol: " + col;

          showError(message, ex);
        } catch (InvocationTargetException ex) {
          String message = "Setter: " + setter + "\nArgument: " +
                  value.getClass().toString() + ", " + value + "\nrow: " + row +
                  "\ncol: " + col;
          showError(message, ex);
        } catch (NoSuchMethodError ex) {
          String message = "Setter: " + setter + "\nArgument: " +
                  value.getClass().toString() + ", " + value + "\nrow: " + row +
                  "\ncol: " + col;
          showError(message, ex);
        }
      }
    }
  }
}
